/**
 * @Filename XmlDocument.hpp
 * @Description Class used to parse a .xml file.
 * Creates a tree of XmlElement classes.
 * @Status Finished
 * @Version $Id: XmlDocument.hpp 5913 2013-02-10 12:15:21Z jreis $
 * @Maintainer Nelson Ramos (nmsra@mega.ist.utl.pt)
 */

#ifndef __XML_XMLDOCUMENT_H
#define __XML_XMLDOCUMENT_H



namespace mermaid
{
  namespace support
  {
    namespace xml
    {
      class XmlDocument;
    }
  }
}

#include <boost/shared_ptr.hpp>

#include "XmlElement.hpp"



namespace mermaid
{
  namespace support
  {
    namespace xml
    {
      using boost::shared_ptr;
      
      /**
      * @Class XmlDocument XmlDocument.hpp "XmlDocument.hpp"
      * @Description Class representing an XML document.
      * @Author Marco Barbosa + Nelson Ramos
      */
      class XmlDocument
      {
        public:
          /**
          * @Description Constructor.
          * @Description This method creates a new XmlDocument instance.
          * @Author Marco Barbosa + Nelson Ramos
          */
          XmlDocument();
          
          /**
          * @Description Copy constructor
          * @Author Marco Barbosa
          */
          XmlDocument (const XmlDocument& xd);
          
          /**
          * Destructor.
          * This method makes deinitializations for objects of this class.
          * @Author Marco Barbosa + Nelson Ramos
          */
          ~XmlDocument();
          
          XmlDocument* clone();
          
          /**
          * @Description Assignment operator
          * @Author Marco Barbosa
          */
          XmlDocument& operator= (const XmlDocument& xd);
          
          /**
          * @Description Parses XML file.
          * @Argument name - Name of the file to be parsed.
          * @Argument enforceDtdValidation - if true, parsing will only be
          * done if file is valid according to its DTD
          * @throws Exception If an error occurs while parsing the file
          * @Author Marco Barbosa + Nelson Ramos
          */
          void parseFile (string name, bool enforceDtdValidation = true);
          
          /**
          * @Description Parses an XML string
          * @TODO Implement DTD validation
          * @Argument xml XML string
          * @throws Exception If an error occurs while parsing the string
          * @Author Marco Barbosa + Nelson Ramos
          */
          void parseString (string xml);
          
          /**
          * Get the root element of the document.
          * @Returns bool TRUE if successful, FALSE otherwise.
          * @Author Marco Barbosa + Nelson Ramos
          */
          shared_ptr<XmlElement> getRootElement();
          
          /**
          * Checks if the document is well formed.
          * @Returns bool TRUE if document is well formed, FALSE otherwise.
          * @Author Marco Barbosa
          */
          bool isWellFormed();
          
          /**
          * Checks if the document is valid.
          * @Returns bool TRUE if document is valid, FALSE otherwise.
          * @Author Marco Barbosa
          */
          bool isValid();
          
          //! @TODO implement operator==
          
        private:
          bool isWellFormed_;
          bool isValid_;
          shared_ptr<XmlElement> rootElement_;
          
          void copyFrom (const XmlDocument& xd);
          void cleanUp();
      }; // class XmlDocument
    } // namespace xml
  } // namespace support
} // namespace mermaid

#endif
