/**
 * @Filename XmlElement.hpp
 * @Description XmlElement class definition.
 * @Status Implementing
 * @Version $Id: XmlElement.hpp 5913 2013-02-10 12:15:21Z jreis $
 * @Maintainer Nelson Ramos (nmsra@mega.ist.utl.pt)
 */

#ifndef __XML_XMLELEMENT_H_
#define __XML_XMLELEMENT_H_



namespace mermaid
{
  namespace support
  {
    namespace xml
    {
      class XmlElement;
    }
  }
}

#include <vector>

#include "XmlAttribute.hpp"
#include "XmlAttributeVector.hpp"
#include "XmlCharData.hpp"
#include "XmlElementVector.hpp"
#include "XmlItem.hpp"



typedef struct _xmlNode xmlNode;
typedef xmlNode* xmlNodePtr;



namespace mermaid
{
  namespace support
  {
    namespace xml
    {
      using std::string;
      
      /**
      * @Class XmlElement XmlElement.hpp "XmlElement.hpp"
      * @Description Class representing an XML Element.
      * @Author Marco Barbosa
      */
      class XmlElement : public XmlItem
      {
          friend class XmlDocument;
          
        public:
          /**
          * @Description Constructor.
          * @Description This method creates a new XmlElement instance.
          * @Author Marco Barbosa
          */
          XmlElement();
          
          /**
                * @Description Copy constructor
                * @Author Marco Barbosa
                */
          XmlElement (const XmlElement& e);
          
          /**
          * @Description Constructor.
          * @Description This method build an XmlElement from a string
          * @Author Marco Barbosa
          */
          XmlElement (string xmlString);
          
          /**
          * @Description Destructor.
          * @Description This method makes deinitializations for objects
          * of this class.
          * @Author Marco Barbosa
          */
          virtual ~XmlElement();
          
          virtual XmlElement* clone();
          
          /**
          * @Description Assignment operator
          * @Author Marco Barbosa
          */
          XmlElement& operator= (const XmlElement& e);
          
          /**
          * XmlElement's name getter.
          * @Returns string Returns the name of this element.
          * @Author Marco Barbosa
          */
          const string getName() const;
          
          /**
          * XmlElement's CharData getter.
          * @Returns XmlCharDataVector Returns all the CharData belonging
          * to this element
          * @Author Marco Barbosa
          */
          XmlCharDataVector getCharData();
          
          /**
          * XmlElement's CharData getter.
          * @Returns XmlCharDataVector Returns only the first CharData
          * which is not blank
          * @Author Marco Barbosa
          */
          shared_ptr<XmlCharData> getFirstNonBlankCharData();
          
          /**
          * XmlElement's children elements getter.
          * @Returns XmlElementVector Ret    urns all the child elements.
          * @Author Marco Barbosa
          */
          XmlElementVector getChildrenElements();
          
          /**
          * XmlElement's attribute getter.
          * @Returns XmlAttributeVector Returns all the attributes.
          * @Author Marco Barbosa
          */
          const XmlAttributeVector getAttributes() const;
          
          /**
          * @Description Method to convert value to string
          */
          virtual operator string() const;
          
          /**
           * @Description Method to check if XmlItem is of type XmlElement
           * @Returns True
           */
          virtual bool isXmlElement() const;
          
        protected:
        
        private:
          /**
          * Constructor.
          * This method creates a new XmlElement instance.
          * @Argument node libxml2 node pointer.
          * @Author Marco Barbosa
          */
          XmlElement (xmlNodePtr node);
          
          void buildFromLibxml2Node (xmlNodePtr node);
          
          void copyFrom (const XmlElement& e);
          void cleanUp();
          
          string name_;
          XmlCharDataVector charData_;
          XmlElementVector childrenElements_;
          XmlAttributeVector attributes_;
          XmlItemVector items_;
      }; // class XmlElement
    } // namespace xml
  } // namespace support
} // namespace mermaid

#endif
