/**
 * @Filename XmlElement.cpp
 * @Description Implementation of the XmlElement class
 * @Status Implementing
 * @Version $Id: XmlElement.cpp 5913 2013-02-10 12:15:21Z jreis $
 * @Maintainer Nelson Ramos (nmsra@mega.ist.utl.pt)
 */

#include "XmlElement.hpp"

#include <string>
#include <sstream>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include <Exception.hpp>

#include "XmlItem.hpp"



using namespace mermaid::support::xml;
using std::string;
using std::stringstream;
using mermaid::support::errorhandling::Exception;



XmlElement::XmlElement() :
  name_ ("UNNAMED"),
  charData_(),
  childrenElements_(),
  attributes_(),
  items_()
{
}



XmlElement::XmlElement (const XmlElement& e) :
  name_(),
  charData_(),
  childrenElements_(),
  attributes_(),
  items_()
{
  copyFrom (e);
}



XmlElement::XmlElement (string xml) :
  name_(),
  charData_(),
  childrenElements_(),
  attributes_(),
  items_()
{
  xmlNodePtr rootElementNode;
  
  xmlParserCtxtPtr parserCtxtPtr = xmlNewParserCtxt();
  xmlDocPtr docPtr = xmlCtxtReadMemory (parserCtxtPtr,
                                        xml.data(),
                                        xml.size(),
                                        NULL,
                                        NULL,
                                        0);
                                        
  if ( (parserCtxtPtr->wellFormed == 0) || (docPtr == NULL)) {
    XmlElement();
    throw Exception ("XmlElement::XmlElement(string xml) : unable to "
                     "build from string: \"" + xml + "\"");
  }
  
  rootElementNode = xmlDocGetRootElement (docPtr);
  
  if (rootElementNode == NULL) {
    xmlFreeDoc (docPtr);
    XmlElement();
    throw Exception ("XmlElement::XmlElement(string xml) : unable to "
                     "build from string: \"" + xml + "\"");
  }
  buildFromLibxml2Node (rootElementNode);
  
  if (parserCtxtPtr != NULL) {
    xmlFreeParserCtxt (parserCtxtPtr);
  }
  
  if (docPtr != NULL) {
    xmlFreeDoc (docPtr);
  }
}



XmlElement::XmlElement (xmlNodePtr node) :
  name_(),
  charData_(),
  childrenElements_(),
  attributes_(),
  items_()
{
  buildFromLibxml2Node (node);
}



void XmlElement::buildFromLibxml2Node (xmlNodePtr node)
{
  // read name
  name_ = string ( (char*) node->name);
  
  // get all children items
  xmlNodePtr child = node->children;
  while (child != NULL) {
    if (child->type == XML_ELEMENT_NODE) {
      shared_ptr<XmlElement> e =
        shared_ptr<XmlElement> (new XmlElement (child));
      if (!e) {
        throw Exception ("XmlElement::buildFromLibxml2Node : unable to "
                         "create XmlElement");
      }
      childrenElements_.push_back (e);
      items_.push_back (static_cast<XmlItem*> (e.get()));
      //! @todo remove dereferencing of shared_ptr
    }
    else if ( (child->type == XML_TEXT_NODE)
              || (child->type == XML_CDATA_SECTION_NODE)) {
      shared_ptr<XmlCharData> c =
        shared_ptr<XmlCharData> (new XmlCharData ( (char*) child->content));
      if (!c) {
        throw Exception ("XmlElement::buildFromLibxml2Node : unable to "
                         "create XmlCharData");
      }
      charData_.push_back (c);
      items_.push_back (static_cast<XmlItem*> (c.get()));
      //! @todo remove dereferencing of shared_ptr
    }
    child = child->next;
  }
  
  // get all attributes
  xmlAttrPtr attr = node->properties;
  while ( (attr != NULL)) {
    if (attr->type == XML_ATTRIBUTE_NODE) {
      attributes_.push_back (
        shared_ptr<XmlAttribute> (new XmlAttribute (attr)));
    }
    attr = attr->next;
  }
}



XmlElement::~XmlElement()
{
  cleanUp();
}



XmlElement* XmlElement::clone()
{
  return new XmlElement (*this);
}



XmlElement& XmlElement::operator= (const XmlElement& e)
{
  // make sure not same object
  if (this != &e) {
    cleanUp();
    copyFrom (e);
  }
  return *this;
}



const string XmlElement::getName() const
{
  return name_;
}



XmlCharDataVector XmlElement::getCharData()
{
  return charData_;
}



shared_ptr<XmlCharData> XmlElement::getFirstNonBlankCharData()
{
  XmlCharDataVector vector = getCharData();
  
  for (XmlCharDataVector::iterator it = vector.begin();
       it != vector.end();
       it++) {
    shared_ptr<XmlCharData> xc = *it;
    
    if (!xc->isBlank()) {
      return xc;
    }
  }
  
  return shared_ptr<XmlCharData>();
}



XmlElementVector XmlElement::getChildrenElements()
{
  return childrenElements_;
}



const XmlAttributeVector XmlElement::getAttributes() const
{
  return attributes_;
}



XmlElement::operator string() const
{
  stringstream ss;
  ss << "<";
  ss << getName();
  
  XmlAttributeVector attributes = getAttributes();
  
  if (attributes.size() > 0) {
    ss << " ";
    XmlAttributeVector::iterator it;
    
    for (it = attributes.begin(); it != attributes.end(); it++) {
      shared_ptr<XmlAttribute> a = *it;
      ss << a->getName();
      ss << "=\"";
      ss << a->getValue();
      ss << "\" ";
    }
  }
  
  ss << ">";
  
  XmlItemVector::const_iterator it;
  
  for (it = items_.begin(); it != items_.end(); it++) {
    XmlItem* i = *it;
    ss << ( (string) *i);
  }
  
  ss << "</";
  ss << getName();
  ss << ">";
  
  return ss.str();
}



bool XmlElement::isXmlElement() const
{
  return true;
}



void XmlElement::copyFrom (const XmlElement& e)
{
  name_ = e.name_;
  
  // copy XmlItems
  for (XmlItemVector::const_iterator iit = e.items_.begin();
       iit != e.items_.end();
       iit++) {
    XmlItem* i = *iit;
    if (i->isXmlElement()) {
      XmlElement* xe = (XmlElement*) i;
      XmlElement* ne = new XmlElement (*xe);
      childrenElements_.push_back (shared_ptr<XmlElement> (ne));
      items_.push_back ( (XmlItem*) ne);
    }
    else if (i->isXmlCharData()) {
      XmlCharData* xc = (XmlCharData*) i;
      XmlCharData* nc = new XmlCharData (*xc);
      charData_.push_back (shared_ptr<XmlCharData> (nc));
      items_.push_back ( (XmlItem*) nc);
    }
    else {
      throw Exception ("XmlElement::copyFrom : got unidentifiable item");
    }
  }
  
  // copy attributes
  XmlAttributeVector::const_iterator ait;
  for (ait = e.attributes_.begin(); ait != e.attributes_.end(); ait++) {
    shared_ptr<XmlAttribute> a = *ait;
    XmlAttribute* na = new XmlAttribute (*a);
    attributes_.push_back (shared_ptr<XmlAttribute> (na));
  }
}



void XmlElement::cleanUp()
{
  name_ = string ("UNNAMED");
  charData_ = XmlCharDataVector();
  childrenElements_ = XmlElementVector();
  attributes_ = XmlAttributeVector();
  items_ = XmlItemVector();
}
