#ifndef _SOCROB_MATH_H_
#define _SOCROB_MATH_H_

#include <cmath>
#include <cfloat>
#include <algorithm>



namespace socrob
{
  /** The inifinte blue... */
  const float INFTY = 1000000.0f;
  // Could be FLT_MAX, but it is dangerous to change a constant
  
  /** PI value (3,14..). */
  const float PI = M_PI;
  
  /** An arbitrarily small value used for comparisons */
  const float INFINITESIMAL = 1e-4f;
  // Could be FLT_EPSILON, but it is dangerous to change a constant
  
  
  
  using std::max;
  using std::min;
  
  
  
  /**
   * Compute the square value of the argument given
   * @Argument value - value to be squared
   * @Returns the computed square value
   */
  template<class T>
  inline T
  sqr (T const& value)
  {
    return value * value;
  }
  
  
  
  /**
   * Compute the maximum of the given array
   * @Argument valuesArray - array with the values to check
   * @Argument numValues - number of values in the array
   * @Returns the maximum value in the given array
   */
  template<class T>
  inline T const&
  max (T* values, std::size_t num)
  {
    return *std::max_element (values, values + num);
  }
  
  
  
  /**
   * Clip the given value to the given limits.
   * @Argument value - value to be clipped
   * @Argument minValue - lower limit
   * @Argument maxValue - upper limit
   * @Returns the clipped value
   */
  template<class T>
  inline T const&
  clipBetween (T const& value, T const& minValue, T const& maxValue)
  {
    return value < maxValue ? (value > minValue ? value : minValue) : maxValue;
  }
  
  
  
  /**
   * Adjust angle between minValue and maxValue (by adding/removing 2*PI to
   * the angle).
   * @Argument value - value to be adjusted
   * @Argument minValue - lower limit
   * @Argument maxValue - upper limit
   * @Returns the clipped angle value
   */
  template<class T>
  inline T
  clipAngleBetween (T const& value, T const& minValue, T const& maxValue)
  {
    T interval = maxValue - minValue;
    T delta = (interval / 2) + minValue;
    return remainder (value + delta, interval) + delta;
  }
  
  
  
  /**
   * Adjust angle between -PI and PI (by adding/removing 2*PI to the angle).
   * @Argument angle - angle to be adjusted
   * @Returns the clipped angle
   */
  template<class T>
  inline T
  clipAngle (T const& angle)
  {
    return remainder (angle, 2 * M_PI);
  }
  
  
  
  /**
   * Check if angle is between minValue and maxValue. All the angles should
   * be bewteen -PI and PI (the standard in the architecture). This method
   * takes into account cases such as PI beeing belonging to the interval
   * [-PI+0.1; PI-0.1].
   * @Argument value - value to be checked
   * @Argument minValue - lower limit
   * @Argument maxValue - upper limit
   * @Returns true if the given angle is between the desired values
   */
  template<class T>
  inline bool
  isAngleBetween (T const& value, T const& minValue, T const& maxValue)
  {
    T deltaMin = clipAngle (value - minValue);
    T deltaMax = clipAngle (value - maxValue);
    return (deltaMin >= 0) && (deltaMax <= 0);
  }
  
  
  
  /**
   * Check if a number is between minValue and maxValue.
   * @Argument value - value to be checked
   * @Argument minValue - lower limit
   * @Argument maxValue - upper limit
   * @Returns true if the given angle is between the desired values
   */
  template<class T>
  inline bool
  isBetween (T const& value, T const& minValue, T const& maxValue)
  {
    return (value >= minValue) && (value < maxValue);
  }
  
  
  
  /**
   * Convert given angle from degrees to radians.
   * @Argument angle - angle to be converted
   * @Returns angle in radians
   */
  template<class T>
  inline T
  deg2rad (T const& angle)
  {
    return angle * M_PI / 180;
  }
  
  
  
  /**
   * Convert given angle from radians to degrees.
   * @Argument angle - angle to be converted
   * @Returns angle in degrees
   */
  template<class T>
  inline T
  rad2deg (T const& angle)
  {
    return angle * 180 / M_PI;
  }
  
  
  
  /**
   * Calculates the L2 norm of a vector, represented as an array.
   * @Argument array - n-dimensional vector
   * @Returns L2 norm of the given vector
   */
  template<class T>
  inline T
  norm (T* array, std::size_t size)
  {
    T norm = 0;
    for (std::size_t i = 0; i < size; i++) {
      norm += sqr (array[i]);
    }
    return sqrt (norm);
  }
  
  
  
  /**
   * Calculates the L2 norm of a vector, represented as a pair.
   * @Argument array - n-dimensional float vector. Do not pass implicitly
   * casted arrays of other types than float (create a dummy float array for these cases)
   * @Returns L2 norm of the given vector
   */
  template<class T>
  inline T
  norm (std::pair<T, T> const& point)
  {
    return hypot (point.first, point.second);
  }
  
  
  
  /**
   * Verifies if a given value is, for all practical purposes, zero.
   * This is useful when handling results of long chains of arithmetic operations, that may
   * accumulate arithmetic noise and therefore fail trivial comparisons.
   *
   * Please see (or any other article about floating point comparison):
   * http://randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/
   * @Argument value - the value that must be checked
   * @Returns indicates wheter or not a value is practically zero
   */
  template<class T>
  inline bool
  isZero (T const& value)
  {
    return abs (value) <= INFINITESIMAL;
  }
  
  
  
  /**
   * Calculate the smallest difference between any two angles (angle1-angle2).
   * @Argument angle1 - the minuend
   * @Argument angle2 - the subtrahend
   * @Returns the smallest difference between the angles
   */
  template<class T>
  inline T
  angleDiff (T angle1, T angle2)
  {
    angle1 = clipAngle (angle1);
    angle2 = clipAngle (angle2);
    return clipAngle (angle1 - angle2);
  }
  
  
  
  /**
   * Check if an angle is between the interval [ minValue, maxValue ].
   * The interval starts at minValue and finish at maxValue and it contains
   * all the angles from minValue until maxValue in a anticlokwise direction.
   * For example, the interval [-100º, -150º] (in the method angles are in rad)
   * contains {-100º,-90º,-45º,0º,45º,90º,135º,180º,-180º,-170º,-150º} and doesn't
   * contain {-149º,-130º,-120º,-110º,-101º}.
   * All the angles doesn't need to be between -PI and PI, the adjustment is done
   * at the beginning of the method with clipAngle.
   * @Argument value - angle to be checked
   * @Argument minValue - Beginning of the interval
   * @Argument maxValue - Ending of the interval
   * @Returns true if the given angle is between the desired values
   */
  template<class T>
  inline bool
  isAngleStrictlyBetween (T value, T minValue, T maxValue)
  {
    value = clipAngle (value);
    minValue = clipAngle (minValue);
    maxValue = clipAngle (maxValue);
    if (minValue > maxValue) {
      if ( ( (value >= minValue) && (value <= M_PI)) || ( (value >= -M_PI) && (value <= maxValue))) {
        return true;
      }
      else {
        return false;
      }
    }
    else if ( (value >= minValue) && (value <= maxValue)) {
      return true;
    }
    else {
      return false;
    }
  }
  
  
  
  /**
   * Compute the signum function of the given value.
   * Returns: -1 if value < 0; 0 if value = 0; 1 if value > 0;
   * @Argument value - number to check
   * @Returns the signum value
   */
  template<class T>
  inline T
  sign (T const& value)
  {
    return value < 0 ? -1 : (value > 0 ? 1 : 0);
  }
}

#endif
