/**
 * @Filename FtdiDevice.cpp
 * @Description Helper to use FTDI devices
 * @Status unknown
 * @Version $Id: FtdiDevice.cpp 6070 2013-06-03 14:38:14Z jreis $
 * @Maintainer Joao Reis (joaocgreis@gmail.com)
 */

#include "FtdiDevice.hpp"

#include <boost/thread/shared_mutex.hpp>



#undef FTDI_DEBUG
// #define FTDI_DEBUG



FtdiDevice::
FtdiDevice (string const& deviceName,
            long const usecTimeout,
            float const rate) :
  rate_ (rate),
  connected_ (false),
  timeout_ (usecTimeout),
  serials_(),
  ctx_(),
  name_ (deviceName)
{
}



bool
FtdiDevice::
initialize()
{
  connected_ = connectDevice();
  
  if (! connected_) {
    ROS_FATAL_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " Could not open " + name_ + " device");
    setServiceDefaultActive (true, 2);
    return false;
  }
  
  setServiceDefaultActive (true, rate_);
  
  return true;
}



namespace
{
  boost::shared_mutex mutex_;
}



bool
FtdiDevice::
connectDevice()
{
  boost::lock_guard<boost::shared_mutex> _ (mutex_);
  
  ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "FtdiDevice::connectDevice()");
  
  using namespace Ftdi;
  
  List* list = List::find_all (0x0403, 0x6001);
  
  ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "Found " << list->size() << " devices.");
  
  for (List::iterator it = list->begin(); it != list->end(); it++) {
    string serial = it->serial();
    
    if (serial == "") {
      ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " FtdiDevice cannot read a device serial. Please check "
                              "http://socrob.isr.ist.utl.pt/private/wiki/doku.php?id=msl:"
                              "software:robot_installation#udev_configuration");
      continue;
    }
    
    map<string, int>::iterator iter = serials_.find (serial);
    if (iter == serials_.end()) {
      continue;
    }
    
    ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "Found a device with serial " + serial);
    
    int ret = it->open();
    if (ret == 0) {
      ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "Open OK. Connected to device with serial " + serial);
      ctx_ = *it;
      
      int baudrate = serials_[serial];
      
      int ret;
      ret = ctx_.set_baud_rate (baudrate);
      if (ret < 0) {
        ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " set_baud_rate failed : " << ctx_.error_string());
      }
      
      if (baudrate == 115200) {
        ret = ctx_.set_latency (1);
        if (ret < 0) {
          ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " set_latency failed : " << ctx_.error_string());
        }
      }
      else {
        ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "FtdiDevice: Leaving latency untouched because baudrate is not 115200.");
      }
      
      delete list;
      
      // afterConnect may disconnect the device.
      connected_ = true;
      afterConnect();
      return connected_;
    }
    else {
      ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "Open failed returning " << ret);
    }
    
    it->close();
  }
  
  ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "Could not connect.");
  
  delete list;
  return false;
}



void
FtdiDevice::
update()
{
  if (connected_) {
    update3();
    
    if (! connected_) {
      setServiceDefaultActive (true, 2);
    }
  }
  else {
    if ( (connected_ = connectDevice()) == true) {
      if (rate_ >= 0) {
        setServiceDefaultActive (true, rate_);
      }
      else {
        setServiceDefaultActive (true, 1);  //to verify the conditions
      }
    }
    else {
      setServiceDefaultActive (true, 1);
    }
  }
}



void
FtdiDevice::
addSerial (string const& serial, int baudrate)
{
  serials_[serial] = baudrate;
}



void
FtdiDevice::
write (uint8_t* const buf,
       unsigned const count)
{
  if (! connected_) {
    return;
  }
  
  boost::shared_lock<boost::shared_mutex> _ (mutex_);
  
  TimeStamp tstart;
  
  unsigned done = 0;
  do {
    {
      TimeStamp tnow;
      long elapsed;
      TimeStamp::timeDiff (tnow, tstart, &elapsed);
      if (elapsed > timeout_) {
        ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " FTDI write timeout.");
        ftdiError();
        return;
      }
    }
    
    int ret = ctx_.write (buf + done, count - done);
    if (ret < 0) {
      ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " FTDI write error: " << ctx_.error_string());
      ftdiError();
      return;
    }
    
    done += ret;
  }
  while (done < count);
  
#ifdef FTDI_DEBUG
  TimeStamp tend;
  long elapsed;
  TimeStamp::timeDiff (tend, tstart, &elapsed);
  ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "FTDI write took " << elapsed << " µs");
#endif
}



void
FtdiDevice::
read (uint8_t* const buf,
      unsigned const count)
{
  if (! connected_) {
    return;
  }
  
  boost::shared_lock<boost::shared_mutex> _ (mutex_);
  
  TimeStamp tstart;
  
  unsigned done = 0;
  do {
    {
      TimeStamp tnow;
      long elapsed;
      TimeStamp::timeDiff (tnow, tstart, &elapsed);
      if (elapsed > timeout_) {
        ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " FTDI read timeout.");
        ftdiError();
        return;
      }
    }
    
    int ret = ctx_.read (buf + done, count - done);
    if (ret < 0) {
      ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " FTDI read error: " << ctx_.error_string());
      ftdiError();
      return;
    }
    
    done += ret;
  }
  while (done < count);
  
#ifdef FTDI_DEBUG
  TimeStamp tend;
  long elapsed;
  TimeStamp::timeDiff (tend, tstart, &elapsed);
  ROS_DEBUG_STREAM_NAMED ("FtdiDevice", "FTDI read took " << elapsed << " µs");
#endif
}



void
FtdiDevice::
ftdiError()
{
  if (! connected_) {
    return;
  }
  
  boost::shared_lock<boost::shared_mutex> _ (mutex_);
  
  /*  int ret = ctx_.close();
  if (ret < 0) {
    ROS_ERROR_STREAM("FtdiDevice", ros::this_node::getName() << " Close failed : " << ctx_.error_string());
  }*/
  
  ROS_ERROR_STREAM_NAMED ("FtdiDevice", ros::this_node::getName() << " FTDI device disconnected.");
  
  connected_ = false;
  ctx_ = Ftdi::Context();
}
