/**
 * @Filename SousaMotors.cpp
 * @Description ...to be described...
 * @Status unknown
 * @Version $Id: SousaMotors.cpp 6107 2013-06-27 13:22:05Z jreis $
 * @Maintainer Joao Reis (joaocgreis@gmail.com)
 */

#include "SousaMotors.hpp"

#include <fcntl.h>
#include <errno.h>
#include <termios.h>

#include "BinaryHelpers.hpp"
#include <socrob/math.h>
using namespace socrob;

#include <DataMap.hpp>
#include <String.hpp>
#include <Float.hpp>
#include <Boolean.hpp>
#include <Logger.hpp>
#include <WorldDefinition.hpp>

#include <MessageDataOdometry.hpp>



const double ODOZERO_X = 0.001;
const double ODOZERO_Y = 0.001;
const double ODOZERO_T = 0.001;
const ros::Duration ODOZERO_DURATION = ros::Duration (5.0);



const unsigned ENCODER_TICKS = 2000; //4x500
const double BASE_TIME = 6.5; //ms
const double INCREMENTAL_TIME = 0.436; //ms
const std::size_t ODOMETRY_HISTORY_LENGTH = 10; //minimum 1 value must be kept for acceleration control
const double GEARBOX_RELATION = 21.0;
const double ACCELERATION_RESET_THRESHOLD = 0.3;



SousaMotors::
SousaMotors () :
  FtdiDevice ("motors",
              100000,
              0),
  control_enabled_ (true), // If nothing is received from ShowStopper, it is probably not running and we should ignore it... or not?
  wishedControls_(),
  wheelRadius_ (WD::WheelRadius()),
  robotRadius_ (WD::RobotRadius()),
  maxVelPerWheel_ (sqrt (3) / 2.0 * WD::MaximumSpeed() / wheelRadius_),
  //hardware maximum (52.7 - by IDMind): maxAccelPerWheel_ ( ( (1 / (BASE_TIME / 1000)) * 2 * M_PI / (ENCODER_TICKS* GEARBOX_RELATION)) / (INCREMENTAL_TIME / 1000)),
  maxAccelPerWheel_ (WD::MaximumAccelerationPerWheel()),
  Ka_ (1),
  odometryHistory_(),
  controlReference_(),
  lastControls_(),
  timeReference_(),
  last_moving_odometry_ (ros::Time::now()),
  nh_(),
  all_ok_sub_ (nh_.subscribe ("watchdogs/all_ok", 1, &SousaMotors::allOkCallback, this))
{
  addSerial ("FTE48B12", 115200); // test
  addSerial ("FTF55HD5", 115200); // OMNI5
  addSerial ("FTF55H45", 115200); // OMNI4
  addSerial ("FTF55GXJ", 115200); // OMNI3
  addSerial ("FTF55IBB", 115200);
  addSerial ("FTF55HG5", 115200);
  addSerial ("FTF55GYP", 115200);
}



void
SousaMotors::
allOkCallback (std_msgs::Bool const& ok)
{
  if (ok.data) {
    ROS_ERROR_STREAM ("Motors: Reenabling motors.");
    control_enabled_ = true;
  }
  else {
    ROS_ERROR_STREAM ("Motors: DISABLING CONTROL OF MOTORS");
    control_enabled_ = false;
    sendSpeed (0, 0, 0);
    sendStopControl();
  }
}



void
SousaMotors::
update3()
{
  int16_t d1, d2, d3;
  
  readOdometry (d1, d2, d3);
  
  if (connected_) {
    broadcastOdometry (d1, d2, d3);
  }
  
  ROS_DEBUG_STREAM (
    "Received odometry binary: "
    << (int) d1 << " "
    << (int) d2 << " "
    << (int) d3);
    
  if (control_enabled_) {
    calculateWheelSpeed (d1, d2, d3);
    
    bool control_is_zero = (d1 == 0) && (d2 == 0) && (d3 == 0);
    bool odometry_is_zero = ros::Time::now() - last_moving_odometry_ > ODOZERO_DURATION;
    
    ROS_DEBUG_STREAM_THROTTLE (1, (control_is_zero ? (odometry_is_zero ? "Both control and odometry are 0" : "Control is 0") : (odometry_is_zero ? "Odometry is 0" : "Both control and odometry are not zero")));
    
    if (control_is_zero && odometry_is_zero) {
      sendStopControl();
    }
    else {
      sendSpeed (d1, d2, d3);
    }
  }
  else {
    sendStopControl();
  }
  
  bool buttonState = readButtonState();
  broadcastButtonState (buttonState && connected_);
}



void
SousaMotors::
setRobotAcceleration (float ax,
                      float ay,
                      float alpha,
                      CONTROL_PRIORITY prio)
{
  ROS_DEBUG_STREAM (
    "setRobotAcceleration(ax=" << ax
    << ", ay=" << ay
    << ", alpha=" << alpha
    << ")");
    
  wishedControls_.setLinearX (ax);
  wishedControls_.setLinearY (ay);
  wishedControls_.setAngular (alpha);
  wishedControls_.setPriority (prio);
  wishedControls_.setType (ACCELERATION);
}



void
SousaMotors::
setRobotVelocity (float vx,
                  float vy,
                  float w,
                  CONTROL_PRIORITY prio)
{
  ROS_DEBUG_STREAM (
    "setRobotVelocity(vx=" << vx
    << ", vy=" << vy
    << ", w=" << w
    << ")");
    
  wishedControls_.setLinearX (vx);
  wishedControls_.setLinearY (vy);
  wishedControls_.setAngular (w);
  wishedControls_.setPriority (prio);
  wishedControls_.setType (VELOCITY);
}



void
SousaMotors::
setRobotControls (ControlInput controls)
{
  ROS_DEBUG_STREAM (
    "setRobotControls(linearX=" << controls.getLinearX()
    << ", linearY=" << controls.getLinearY()
    << ", angular=" << controls.getAngular()
    << ")");
    
  wishedControls_ = controls;
}



WheelControlTriplet
SousaMotors::
saturateControls (WheelControlTriplet linear, WheelControlTriplet angular)
{
  WheelControlTriplet controls;
  
  controls = linear + angular;
  float k = 1;
  
  //Determine the fastest wheel
  float maxVel = abs (controls.right);
  if (abs (controls.rear) > maxVel) {
    maxVel = abs (controls.rear);
  }
  if (abs (controls.left) > maxVel) {
    maxVel = abs (controls.left);
  }
  
  if (maxVel > maxVelPerWheel_) {
    switch (wishedControls_.getPriority()) {
      case BALANCED: { //No priority
        k = maxVelPerWheel_ / maxVel;
        
        controls = controls * k;
        
        break;
      }
      case LINEAR: { //Linear velocity priority
        //Determine the fastest wheel in linear velocity
        float maxVel_v = abs (linear.right);
        if (abs (linear.rear) > maxVel_v) {
          maxVel_v = abs (linear.rear);
        }
        if (abs (linear.left) > maxVel_v) {
          maxVel_v = abs (linear.left);
        }
        if (maxVel_v >= this->maxVelPerWheel_) {
          k = maxVelPerWheel_ / maxVel_v;
          //notice that the angular velocity has no contribution in this case (w=0).
          controls = linear * k;
        }
        else {
          /*in the following expression, (maxVel-maxVel_v) relates to the available control effort
            for angular velocity. Suppose that you have a saturation limit maxVelPerWheel_ of 10rad/s
            in each wheel, and that you have the following wished velocities:
            rightW_v = 8  rightW_w = 3
            rearW_v = 6   rearW_w = 3
            leftW_v = 4   leftW_w = 3
            so, evidently,
            maxVel = 11   maxVel_v = rightW_v = 8
            Also, note that the contribution for the angular velocity is always the same for each
            wheel. In this situation, it is clear that by giving priority to linear velocity, the
            available control effort is 2, which is maxVelPerWheel_-maxVel_v. So we need to multiply
            all the w contributions by 2/3, which is (maxVelPerWheel_-maxVel_v)/(maxVel-maxVel_v)*/
          k = (maxVelPerWheel_ - maxVel_v) / (maxVel - maxVel_v);
          angular = angular * k;
          
          controls = linear + angular;
        }
        break;
      }
      case ANGULAR: { //Angular velocity priority
        if (angular.right >= this->maxVelPerWheel_) {
          k = maxVelPerWheel_ / angular.right;
          //notice that the linear velocity has no contribution in this case (v=0).
          controls = angular * k;
        }
        else {
          k = (maxVelPerWheel_ - angular.right) / (maxVel - angular.right);
          linear = linear * k;
          
          controls = linear + angular;
        }
        break;
      }
      default:
        break;
    }
  }
  
  
  float deltaTime = (1.0 / 30.0);
  WheelControlTriplet zeroControls;
  bool solution = false;
  
  if (controls == zeroControls && controls == lastControls_) {
    Ka_ = 0;
    return controls;
  }
  else if (controls == zeroControls && controls != lastControls_) {
    solution = minimizeK (lastControls_, deltaTime);
    lastControls_ = lastControls_ * Ka_;
    return lastControls_;
  }
  else {
    solution = preserveK (controls, deltaTime, lastControls_);
    if (!solution) {
      minimizeK (lastControls_, deltaTime);
      lastControls_ = lastControls_ * Ka_;
      return lastControls_;
    }
    lastControls_ = controls * Ka_;
    return controls * Ka_;
  }
}



void
SousaMotors::
velocityControl (WheelControlTriplet& controls)
{
  WheelControlTriplet linear;
  linear.left = (1.0 / wheelRadius_) * ( (-0.5 * sqrt (3)) * (wishedControls_.getLinearX()) + 0.5 * (wishedControls_.getLinearY()));
  linear.right = (1.0 / wheelRadius_) * ( (0.5 * sqrt (3)) * (wishedControls_.getLinearX()) + 0.5 * (wishedControls_.getLinearY()));
  linear.rear = (1.0 / wheelRadius_) * (-wishedControls_.getLinearY());
  
  //calculate wheel velocities with vx = 0 and vy = 0;
  WheelControlTriplet angular;
  angular.left = angular.right = angular.rear = (1.0 / wheelRadius_) * (robotRadius_ * wishedControls_.getAngular());
  
  controls = saturateControls (linear, angular);
}



void
SousaMotors::
accelerationControl (WheelControlTriplet& controls)
{
  //default failsafe value (should never be loaded)
  float dtEstimate = 1.0 / 30.0;
  
  if (odometryHistory_.size() > 0) {
    dtEstimate = 0;
    for (unsigned i = 0; i < odometryHistory_.size(); i++) {
      dtEstimate += odometryHistory_[i].dt;
    }
    dtEstimate = dtEstimate / odometryHistory_.size();
    
    if (odometryHistory_.back().dLeft == 0) {
      controlReference_.left = 0;
    }
    else if (abs (1 - controlReference_.left / odometryHistory_.back().dLeft / odometryHistory_.back().dt) < ACCELERATION_RESET_THRESHOLD) {
      controlReference_.left = odometryHistory_.back().dLeft / odometryHistory_.back().dt;
    }
    if (odometryHistory_.back().dRear == 0) {
      controlReference_.rear = 0;
    }
    else if (abs (1 - controlReference_.rear / odometryHistory_.back().dRear / odometryHistory_.back().dt) < ACCELERATION_RESET_THRESHOLD) {
      controlReference_.rear = odometryHistory_.back().dRear / odometryHistory_.back().dt;
    }
    if (odometryHistory_.back().dRight == 0) {
      controlReference_.right = 0;
    }
    else if (abs (1 - controlReference_.right / odometryHistory_.back().dRight / odometryHistory_.back().dt) < ACCELERATION_RESET_THRESHOLD) {
      controlReference_.right = odometryHistory_.back().dRight / odometryHistory_.back().dt;
    }
  }
  
  //now we have to project to robot velocities and back to wheel speeds again, to separate linear and angular components:
  float robotVx = (wheelRadius_ / sqrt ( (float) 3)) * (-controlReference_.left + controlReference_.right);
  float robotVy = (wheelRadius_ / (float) 3) * (-2 * controlReference_.rear + controlReference_.left + controlReference_.right);
  float robotW = (wheelRadius_ / ( (float) 3 * robotRadius_)) * (controlReference_.rear + controlReference_.left + controlReference_.right);
  
  WheelControlTriplet linear;
  linear.left = (1.0 / wheelRadius_) * ( (-0.5 * sqrt (3)) * (robotVx + wishedControls_.getLinearX() * dtEstimate) + 0.5 * (robotVy + wishedControls_.getLinearY() * dtEstimate));
  linear.right = (1.0 / wheelRadius_) * ( (0.5 * sqrt (3)) * (robotVx + wishedControls_.getLinearX() * dtEstimate) + 0.5 * (robotVy + wishedControls_.getLinearY() * dtEstimate));
  linear.rear = (1.0 / wheelRadius_) * (- (robotVy + wishedControls_.getLinearY() * dtEstimate));
  
  //calculate wheel velocities with vx = 0 and vy = 0;
  WheelControlTriplet angular;
  angular.left = angular.right = angular.rear = (1.0 / wheelRadius_) * (robotRadius_ * (robotW + wishedControls_.getAngular() * dtEstimate));
  
  //   WheelControlTriplet controlsVect = (WheelControlTriplet *) malloc(2 * sizeof(WheelControlTriplet));
  
  controls = controlReference_ = saturateControls (linear, angular);
}



void
SousaMotors::
calculateWheelSpeed (int16_t& d1,
                     int16_t& d2,
                     int16_t& d3)
{
  ROS_DEBUG_STREAM ("Crude wheel speeds: "
                    << odometryHistory_.back().dLeft / odometryHistory_.back().dt << " "
                    << odometryHistory_.back().dRear / odometryHistory_.back().dt << " "
                    << odometryHistory_.back().dRight / odometryHistory_.back().dt);
                    
  WheelControlTriplet wheelControls;
  if (wishedControls_.getType() == ACCELERATION) {
    accelerationControl (wheelControls);
  }
  else {
    velocityControl (wheelControls);
  }
  
  //convert rad/s to encoder pulse reference (int)
  d1 = round (wheelControls.rear * GEARBOX_RELATION * ENCODER_TICKS * BASE_TIME / (2 * M_PI * 1000));
  d2 = round (wheelControls.left * GEARBOX_RELATION * ENCODER_TICKS * BASE_TIME / (2 * M_PI * 1000));
  d3 = round (wheelControls.right * GEARBOX_RELATION * ENCODER_TICKS * BASE_TIME / (2 * M_PI * 1000));
}



void
SousaMotors::
broadcastOdometry (int16_t const& d1,
                   int16_t const& d2,
                   int16_t const& d3)
{
  TimeStamp odometryDataTime;
  float dPosxRobot, dPosyRobot, dThetaRobot;
  OdometryDataPoint currentData;
  
  //convert encoder pulses to rad
  currentData.dRear = d1 * 2 * M_PI / (ENCODER_TICKS * GEARBOX_RELATION);
  currentData.dLeft = d2 * 2 * M_PI / (ENCODER_TICKS * GEARBOX_RELATION); //wheel displacements, radians
  currentData.dRight = d3 * 2 * M_PI / (ENCODER_TICKS * GEARBOX_RELATION);
  
  double dt;
  TimeStamp::timeDiff (odometryDataTime, timeReference_, &dt);
  currentData.dt = dt;
  timeReference_.setTimeToNow();
  
  odometryHistory_.push_back (currentData);
  if (odometryHistory_.size() > ODOMETRY_HISTORY_LENGTH) {
    odometryHistory_.pop_front();
  }
  
  dPosxRobot = (wheelRadius_ / sqrt (3.0)) * (-currentData.dLeft + currentData.dRight);
  dPosyRobot = (wheelRadius_ / 3.0) * (-2 * currentData.dRear + currentData.dLeft + currentData.dRight);
  dThetaRobot = (wheelRadius_ / (3.0 * robotRadius_)) * (currentData.dRear + currentData.dLeft + currentData.dRight);
  
  if ( (abs (dPosxRobot) > ODOZERO_X)
       || (abs (dPosyRobot) > ODOZERO_Y)
       || (abs (dThetaRobot) > ODOZERO_T)) {
    last_moving_odometry_ = ros::Time::now();
  }
  
  ROS_DEBUG_STREAM ("Odometry: " << dPosxRobot << " " << dPosyRobot << " " << dThetaRobot);
  
  MessageDataOdometry::buildAndSend (dPosxRobot,
                                     dPosyRobot,
                                     dThetaRobot,
                                     odometryDataTime);
}



void
SousaMotors::
broadcastButtonState (bool buttonState)
{
  static bool firstTime = true;
  static bool prevState;
  
  if (! control_enabled_) {
    buttonState = false;
  }
  
  if (firstTime) {
    firstTime = false;
    prevState = buttonState;
    putDataWI ("isRunning", shared_ptr<Boolean> (new Boolean (buttonState)));
    return;
  }
  
  if (prevState != buttonState) {
    prevState = buttonState;
    putDataWI ("isRunning", shared_ptr<Boolean> (new Boolean (buttonState)));
  }
}



void
SousaMotors::
sendSpeed (int16_t const& d1,
           int16_t const& d2,
           int16_t const& d3)
{
  ROS_DEBUG_STREAM ("Sending speed");
  
  uint8_t wb[7];
  wb[0] = 'V';
  uint_split (wb[1], wb[2], static_cast<uint16_t> (d1));
  uint_split (wb[3], wb[4], static_cast<uint16_t> (d2));
  uint_split (wb[5], wb[6], static_cast<uint16_t> (d3));
  
  write (wb, 7);
  
  uint8_t rb[4];
  read (rb, 4);
  
  ROS_DEBUG_STREAM (
    "Speed answer bytes: "
    << (int) rb[0] << " "
    << (int) rb[1] << " "
    << (int) rb[2] << " "
    << (int) rb[3]);
    
  if (! connected_) {
    return;
  }
  
  if (! checkChecksum (rb, 4)) {
    ROS_ERROR_STREAM ("Checksum error");
    ftdiError();
    return;
  }
}



void
SousaMotors::
sendStopControl()
{
  ROS_DEBUG_STREAM ("Sending stop control");
  
  uint8_t wb = 'S';
  write (&wb, 1);
  
  uint8_t rb[4];
  read (rb, 4);
  
  ROS_DEBUG_STREAM (
    "Stop control answer bytes: "
    << (int) rb[0] << " "
    << (int) rb[1] << " "
    << (int) rb[2] << " "
    << (int) rb[3]);
    
  if (! connected_) {
    return;
  }
  
  if (! checkChecksum (rb, 4)) {
    ROS_ERROR_STREAM ("Checksum error");
    ftdiError();
    return;
  }
}



void
SousaMotors::
readOdometry (int16_t& d1,
              int16_t& d2,
              int16_t& d3)
{
  ROS_DEBUG_STREAM ("Reading odometry");
  
  uint8_t wb = 'J';
  write (&wb, 1);
  
  uint8_t rb[10];
  read (rb, 10);
  
  ROS_DEBUG_STREAM (
    "Odometry bytes: "
    << (int) rb[0] << " "
    << (int) rb[1] << " "
    << (int) rb[2] << " "
    << (int) rb[3] << " "
    << (int) rb[4] << " "
    << (int) rb[5] << " "
    << (int) rb[6] << " "
    << (int) rb[7] << " "
    << (int) rb[8] << " "
    << (int) rb[9]);
    
  // Make sure that odometry is = 0 if anything went wrong.
  if (! connected_) {
    d1 = d2 = d3 = 0;
    return;
  }
  
  if (! checkChecksum (rb, 10)) {
    ROS_ERROR_STREAM ("Checksum error");
    ftdiError();
    d1 = d2 = d3 = 0;
    return;
  }
  
  uint16_t tmp;
  uint_cat (tmp, rb[1], rb[2]);
  d1 = static_cast<int16_t> (tmp);
  uint_cat (tmp, rb[3], rb[4]);
  d2 = static_cast<int16_t> (tmp);
  uint_cat (tmp, rb[5], rb[6]);
  d3 = static_cast<int16_t> (tmp);
}



bool
SousaMotors::
readButtonState()
{
  ROS_DEBUG_STREAM ("Reading button state");
  
  uint8_t wb = 'O';
  write (&wb, 1);
  
  uint8_t rb[5];
  read (rb, 5);
  
  ROS_DEBUG_STREAM (
    "Button bytes: "
    << (int) rb[0] << " "
    << (int) rb[1] << " "
    << (int) rb[2] << " "
    << (int) rb[3] << " "
    << (int) rb[4]);
    
  if (! connected_) {
    return false;
  }
  
  if (! checkChecksum (rb, 5)) {
    ROS_ERROR_STREAM ("Checksum error");
    ftdiError();
    return false;
  }
  
  bool buttonState = rb[1];
  
  if (!buttonState) { //resetting acceleration control
    Ka_ = 1.0;
    lastControls_ = WheelControlTriplet(); //zero
  }
  
  return buttonState;
}



bool
SousaMotors::
checkChecksum (uint8_t const* const buf,
               unsigned const count)
{
  uint16_t received;
  uint16_t calculated = 0;
  
  uint_cat (received, buf[count - 2], buf[count - 1]);
  
  for (unsigned i = 0; i < (count - 2); i++) {
    calculated += buf[i];
  }
  
  return calculated == received;
}



bool
SousaMotors::
minimizeK (WheelControlTriplet lastWheelVel, float deltaTime)
{
  float max, min;
  bool solution = false;
  solution = lp_Solver (lastWheelVel, deltaTime, lastWheelVel, max, min);
  if (!solution) {
    return false;
  }
  Ka_ = min;
  return true;
}



bool
SousaMotors::
preserveK (WheelControlTriplet desiredWheelVel, float deltaTime, WheelControlTriplet lastWheelVel)
{
  float max, min;
  bool solution = false;
  solution = lp_Solver (desiredWheelVel, deltaTime, lastWheelVel, max, min);
  if (!solution) {
    return false;
  }
  
  if (min >= 1) {
    Ka_ = min;
  }
  else if (max <= 1) {
    Ka_ = max;
  }
  else {
    Ka_ = 1;
  }
  return true;
}



bool
SousaMotors::
lp_Solver (WheelControlTriplet desiredWheelVel, float deltaTime, WheelControlTriplet lastWheelVel, float& max, float& min)
{
  float* a = NULL, *b = NULL, *ans = NULL;
  bool* c = NULL;
  int j = 0;
  
  a = (float*) malloc (6 * sizeof (float));
  b = (float*) malloc (6 * sizeof (float));
  c = (bool*) calloc (6, sizeof (bool));
  ans = (float*) malloc (6 * sizeof (float));
  
  a[0] = desiredWheelVel.left;
  a[1] = desiredWheelVel.right;
  a[2] = desiredWheelVel.rear;
  a[3] = -desiredWheelVel.left;
  a[4] = -desiredWheelVel.right;
  a[5] = -desiredWheelVel.rear;
  
  b[0] = maxAccelPerWheel_ * deltaTime + lastWheelVel.left;
  b[1] = maxAccelPerWheel_ * deltaTime + lastWheelVel.right;
  b[2] = maxAccelPerWheel_ * deltaTime + lastWheelVel.rear;
  b[3] = maxAccelPerWheel_ * deltaTime - lastWheelVel.left;
  b[4] = maxAccelPerWheel_ * deltaTime - lastWheelVel.right;
  b[5] = maxAccelPerWheel_ * deltaTime - lastWheelVel.rear;
  
  for (j = 0; j < 3; j++) {
    if (a[j] < 0) {
      c[j] = 1; //c is 1 for a greater or equal than (>=) constraint, and 0 otherwise.
      c[j + 3] = 0;
    }
    else {
      c[j] = 0;
      c[j + 3] = 1;
    }
  }
  
  for (j = 0; j < 6; j++) {
    if (c[j] == 0) {
      if (a[j] == 0) {
        ans[j] = INFTY;
      }
      else {
        ans[j] = b[j] / a[j];
      }
    }
    else {
      if (a[j] == 0) {
        ans[j] = -INFTY;
      }
      else {
        ans[j] = b[j] / a[j];
      }
    }
  }
  
  max = INFTY;
  min = -INFTY;
  for (j = 0; j < 6; j++) {
    if (c[j] == 0) {
      if (ans[j] < max) {
        max = ans[j];
      }
    }
    else {
      if (ans[j] > min) {
        min = ans[j];
      }
    }
  }
  
  if (min < 0) {
    min = 0;
  }
  
  if (max > min) {
    return true;
  }
  else {
    return false;
  }
}




#define ROS_NAME "motors"
#define ROS_CLASS SousaMotors
#include <RosMain.cpp.part>
