/**
 * @Filename SousaMotors.hpp
 * @Description ...to be described...
 * @Status unknown
 * @Version $Id: SousaMotors.hpp 6089 2013-06-26 08:18:47Z jreis $
 * @Maintainer Joao Reis (joaocgreis@gmail.com)
 */

#ifndef _SOUSA_MOTORS_HPP_
#define _SOUSA_MOTORS_HPP_

#include <ros/ros.h>
#include <std_msgs/Bool.h>



class SousaMotors;

#include <stdint.h>
#include <deque>
#include "FtdiDevice.hpp"
#include <ControlInput.hpp>
#include <Logger.hpp>

#include <MessageSetRobotFinalVelocity.hpp>
#include <MessageSetRobotFinalAcceleration.hpp>



using namespace mermaid::support::logging;



struct OdometryDataPoint {
  float dLeft, dRear, dRight, dt;
  
  OdometryDataPoint() : dLeft (0), dRear (0), dRight (0), dt (0) {};
};



struct WheelControlTriplet {
  float left, rear, right;
  
  WheelControlTriplet (float leftC = 0, float rearC = 0, float rightC = 0) : left (leftC), rear (rearC), right (rightC) {};
  WheelControlTriplet operator+ (const WheelControlTriplet& controls) {
    return WheelControlTriplet (left + controls.left, rear + controls.rear, right + controls.right);
  };
  WheelControlTriplet operator* (const float k) {
    return WheelControlTriplet (left * k, rear * k, right * k);
  };
  bool operator== (const WheelControlTriplet& controls) {
    return controls.left == left && controls.rear == rear && controls.right == right;
  };
  bool operator!= (const WheelControlTriplet& controls) {
    return controls.left != left || controls.rear != rear || controls.right != right;
  };
};



/**
 * @Class SousaMotors SousaMotors.hpp "SousaMotors.hpp"
 * ?
 * @Author Joao Reis
 */
class SousaMotors :
  public FtdiDevice,
  public Receiver<MessageSetRobotFinalVelocity>,
  public Receiver<MessageSetRobotFinalAcceleration>
{
  public:
    SousaMotors ();
    
    void
    setRobotVelocity (float vx,
                      float vy,
                      float w,
                      CONTROL_PRIORITY prio);
                      
    void
    setRobotAcceleration (float ax,
                          float ay,
                          float alpha,
                          CONTROL_PRIORITY prio);
                          
    void
    setRobotControls (ControlInput controls);
    
  protected:
    void update3();
    
  private:
    void
    allOkCallback (std_msgs::Bool const& ok);
    
    WheelControlTriplet
    saturateControls (WheelControlTriplet linear,
                      WheelControlTriplet angular);
                      
    void
    velocityControl (WheelControlTriplet& controls);
    
    void
    accelerationControl (WheelControlTriplet& controls);
    
    
    void
    calculateWheelSpeed (int16_t& d1,
                         int16_t& d2,
                         int16_t& d3);
                         
    void
    broadcastOdometry (int16_t const& d1,
                       int16_t const& d2,
                       int16_t const& d3);
                       
    void broadcastButtonState (bool buttonState);
    
    
    void
    sendSpeed (int16_t const& d1,
               int16_t const& d2,
               int16_t const& d3);
               
    void
    sendStopControl();
    
    void
    readOdometry (int16_t& d1,
                  int16_t& d2,
                  int16_t& d3);
                  
    bool
    readButtonState();
    
    
    bool
    checkChecksum (uint8_t const* const buf,
                   unsigned const count);
                   
    bool
    minimizeK (WheelControlTriplet lastWheelVel, float deltaTime);
    
    bool
    preserveK (WheelControlTriplet desiredWheelVel, float deltaTime, WheelControlTriplet lastWheelVel);
    
    bool
    lp_Solver (WheelControlTriplet desiredWheelVel, float deltaTime, WheelControlTriplet lastWheelVel, float& max, float& min);
    
    
    bool control_enabled_;
    ControlInput wishedControls_;
    
    // Keeping this private so we don't have to access WI everytime (which is slow)
    float wheelRadius_, robotRadius_;
    float maxVelPerWheel_, maxAccelPerWheel_;
    float Ka_;
    
    // These are necessary to make parabolic fits of the odometry data to infer acceleration
    /// TODO Consider using a std::list or a boost::circular_buffer
    deque <OdometryDataPoint> odometryHistory_;
    
    // We need all these for acceleration control (which works at a lower frequency)
    WheelControlTriplet controlReference_;
    WheelControlTriplet lastControls_;
    TimeStamp timeReference_;
    
    ros::Time last_moving_odometry_;
    
    ros::NodeHandle nh_;
    ros::Subscriber all_ok_sub_;
};

#endif
