#include "driver_view.h"

#include <cv_bridge/cv_bridge.h>

#include <WorldDefinition.hpp>

#include <socrob/math.h>
using namespace socrob;



DriverView::
DriverView () :
  nh_(),
  it_ (nh_),
  display_ ("driver_view"),
  image_sub_ (),
  activation_timer_ (nh_.createTimer (ros::Duration (1.0), &DriverView::check_active, this)),
  ROTATION_ANGLE_ (atan2 (WD::RobotCenterImgRow() - WD::CameraCenterImgRow(), WD::CameraCenterImgCol() - WD::RobotCenterImgCol()))
{
  update_params();
}



void
DriverView::
imageCallback (const sensor_msgs::ImageConstPtr& image)
{
  cv_bridge::CvImageConstPtr cv_image = cv_bridge::toCvShare (image, "bgr8");
  
  const int clipped_view_height = height_ * (front_radial_top_ - front_radial_bottom_);
  const int clipped_rear_height = rear_height_ * (rear_radial_top_ - rear_radial_bottom_);
  
  cv::Mat view (clipped_view_height + clipped_rear_height, width_, CV_8UC3);
  
  const cv::Point2d IMAGE_CENTER ( (image->width - 1) / 2, (image->height - 1) / 2);
  const double IMAGE_RAY = min (IMAGE_CENTER.x, IMAGE_CENTER.y);
  
  const int last_view_col = width_ - 1;
  const int last_view_row = clipped_view_height - 1;
  const double RAY_BASE = IMAGE_RAY * front_radial_bottom_;
  const double VIEW_RAY = IMAGE_RAY * front_radial_top_ - RAY_BASE;
  
  for (int col = 0; col <= last_view_col; ++col) {
    const double angle = front_angular_left_ - (col * front_angular_span_ / last_view_col);
    const cv::Point2d ray_norm (cos (angle), -sin (angle));
    
    for (int row = 0; row <= last_view_row; ++row) {
      const double ray = RAY_BASE + VIEW_RAY * (last_view_row - row) / last_view_row;
      const cv::Point2d pixel = IMAGE_CENTER + ray * ray_norm;
      
      view.at<cv::Vec3b> (clipped_rear_height + row, col) = cv_image->image.at<cv::Vec3b> (pixel);
    }
  }
  
  const int last_rear_col = rear_width_ - 1;
  const int last_rear_row = clipped_rear_height - 1;
  const double REAR_RAY_BASE = IMAGE_RAY * rear_radial_bottom_;
  const double REAR_RAY = IMAGE_RAY * rear_radial_top_ - REAR_RAY_BASE;
  
  for (int col = 0; col <= last_rear_col; ++col) {
    const double angle = rear_angular_left_ + (col * rear_angular_span_ / last_rear_col);
    const cv::Point2d ray_norm (cos (angle), -sin (angle));
    
    for (int row = 0; row <= last_rear_row; ++row) {
      const double ray = REAR_RAY_BASE + REAR_RAY * (last_rear_row - row) / last_rear_row;
      const cv::Point2d pixel = IMAGE_CENTER + ray * ray_norm;
      
      view.at<cv::Vec3b> (row, rear_start_col_ + col) = cv_image->image.at<cv::Vec3b> (pixel);
    }
  }
  
  display_.set (view);
  
  update_params();
}



void
DriverView::
check_active (ros::TimerEvent const&)
{
  if (display_.active()) {
    if (! image_sub_) {
      image_sub_ = it_.subscribe ("devices/topcam/image_raw", 1, &DriverView::imageCallback, this);
    }
  }
  else {
    if (image_sub_) {
      image_sub_ = image_transport::Subscriber();
    }
  }
}



#define PARAM_DIRECT(param_name, param_var, param_default) \
  do { \
    if (! ros::param::getCached (param_name, param_var)) { \
      param_var = (param_default); \
      ros::param::set (param_name, param_var); \
    } \
  } while (false)



void
DriverView::
update_params ()
{
  PARAM_DIRECT ("~width", width_, 1000);
  PARAM_DIRECT ("~height", height_, 500);
  
  PARAM_DIRECT ("~front_span", front_angular_span_, M_PI);
  front_angular_left_ = clipAngle (ROTATION_ANGLE_ + (front_angular_span_ / 2));
  PARAM_DIRECT ("~front_bottom", front_radial_bottom_, 0.12);
  PARAM_DIRECT ("~front_top", front_radial_top_, 0.97);
  
  PARAM_DIRECT ("~rear_mul", rear_mul_, 0.5);
  rear_width_ = width_ * rear_mul_;
  rear_height_ = height_ * rear_mul_;
  rear_start_col_ = (width_ - rear_width_) / 2;
  
  PARAM_DIRECT ("~rear_span", rear_angular_span_, M_PI);
  rear_angular_left_ = clipAngle (ROTATION_ANGLE_ + (rear_angular_span_ / 2));
  PARAM_DIRECT ("~rear_bottom", rear_radial_bottom_, 0.35);
  PARAM_DIRECT ("~rear_top", rear_radial_top_, front_radial_top_);
}
