/**
 * @Filename LinePoints.cpp
 * LinePoints class implementation.
 * @Status Implementing
 * @Version $Id: line_points.cpp 6053 2013-05-09 15:55:57Z jreis $
 * @Maintainer João Estilita (jestilita@isr.ist.utl.pt)
 */

#include "line_points.h"
#include "sensors_consts.hpp"

#include <queue>

#include <cv_bridge/cv_bridge.h>
#include <std_srvs/Empty.h>

#include <opencv2/core/core_c.h> // For CV_RGB only

#include <MessageDataAllLinePoints.hpp>

#include <WorldDefinition.hpp>

#include <socrob/math.h>
using namespace socrob;



#define SHOW_ALGORITHM_TIME 0



namespace
{
  uint8_t
  jumpPoint (uint8_t p)
  {
    if (p == 255) {
      return 255;
    }
    return 128 + 0.5 * p;
  }
}



LinePoints::
LinePoints () :
  W (IMAGE_WIDTH),
  H (IMAGE_HEIGHT),
  searchPoints_ (W, H),
  cameraTransform_(),
  nh_(),
  it_ (nh_),
  debug_ ("debug"),
  debug_mask_ ("mask_debug"),
  image_sub_ (it_.subscribe ("devices/topcam/image_raw", 1, &LinePoints::imageCallback, this))
{
}



bool
LinePoints::
initialize ()
{
  setServiceDefaultActive (false, 0);
  
  highestBlackValue_ = constants::LinePoints::highest_black;
  
  return true;
}



void
LinePoints::
imageCallback (const sensor_msgs::ImageConstPtr& image)
{
  if ( (image->width != IMAGE_WIDTH) || (image->height != IMAGE_HEIGHT)) {
    ROS_FATAL_STREAM ("Received image with a wrong size: " << image->width << "x" << image->height << " (expecting " << IMAGE_WIDTH << "x" << IMAGE_HEIGHT << ")");
    abort();
  }
  
  TimeStamp start_time;
  
#if OPTIMIZE_LINE_POINTS
  cv_bridge::CvImage::ConstPtr cv_ptr = cv_bridge::toCvShare (image);
  if ( (cv_ptr->encoding != "bgr8") && (cv_ptr->encoding != "rgb8")) {
    ROS_FATAL_STREAM ("line_points prepared to receive rgb8 or bgr8 images but received " << cv_ptr->encoding);
    abort();
  }
#else
  cv_bridge::CvImageConstPtr cv_ptr = cv_bridge::toCvShare (image, "mono8");
#endif
  
  vector<pair<cv::Point, cv::Point> > transitions;
  vector<cv::Point> obstacles;
  
  detectTransitions (transitions, obstacles, cv_ptr->image);
  
  MessageDataAllLinePoints::buildAndSend (toRobotFrame (transitions), toRobotFrame (obstacles));
  
  
#if SHOW_ALGORITHM_TIME
  {
    TimeStamp end_time;
    long this_time = end_time.timeDiff (start_time);
    static long minmax = this_time;
    static double movavg = this_time;
    minmax = min (this_time, minmax);
    movavg = (movavg * 0.99) + (this_time * 0.01);
    cout << "Minimum ever: " << minmax << " µs. Moving average: " << static_cast<long> (movavg) << " µs. This time: " << this_time << " µs." << endl;
  }
#endif
  
  if (debug_mask_.active()) {
#if OPTIMIZE_LINE_POINTS
    cv::Mat image = cv_ptr->image.clone();
    for (int row = 0; row < image.rows; row++) {
      for (int col = 0; col < image.cols; col++) {
        cv::Point point (col, row);
        cv::Vec3b& pixelVec = image.at<cv::Vec3b> (point);
        uint8_t pixelValue = (pixelVec[0] >> 2) + (pixelVec[1] >> 1) + (pixelVec[2] >> 2);
        pixelVec[0] = pixelVec[1] = pixelVec[2] = pixelValue;
      }
    }
#else
    cv::Mat image;
    cv::cvtColor (cv_ptr->image, image, CV_GRAY2BGR);
#endif
    
    const cv::Point ORIGIN (WD::CameraCenterImgCol(), WD::CameraCenterImgRow());
    const cv::Point ROBOT_CENTER (WD::RobotCenterImgCol(), WD::RobotCenterImgRow());
    unsigned radiusOfSearch = WD::SearchableRadius() + constants::LinePoints::radius_increment;
    
    const double SUPPORT_1_MIN = WD::Support1Min();
    const double SUPPORT_1_MAX = WD::Support1Max();
    const double SUPPORT_2_MIN = WD::Support2Min();
    const double SUPPORT_2_MAX = WD::Support2Max();
    const double SUPPORT_3_MIN = WD::Support3Min();
    const double SUPPORT_3_MAX = WD::Support3Max();
    
    for (int row = 0; row < image.rows; row++) {
      for (int col = 0; col < image.cols; col++) {
        cv::Point point (col, row);
        
        double angle = atan2 (- (point.y - ORIGIN.y), point.x - ORIGIN.x);
        bool in_support = isAngleBetween (angle, SUPPORT_1_MIN, SUPPORT_1_MAX)
                          || isAngleBetween (angle, SUPPORT_2_MIN, SUPPORT_2_MAX)
                          || isAngleBetween (angle, SUPPORT_3_MIN, SUPPORT_3_MAX);
        bool out_of_support = ! in_support;
        
        bool in_robot = (sqr (ROBOT_CENTER.x - point.x)
                         + sqr (ROBOT_CENTER.y - point.y))
                        <= sqr (constants::LinePoints::robot_radius);
        bool out_of_robot = ! in_robot;
        
        bool in_radius = hypot (ORIGIN.x - point.x, ORIGIN.y - point.y) < radiusOfSearch;
        
        if (! (out_of_support && out_of_robot && in_radius)) {
          image.at<cv::Vec3b> (point) [1] = 255;
        }
      }
    }
    
    foreach (vector<cv::Point> const & ray, searchPoints_.getPoints()) {
      foreach (cv::Point const & point, ray) {
        image.at<cv::Vec3b> (point) = cv::Vec3b (0, 0, 255);
      }
    }
    
    debug_mask_.set (image);
  }
  
  if (debug_.active()) {
#if OPTIMIZE_LINE_POINTS
    cv::Mat image = cv_ptr->image.clone();
#else
    cv::Mat image;
    cv::cvtColor (cv_ptr->image, image, CV_GRAY2BGR);
#endif
    
    // Red lines
    // Attention: The yellow point of transition is calculated in the robot
    // frame, the red lines are drawn directly in the image (therefore the
    // yellow point may be slightly off the line, but correct).
    typedef pair<cv::Point, cv::Point> pair_point_point;
    foreach (pair_point_point const & transition, transitions) {
      cv::line (image, transition.first, transition.second, CV_RGB (255, 0, 0));
    }
    
    foreach (cv::Point const & point, obstacles) {
      cv::circle (image, point, 3, CV_RGB (0, 0, 0), -1);
      cv::circle (image, point, 1, CV_RGB (0, 0, 255), -1);
    }
    
    foreach (cv::Point const & point, toImageFrame (toRobotFrame (transitions))) {
      cv::circle (image, point, 3, CV_RGB (0, 0, 0), -1);
      cv::circle (image, point, 1, CV_RGB (255, 255, 0), -1);
    }
    
    debug_.set (image);
  }
}



void
LinePoints::
detectTransitions (vector<pair<cv::Point, cv::Point> >& transitions,
                   vector<cv::Point>& obstacles,
                   cv::Mat const& img)
{
  const size_t SEARCH_DIFF = 2;
  
  transitions.clear();
  obstacles.clear();
  
  bool found_some_non_black_in_VLPNC = false;
  
  typedef enum {LIN, LOUT} lstate_t;
  
  queue<int> searchRays;
  lstate_t lstate[constants::LinePoints::num_radial_rays_to_search];
  unsigned lines_in_ray[constants::LinePoints::num_radial_rays_to_search];
  size_t point_it[constants::LinePoints::num_radial_rays_to_search];
  
  for (size_t i = 0; i < constants::LinePoints::num_radial_rays_to_search; i++) {
    searchRays.push (i);
    lstate[i] = LOUT;
    lines_in_ray[i] = 0;
    point_it[i] = SEARCH_DIFF;
  }
  
  while (! searchRays.empty()) {
    int ray_num = searchRays.front();
    searchRays.pop();
    
    vector<cv::Point> const& ray = searchPoints_.getPoints() [ray_num];
    
    cv::Point entry (0, 0);
    uint8_t exit_value = 255;
    
    for (; point_it[ray_num] < ray.size(); point_it[ray_num]++) {
      bool continue_next = false;
      cv::Point pixelPosM = ray[point_it[ray_num] - SEARCH_DIFF];
      cv::Point pixelPos = ray[point_it[ray_num]];
#if OPTIMIZE_LINE_POINTS
      cv::Vec3b pixelVecM = img.at<cv::Vec3b> (pixelPosM);
      uint8_t pixelValueM = (pixelVecM[0] >> 2) + (pixelVecM[1] >> 1) + (pixelVecM[2] >> 2);
      cv::Vec3b pixelVec = img.at<cv::Vec3b> (pixelPos);
      uint8_t pixelValue = (pixelVec[0] >> 2) + (pixelVec[1] >> 1) + (pixelVec[2] >> 2);
#else
      uint8_t pixelValueM = img.at<uint8_t> (pixelPosM);
      uint8_t pixelValue = img.at<uint8_t> (pixelPos);
#endif
      
      // Detect black
      if (pixelValue <= highestBlackValue_) {
        obstacles.push_back (pixelPos);
        if (point_it[ray_num] != SEARCH_DIFF) {
          found_some_non_black_in_VLPNC = true;
        }
        // do not insert in queue again
        break;
      }
      
      // Detect lines
      switch (lstate[ray_num]) {
        case LIN:
          if (pixelValue <= exit_value) {
            lstate[ray_num] = LOUT;
            transitions.push_back (make_pair (entry, ray[point_it[ray_num] - 1]));
            lines_in_ray[ray_num]++;
            continue_next = true;
          }
          break;
        case LOUT:
          if ( (pixelValueM != 255) && (pixelValue >= jumpPoint (pixelValueM))) {
            exit_value = (pixelValue * 0.75) + (pixelValueM * 0.25);
            lstate[ray_num] = LIN;
            entry = pixelPos;
          }
          break;
      }
      
      
      if (lines_in_ray[ray_num] >= constants::LinePoints::max_lines_on_a_ray) {
        break;
      }
      
      if (continue_next) {
        searchRays.push (ray_num);
        break;
      }
    }
  }
  
  {
    static unsigned numtimesnotfount = 0;
    if (found_some_non_black_in_VLPNC) {
      numtimesnotfount = 0;
    }
    else {
      numtimesnotfount++;
    }
    if (numtimesnotfount == constants::LinePoints::num_black_frames_before_crash_hack) {
      std_srvs::Empty srv;
      if (ros::service::call ("localization/reset", srv)) {
        ROS_ERROR_STREAM ("line_points sent reset request to localization");
      }
      else {
        ROS_ERROR_STREAM ("Could not call service to reset localization!");
      }
      system ("killall -9 cortex");
      ROS_ERROR_STREAM ("line_points killed cortex");
    }
  }
}



/// DEPERCATED
shared_ptr<set<pair<float, float> > const>
LinePoints::
toRobotFrame_set (vector<pair<cv::Point, cv::Point> > const& transitions)
{
  shared_ptr<set<pair<float, float> > > transitionList = shared_ptr<set<pair<float, float> > > (new set<pair<float, float> >());
  
  shared_ptr<vector<pair<float, float> > const> tv = toRobotFrame (transitions);
  
  transitionList->insert (tv->begin(), tv->end());
  
  return transitionList;
}



shared_ptr<vector<pair<float, float> > const>
LinePoints::
toRobotFrame (vector<cv::Point> const& obstaclePoints)
{
  shared_ptr<vector<pair<float, float> > > obstaclePointList = shared_ptr<vector<pair<float, float> > > (new vector<pair<float, float> >());
  
  foreach (cv::Point const & point, obstaclePoints) {
    pair<float, float> const& obstacleRobotFrame = cameraTransform_.get (point.x, point.y);
    obstaclePointList->push_back (obstacleRobotFrame);
  }
  
  return obstaclePointList;
}



shared_ptr<vector<pair<float, float> > const>
LinePoints::
toRobotFrame (vector<pair<cv::Point, cv::Point> > const& transitions)
{
  shared_ptr<vector<pair<float, float> > > transitionList = shared_ptr<vector<pair<float, float> > > (new vector<pair<float, float> >());
  
  typedef pair<cv::Point, cv::Point> pair_point_point;
  foreach (pair_point_point const & transition, transitions) {
    pair<float, float> const& transition1 = cameraTransform_.get (transition.first.x, transition.first.y);
    pair<float, float> const& transition2 = cameraTransform_.get (transition.second.x, transition.second.y);
    
    float length = sqrt (sqr (transition1.first - transition2.first) + sqr (transition1.second - transition2.second));
    if (length > constants::LinePoints::maximum_transition_length) {
      continue;
    }
    
    pair<float, float> middle ( (transition1.first + transition2.first) / 2, (transition1.second + transition2.second) / 2);
    
    transitionList->push_back (make_pair (middle.first, middle.second));
  }
  
  return transitionList;
}



vector<cv::Point>
LinePoints::
toImageFrame (shared_ptr<vector<pair<float, float> > const> transitions)
{
  vector<cv::Point> points;
  
  typedef pair<float, float> pair_float_float;
  foreach (pair_float_float const & point, *transitions) {
    points.push_back (cameraTransform_.reverse (point.first, point.second));
  }
  
  return points;
}
