/**
 * @Filename FisheyeImageFrameRobotFrameTransformations.cpp
 * @Description Transform image frame coordinates to robot frame coordinates
 * and vice-versa.
 * @Status Implementing
 * @Version $Id: FisheyeImageFrameRobotFrameTransformations.cpp 5986 2013-04-24 10:53:37Z jreis $
 * @Maintainer Hugo Costelha
 */

#include "FisheyeImageFrameRobotFrameTransformations.hpp"

#include <socrob/math.h>
using namespace socrob;

#include <WorldDefinition.hpp>



void
FisheyeImageFrameRobotFrameTransformations::
recalcParams()
{
  // Assume for now that the camera is aligned with the robot
  xCamera_ = 0;
  yCamera_ = 0;
  
  // Get the robot position in field coordinates
  double tmpXCamera;
  double tmpYCamera;
  fromImageFrame2RobotFrame (robotCenterImgCol_,
                             robotCenterImgRow_,
                             tmpXCamera,
                             tmpYCamera);
                             
  // The camera position is the opposite
  xCamera_ = -tmpXCamera;
  yCamera_ = -tmpYCamera;
  
  // Calculate the angle
  rotationAngle_ = atan2 (robotCenterImgRow_ - cameraCenterImgRow_,
                          cameraCenterImgCol_ - robotCenterImgCol_);
}



FisheyeImageFrameRobotFrameTransformations::
FisheyeImageFrameRobotFrameTransformations () :
  cameraCenterImgCol_ (WD::CameraCenterImgCol()),
  cameraCenterImgRow_ (WD::CameraCenterImgRow()),
  robotCenterImgCol_ (WD::RobotCenterImgCol()),
  robotCenterImgRow_ (WD::RobotCenterImgRow()),
  fovConstant_ (WD::FOVConstant()),
  focalPointHeight_ (WD::CameraHeight())
{
  recalcParams();
  ROS_INFO_STREAM ("Using rotation angle: " << rotationAngle_);
}



void
FisheyeImageFrameRobotFrameTransformations::
cameraCenterImg (cv::Point2d const& cameraCenterImg)
{
  cameraCenterImgCol_ = cameraCenterImg.x;
  cameraCenterImgRow_ = cameraCenterImg.y;
  recalcParams();
}



void
FisheyeImageFrameRobotFrameTransformations::
fromImageFrame2RobotFrame (double imgX,
                           double imgY,
                           double& robotX,
                           double& robotY)
{
  // Origin in camera center and invert Y
  imgX = imgX - cameraCenterImgCol_;
  imgY = cameraCenterImgRow_ - imgY;
  
  double distanceImg = hypot (imgX, imgY);
  double distanceRobot = fromImageFrame2RobotFrameRaw (distanceImg);
  
  double angle = atan2 (imgY, imgX) - rotationAngle_;
  
  robotX = distanceRobot * cos (angle) + xCamera_;
  robotY = distanceRobot * sin (angle) + yCamera_;
}



void
FisheyeImageFrameRobotFrameTransformations::
fromRobotFrame2ImageFrame (double robotX,
                           double robotY,
                           double& imgX,
                           double& imgY)
{
  robotX -= xCamera_;
  robotY -= yCamera_;
  
  double robotDistance = hypot (robotX, robotY);
  double imgDistance = fromRobotFrame2ImageFrameRaw (robotDistance);
  
  double angle = atan2 (robotY, robotX) + rotationAngle_;
  
  // Origin in the top left corner of img and invert Y
  imgX = cameraCenterImgCol_ + (imgDistance * cos (angle));
  imgY = cameraCenterImgRow_ - (imgDistance * sin (angle));
}



double
FisheyeImageFrameRobotFrameTransformations::
fromImageFrame2RobotFrameRaw (double imgDistance)
{
  // return focalPointHeight_ * tan (imgDistance / fovConstant_);
  const double theta = atan2 (imgDistance, (2 * fovConstant_)) * 2;
  return focalPointHeight_ * tan (theta);
}



double
FisheyeImageFrameRobotFrameTransformations::
fromRobotFrame2ImageFrameRaw (double robotDistance)
{
  // return atan2 (robotDistance, focalPointHeight_) * fovConstant_;
  return 2 * fovConstant_ * tan (atan2 (robotDistance, focalPointHeight_) / 2);
}
