/**
 * @Filename FisheyeImageFrameRobotFrameTransformations.hpp
 * @Description Transform image frame coordinates to robot frame coordinates
 * and vice-versa for the fisheye setup.
 * @Status Implementing
 * @Version $Id: FisheyeImageFrameRobotFrameTransformations.hpp 5986 2013-04-24 10:53:37Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _FISHEYE_IMAGE_FRAME_ROBOT_FRAME_TRANSFORMATIONS_HPP_
#define _FISHEYE_IMAGE_FRAME_ROBOT_FRAME_TRANSFORMATIONS_HPP_

#include <opencv2/core/core.hpp>



/**
 * @Class FisheyeImageFrameRobotFrameTransformations ImageFrameRobotFrameTransformations.hpp "ImageFrameRobotFrameTransformations.hpp"
 * Class to transform image frame coordinates to robot frame coordinates
 * and vice-versa for the fisheye setup.
 * @Author Hugo Costelha
 */
class FisheyeImageFrameRobotFrameTransformations
{
  public:
    /**
     * Constructor that uses World Definition.
     */
    FisheyeImageFrameRobotFrameTransformations ();
    
    /**
     * Setter for camera center
     */
    void cameraCenterImg (cv::Point2d const& cameraCenterImg);
    
    /**
     * Transfom a point in x,y coordinates from the image frame to the robot
     * frame. The image points are measured from the center of the image
     * (positive X from left to right, positive Y from up to bottom).
     * @Argument imgX - x value in the image [pixels]
     * @Argument imgY - y value in the image [pixels]
     * @Argument robotX - will hold the X value in the robot frame [meters]
     * @Argument robotY - will hold the Y value in the robot frame [meters]
     */
    void fromImageFrame2RobotFrame (double imgX,
                                    double imgY,
                                    double& robotX,
                                    double& robotY);
                                    
    /**
     * Transfom a point in x,y coordinates from the robot frame to the image
     * frame. The image points are measured from the center of the image
     * (positive X from left to right, positive Y from up to bottom).
     * @Argument robotX - X value in the robot frame [meters]
     * @Argument robotY - Y value in the robot frame [meters]
     * @Argument imgX - will hold the X value in the image [pixels]
     * @Argument imgY - will hold the Y value in the image [pixels]
     */
    void fromRobotFrame2ImageFrame (double robotX,
                                    double robotY,
                                    double& imgX,
                                    double& imgY);
                                    
                                    
  private:
    void recalcParams();
    
    /**
     * Transfom a distance from the image frame to the robot frame. The image
     * distance is measured from the center of the image. In this case (Raw)
     * the camera offset is not taken into account, as we would need the angle
     * for that.
     * @Argument imgDistance - Distance in pixels measured in the image.
     * @Argument robotDistance - will hold the distance in meters.
     * @Author Marco Barbosa
     */
    double fromImageFrame2RobotFrameRaw (double imgDistance);
    
    /**
     * Transfom a distance from the robot frame to the image frame. The image
     * distance is measured from the center of the image. In this case (Raw)
     * the camera offset is not taken into account, as we would need the angle
     * for that.
     * @Argument robotDistance - Distance from the robot in meters.
     * @Argument imgDistance - Distance in pixels measured in the image.
     */
    double fromRobotFrame2ImageFrameRaw (double robotDistance);
    
    
    /// Pixel of camera center (column)
    double cameraCenterImgCol_;
    
    /// Pixel of camera center (row)
    double cameraCenterImgRow_;
    
    /// Pixel of robot center (column)
    double robotCenterImgCol_;
    
    /// Pixel of robot center (row)
    double robotCenterImgRow_;
    
    /// FOV constant
    double fovConstant_;
    
    /// Height of the focal point measured from the ground [meters]
    double focalPointHeight_;
    
    /// Distance along X in the floor from the center of the robot to the camera [meters]
    double xCamera_;
    
    /// Distance along Y in the floor from the center of the robot to the camera [meters]
    double yCamera_;
    
    /// Rotation angle
    double rotationAngle_;
};

#endif
