/**
 * @Filename ImageFrameRobotFrameTransformations.hpp
 * @Description Transform image frame coordinates to robot frame coordinates
 * and vice-versa.
 * @Status Implementing
 * @Version $Id: ImageFrameRobotFrameTransformations.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _IMAGE_FRAME_ROBOT_FRAME_TRANSFORMATIONS_HPP_
#define _IMAGE_FRAME_ROBOT_FRAME_TRANSFORMATIONS_HPP_



/**
 * @Class ImageFrameRobotFrameTransformations ImageFrameRobotFrameTransformations.hpp "ImageFrameRobotFrameTransformations.hpp"
 * Class to transform image frame coordinates to robot frame coordinates
 * and vice-versa.
 * @Author Hugo Costelha
 */
class ImageFrameRobotFrameTransformations
{
  public:
  
    /**
     * Constructor
     */
    ImageFrameRobotFrameTransformations();
    
    /**
     * Destructor
     */
    virtual ~ImageFrameRobotFrameTransformations();
    
    /**
     * Transfom a point in x,y coordinates from the image frame to the robot
     * frame. The image points are measured from the center of the image
     * (positive X from left to right, positive Y from up to bottom).
     * @Argument imgX - x value in the image [pixels]
     * @Argument imgY - y value in the image [pixels]
     * @Argument robotX - will hold the X value in the robot frame [meters]
     * @Argument robotY - will hold the Y value in the robot frame [meters]
     */
    virtual void fromImageFrame2RobotFrame (int imgX, int imgY,
                                            float* robotX, float* robotY) = 0;
                                            
    /**
     * Transfom a distance from the image frame to the robot frame. The image
     * distance is measured from the center of the image. In this case (Raw)
     * the camera offset is not taken into account, as we would need the angle
     * for that.
     * @Argument imgDistance - Distance in pixels measured in the image.
     * @Argument robotDistance - will hold the distance in meters.
     */
    virtual void fromImageFrame2RobotFrameRaw (float imgDistance,
        float* robotDistance) = 0;
        
    /**
     * Transfom a point in x,y coordinates from the robot frame to the image
     * frame. The image points are measured from the center of the image
     * (positive X from left to right, positive Y from up to bottom).
     * @Argument robotX - X value in the robot frame [meters]
     * @Argument robotY - Y value in the robot frame [meters]
     * @Argument imgX - will hold the X value in the image [pixels]
     * @Argument imgY - will hold the Y value in the image [pixels]
     */
    virtual void fromRobotFrame2ImageFrame (float robotX, float robotY,
                                            float* imgX, float* imgY) = 0;
                                            
    /**
     * Transfom a distance from the robot frame to the image frame. The image
     * distance is measured from the center of the image. In this case (Raw)
     * the camera offset is not taken into account, as we would need the angle
     * for that.
     * @Argument robotDistance - Distance from the robot in meters.
     * @Argument imgDistance - Distance in pixels measured in the image.
     */
    virtual void fromRobotFrame2ImageFrameRaw (float robotDistance,
        float* imgDistance) = 0;
        
    /**
     * Get method for the camera distance
     * @Returns float, the distance from the camera to the robot
     */
    float getCameraDistance();
    
  protected:
    int xRobotInImg_; /*!< X position of the robot in the image [pixels] */
    int yRobotInImg_; /*!< Y position of the robot in the image [pixels] */
    
    float xCamera_; /*!< Distance along X in the floor from the center of
                         * the robot to the camera [meters]*/
    float yCamera_; /*!< Distance along Y in the floor from the center of
                         * the robot to the camera [meters]*/
    float cameraDistance_; /*!< Distance, measured in in floor, from the robot
                            * center to the camera [meters] */
}; // class ImageFrameRobotFrameTransformations
#endif
