/**
 * @Filename LocalFrameWorldFrameTransformations.cpp
 * @Description LocalFrameWorldFrame class implementation.
 * @Status Finished
 * @Version $Id: LocalFrameWorldFrameTransformations.cpp 5661 2012-05-11 13:43:55Z jreis $
 * @Maintainer Hugo Costelha
 */

#include "LocalFrameWorldFrameTransformations.hpp"



namespace LocalFrameWorldFrameTransformations
{
  //
  // 2D+theta (Posture and VelocityPosture)
  //
  void rotateFrame (Posture* baseFrame, float rotationTheta)
  {
    baseFrame->setTheta (baseFrame->getTheta() + rotationTheta);
  }
  
  void translateFrame (Posture* frame,
                       float translationX,
                       float translationY)
  {
    frame->setXY (frame->getX() + translationX * cos (frame->getTheta())
                  - translationY * sin (frame->getTheta()),
                  frame->getY() + translationX * sin (frame->getTheta())
                  + translationY * cos (frame->getTheta()));
  }
  
  void translateFrame (Posture* frame, float translation)
  {
    frame->setDistance (frame->getDistance() + translation);
  }
  
  void translateLocalVector (Posture* vectorInBaseFrame,
                             float translationX,
                             float translationY)
  {
    vectorInBaseFrame->setXY (vectorInBaseFrame->getX() + translationX,
                              vectorInBaseFrame->getY() + translationY);
  }
  
  void translateLocalVector (Posture* vectorInBaseFrame, float translation)
  {
    vectorInBaseFrame->setDistance (vectorInBaseFrame->getDistance() +
                                    translation);
  }
  
  void rotateLocalVector (Posture* vectorInBaseFrame, float rotationTheta)
  {
    vectorInBaseFrame->setAzimuth (vectorInBaseFrame->getAzimuth()
                                   + rotationTheta);
    vectorInBaseFrame->setTheta (vectorInBaseFrame->getTheta()
                                 + rotationTheta);
  }
  
  void translateBaseFrameAndVector (Posture* baseFrame,
                                    Posture* vectorInBaseFrame,
                                    float translationX,
                                    float translationY)
  {
    translateFrame (baseFrame, translationX, translationY);
    translateLocalVector (vectorInBaseFrame, -translationX, -translationY);
  }
  
  void rotateBaseFrameAndVector (Posture* baseFrame,
                                 Posture* vectorInBaseFrame,
                                 float rotationTheta)
  {
    rotateFrame (baseFrame, rotationTheta);
    rotateLocalVector (vectorInBaseFrame, -rotationTheta);
  }
  
  void changeToGivenBaseFrame (Posture newBaseFrame,
                               Posture* vectorInBaseFrame)
  {
    translateLocalVector (vectorInBaseFrame,
                          -newBaseFrame.getX(),
                          -newBaseFrame.getY());
    rotateLocalVector (vectorInBaseFrame,
                       -newBaseFrame.getTheta());
  }
  
  void changeFromGivenBaseFrame (Posture oldBaseFrame,
                                 Posture* vectorInBaseFrame)
  {
    rotateLocalVector (vectorInBaseFrame,
                       oldBaseFrame.getTheta());
    translateLocalVector (vectorInBaseFrame,
                          oldBaseFrame.getX(),
                          oldBaseFrame.getY());
  }
  
  void local2World (Posture baseFrame,
                    Posture postureInBaseFrame,
                    Posture& postureInWorldFrame)
  {
    double sinTheta = sin (baseFrame.getTheta());
    double cosTheta = cos (baseFrame.getTheta());
    float x, y, theta;
    
    // Perform the transformation
    x = baseFrame.getX() + postureInBaseFrame.getX() * cosTheta
        - postureInBaseFrame.getY() * sinTheta;
    y = baseFrame.getY() + postureInBaseFrame.getX() * sinTheta
        + postureInBaseFrame.getY() * cosTheta;
    theta = baseFrame.getTheta() + postureInBaseFrame.getTheta();
    
    // Store the result
    postureInWorldFrame.setXYTheta (x, y, theta);
  }
  
  void world2Local (Posture baseFrame,
                    Posture postureInWorldFrame,
                    Posture& postureInBaseFrame)
  {
    double sinTheta = sin (baseFrame.getTheta());
    double cosTheta = cos (baseFrame.getTheta());
    float x, y, theta;
    
    // Do the transformation
    x = (postureInWorldFrame.getX() - baseFrame.getX()) * cosTheta +
        (postureInWorldFrame.getY() - baseFrame.getY()) * sinTheta;
    y =  - (postureInWorldFrame.getX() - baseFrame.getX()) * sinTheta +
         (postureInWorldFrame.getY() - baseFrame.getY()) * cosTheta;
    theta = postureInWorldFrame.getTheta() - baseFrame.getTheta();
    
    // Store the computed values
    postureInBaseFrame.setXYTheta (x, y, theta);
  }
  
  void local2World (Posture baseFrame,
                    VelocityPosture velocityInBaseFrame,
                    VelocityPosture& velocityInWorldFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInWorldFrame.setPolar (
      velocityInBaseFrame.getSpeed(),
      velocityInBaseFrame.getAzimuth() + baseFrame.getTheta(),
      velocityInBaseFrame.getAngularSpeed());
  }
  
  void world2Local (Posture baseFrame,
                    VelocityPosture velocityInWorldFrame,
                    VelocityPosture& velocityInBaseFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInBaseFrame.setPolar (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() - baseFrame.getTheta(),
      velocityInWorldFrame.getAngularSpeed());
  }
  
  void local2World (Posture baseFrame,
                    ControlInput controlInBaseFrame,
                    ControlInput& controlInWorldFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    controlInWorldFrame.setPolar (
      controlInBaseFrame.getMagnitude(),
      controlInBaseFrame.getAzimuth() + baseFrame.getTheta(),
      controlInBaseFrame.getAngular());
  }
  
  void world2Local (Posture baseFrame,
                    ControlInput controlInWorldFrame,
                    ControlInput& controlInBaseFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    controlInBaseFrame.setPolar (
      controlInWorldFrame.getMagnitude(),
      controlInWorldFrame.getAzimuth() - baseFrame.getTheta(),
      controlInWorldFrame.getAngular());
  }
  
  void local2World (Posture baseFrame,
                    Posture postureInBaseFrame,
                    VelocityPosture baseFrameVelocity,
                    VelocityPosture velocityInBaseFrame,
                    VelocityPosture& velocityInWorldFrame)
  {
    float crossX = -baseFrameVelocity.getAngularSpeed() * postureInBaseFrame.getY();
    float crossY = baseFrameVelocity.getAngularSpeed() * postureInBaseFrame.getX();
    // Get the full velocity
    velocityInWorldFrame.setVxVyVang (
      velocityInBaseFrame.getVx() + baseFrameVelocity.getVx() + crossX,
      velocityInBaseFrame.getVy() + baseFrameVelocity.getVy() + crossY,
      velocityInBaseFrame.getAngularSpeed() +
      baseFrameVelocity.getAngularSpeed());
      
    //         // Perform the transformation, which in this case corresponds to a
    //         // simple rotation
    velocityInWorldFrame.setPolar (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() + baseFrame.getTheta());
  }
  
  void world2Local (Posture baseFrame,
                    Posture postureInBaseFrame,
                    VelocityPosture baseFrameVelocity,
                    VelocityPosture velocityInWorldFrame,
                    VelocityPosture& velocityInBaseFrame)
  {
    float crossX = -baseFrameVelocity.getAngularSpeed() * postureInBaseFrame.getY();
    float crossY = baseFrameVelocity.getAngularSpeed() * postureInBaseFrame.getX();
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInBaseFrame.setPolar (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() - baseFrame.getTheta(),
      velocityInWorldFrame.getAngularSpeed());
      
    // Get the full velocity
    velocityInBaseFrame.setVxVyVang (
      velocityInBaseFrame.getVx() - baseFrameVelocity.getVx() - crossX,
      velocityInBaseFrame.getVy() - baseFrameVelocity.getVy() - crossY,
      velocityInBaseFrame.getAngularSpeed() -
      baseFrameVelocity.getAngularSpeed());
  }
  
  //
  // 2D
  //
  
  void translateLocalVector (Position2D* vectorInBaseFrame,
                             float translationX,
                             float translationY)
  {
    vectorInBaseFrame->setXY (vectorInBaseFrame->getX() + translationX,
                              vectorInBaseFrame->getY() + translationY);
  }
  
  void translateLocalVector (Position2D* vectorInBaseFrame,
                             float translation)
  {
    vectorInBaseFrame->setDistance (vectorInBaseFrame->getDistance()
                                    + translation);
  }
  
  void rotateLocalVector (Position2D* vectorInBaseFrame,
                          float rotationTheta)
  {
    vectorInBaseFrame->setAzimuth (vectorInBaseFrame->getAzimuth()
                                   + rotationTheta);
  }
  
  void translateBaseFrameAndVector (Posture* baseFrame,
                                    Position2D* vectorInBaseFrame,
                                    float translationX,
                                    float translationY)
  {
    translateFrame (baseFrame, translationX, translationY);
    translateLocalVector (vectorInBaseFrame, -translationX, -translationY);
  }
  
  void rotateBaseFrameAndVector (Posture* baseFrame,
                                 Position2D* vectorInBaseFrame,
                                 float rotationTheta)
  {
    rotateFrame (baseFrame, rotationTheta);
    rotateLocalVector (vectorInBaseFrame, -rotationTheta);
  }
  
  void changeToGivenBaseFrame (Posture newBaseFrame,
                               Position2D* vectorInBaseFrame)
  {
    translateLocalVector (vectorInBaseFrame,
                          -newBaseFrame.getX(),
                          -newBaseFrame.getY());
    rotateLocalVector (vectorInBaseFrame,
                       -newBaseFrame.getTheta());
  }
  
  void changeFromGivenBaseFrame (Posture oldBaseFrame,
                                 Position2D* vectorInBaseFrame)
  {
    rotateLocalVector (vectorInBaseFrame,
                       oldBaseFrame.getTheta());
    translateLocalVector (vectorInBaseFrame,
                          oldBaseFrame.getX(),
                          oldBaseFrame.getY());
  }
  
  void local2World (Posture baseFrame,
                    Position2D vectorInBaseFrame,
                    Position2D& vectorInWorldFrame)
  {
    double sinTheta = sin (baseFrame.getTheta());
    double cosTheta = cos (baseFrame.getTheta());
    float x, y;
    
    // Perform the transformation
    x = baseFrame.getX() + vectorInBaseFrame.getX() * cosTheta
        - vectorInBaseFrame.getY() * sinTheta;
    y = baseFrame.getY() + vectorInBaseFrame.getX() * sinTheta
        + vectorInBaseFrame.getY() * cosTheta;
        
    // Store the result
    vectorInWorldFrame.setXY (x, y);
  }
  
  void world2Local (Posture baseFrame,
                    Position2D vectorInWorldFrame,
                    Position2D& vectorInBaseFrame)
  {
    double sinTheta = sin (baseFrame.getTheta());
    double cosTheta = cos (baseFrame.getTheta());
    float x, y;
    
    // Do the transformation
    x = (vectorInWorldFrame.getX() - baseFrame.getX()) * cosTheta +
        (vectorInWorldFrame.getY() - baseFrame.getY()) * sinTheta;
    y =  - (vectorInWorldFrame.getX() - baseFrame.getX()) * sinTheta +
         (vectorInWorldFrame.getY() - baseFrame.getY()) * cosTheta;
         
    // Store the computed values
    vectorInBaseFrame.setXY (x, y);
  }
  
  void local2World (Posture baseFrame,
                    Velocity2D velocityInBaseFrame,
                    Velocity2D& velocityInWorldFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInWorldFrame.setPolar (
      velocityInBaseFrame.getSpeed(),
      velocityInBaseFrame.getAzimuth() + baseFrame.getTheta());
  }
  
  void world2Local (Posture baseFrame,
                    Velocity2D velocityInWorldFrame,
                    Velocity2D& velocityInBaseFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInBaseFrame.setPolar (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() - baseFrame.getTheta());
  }
  
  void local2World (Posture baseFrame,
                    Position2D positionInBaseFrame,
                    VelocityPosture baseFrameVelocity,
                    Velocity2D velocityInBaseFrame,
                    Velocity2D& velocityInWorldFrame)
  {
    float crossX = -baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getY();
    float crossY = baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getX();
    // Get the full velocity
    velocityInWorldFrame.setVxVy (
      velocityInBaseFrame.getVx() + baseFrameVelocity.getVx() + crossX ,
      velocityInBaseFrame.getVy() + baseFrameVelocity.getVy() + crossY);
      
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInWorldFrame.setPolar (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() + baseFrame.getTheta());
  }
  
  void world2Local (Posture baseFrame,
                    Position2D positionInBaseFrame,
                    VelocityPosture baseFrameVelocity,
                    Velocity2D velocityInWorldFrame,
                    Velocity2D& velocityInBaseFrame)
  {
    float crossX = -baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getY();
    float crossY = baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getX();
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInBaseFrame.setPolar (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() - baseFrame.getTheta());
      
    // Get the full velocity
    velocityInBaseFrame.setVxVy (
      velocityInBaseFrame.getVx() - baseFrameVelocity.getVx() - crossX,
      velocityInBaseFrame.getVy() - baseFrameVelocity.getVy() - crossY);
  }
  
  //
  // 3D
  //
  
  void translateLocalVector (Position3D* vectorInBaseFrame,
                             float translationX,
                             float translationY,
                             float translationZ)
  {
    vectorInBaseFrame->setXYZ (vectorInBaseFrame->getX() + translationX,
                               vectorInBaseFrame->getY() + translationY,
                               vectorInBaseFrame->getZ() + translationZ);
  }
  
  void translateLocalVector (Position3D* vectorInBaseFrame,
                             float translationX,
                             float translationY)
  {
    vectorInBaseFrame->setX (vectorInBaseFrame->getX() + translationX);
    vectorInBaseFrame->setY (vectorInBaseFrame->getY() + translationY);
  }
  
  void translateLocalVector (Position3D* vectorInBaseFrame,
                             float translation)
  {
    vectorInBaseFrame->setDistance (vectorInBaseFrame->getDistance() +
                                    translation);
  }
  
  void rotateLocalVector (Position3D* vectorInBaseFrame,
                          float rotationTheta)
  {
    vectorInBaseFrame->setAzimuth (vectorInBaseFrame->getAzimuth() +
                                   rotationTheta);
  }
  
  void translateBaseFrameAndVector (Posture* baseFrame,
                                    Position3D* vectorInBaseFrame,
                                    float translationX,
                                    float translationY)
  {
    translateFrame (baseFrame, translationX, translationY);
    translateLocalVector (vectorInBaseFrame, -translationX, -translationY);
  }
  
  void rotateBaseFrameAndVector (Posture* baseFrame,
                                 Position3D* vectorInBaseFrame,
                                 float rotationTheta)
  {
    rotateFrame (baseFrame, rotationTheta);
    rotateLocalVector (vectorInBaseFrame, -rotationTheta);
  }
  
  void changeToGivenBaseFrame (Posture newBaseFrame,
                               Position3D* vectorInBaseFrame)
  {
    translateLocalVector (vectorInBaseFrame,
                          -newBaseFrame.getX(),
                          -newBaseFrame.getY());
    rotateLocalVector (vectorInBaseFrame,
                       -newBaseFrame.getTheta());
  }
  
  void changeFromGivenBaseFrame (Posture oldBaseFrame,
                                 Position3D* vectorInBaseFrame)
  {
    rotateLocalVector (vectorInBaseFrame,
                       oldBaseFrame.getTheta());
    translateLocalVector (vectorInBaseFrame,
                          oldBaseFrame.getX(),
                          oldBaseFrame.getY());
  }
  
  void local2World (Posture baseFrame,
                    Position3D vectorInBaseFrame,
                    Position3D& vectorInWorldFrame)
  {
    double sinTheta = sin (baseFrame.getTheta());
    double cosTheta = cos (baseFrame.getTheta());
    
    // Perform the transformation and store the result
    vectorInWorldFrame.setXYZ (
      baseFrame.getX() + vectorInBaseFrame.getX() *cosTheta
      - vectorInBaseFrame.getY() *sinTheta,
      baseFrame.getY() + vectorInBaseFrame.getX() *sinTheta
      + vectorInBaseFrame.getY() *cosTheta,
      vectorInBaseFrame.getZ());
  }
  
  void world2Local (Posture baseFrame,
                    Position3D vectorInWorldFrame,
                    Position3D& vectorInBaseFrame)
  {
    double sinTheta = sin (baseFrame.getTheta());
    double cosTheta = cos (baseFrame.getTheta());
    
    // Perform the transformation and store the computed values
    vectorInBaseFrame.setXYZ (
      (vectorInWorldFrame.getX() - baseFrame.getX()) *cosTheta +
      (vectorInWorldFrame.getY() - baseFrame.getY()) *sinTheta,
      - (vectorInWorldFrame.getX() - baseFrame.getX()) *sinTheta +
      (vectorInWorldFrame.getY() - baseFrame.getY()) *cosTheta,
      vectorInWorldFrame.getZ());
  }
  
  void local2World (Posture baseFrame,
                    Velocity3D velocityInBaseFrame,
                    Velocity3D& velocityInWorldFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInWorldFrame.setSpherical (
      velocityInBaseFrame.getSpeed(),
      velocityInBaseFrame.getAzimuth() + baseFrame.getTheta(),
      velocityInBaseFrame.getZenith());
  }
  
  void world2Local (Posture baseFrame,
                    Velocity3D velocityInWorldFrame,
                    Velocity3D& velocityInBaseFrame)
  {
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInBaseFrame.setSpherical (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() - baseFrame.getTheta(),
      velocityInWorldFrame.getZenith());
  }
  
  void local2World (Posture baseFrame,
                    Position3D positionInBaseFrame,
                    VelocityPosture baseFrameVelocity,
                    Velocity3D velocityInBaseFrame,
                    Velocity3D& velocityInWorldFrame)
  {
    float crossX = -baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getY();
    float crossY = baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getX();
    // Get the full velocity
    velocityInWorldFrame.setVxVyVz (
      velocityInBaseFrame.getVx() + baseFrameVelocity.getVx() + crossX,
      velocityInBaseFrame.getVy() + baseFrameVelocity.getVy() + crossY,
      velocityInBaseFrame.getVz());
      
    // Perform the transformation, which in this case corresponds to a
    // simple rotation
    velocityInWorldFrame.setSpherical (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() + baseFrame.getTheta(),
      velocityInWorldFrame.getZenith());
  }
  
  void world2Local (Posture baseFrame,
                    Position3D positionInBaseFrame,
                    VelocityPosture baseFrameVelocity,
                    Velocity3D velocityInWorldFrame,
                    Velocity3D& velocityInBaseFrame)
  {
    float crossX = -baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getY();
    float crossY = baseFrameVelocity.getAngularSpeed() * positionInBaseFrame.getX();
    // Perform the transformation, which in this case corresponds to a simple
    // rotation
    velocityInBaseFrame.setSpherical (
      velocityInWorldFrame.getSpeed(),
      velocityInWorldFrame.getAzimuth() - baseFrame.getTheta(),
      velocityInWorldFrame.getZenith());
      
    // Get the full velocity
    velocityInBaseFrame.setVxVyVz (
      velocityInBaseFrame.getVx() - baseFrameVelocity.getVx() - crossX,
      velocityInBaseFrame.getVy() - baseFrameVelocity.getVy() - crossY,
      velocityInBaseFrame.getVz());
  }
}
