/**
 * @Filename LocalFrameWorldFrameTransformations.hpp
 * @Description LocalFrameWorldFrameTransformations class header
 * @Status Finished
 * @Version $Id: LocalFrameWorldFrameTransformations.hpp 5864 2012-12-08 14:40:54Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _LOCAL_FRAME_WORLD_FRAME_TRANSFORMATIONS_HPP_
#define _LOCAL_FRAME_WORLD_FRAME_TRANSFORMATIONS_HPP_

#include <Defines.hpp>

#include <Position2D.hpp>
#include <Position3D.hpp>
#include <Posture.hpp>
#include <Velocity2D.hpp>
#include <Velocity3D.hpp>
#include <VelocityPosture.hpp>
#include <ControlInput.hpp>

#include <socrob/math.h>
using namespace socrob;



/**
 * @Namespace LocalFrameWorldFrameTransformations
 * Class to transform local coordinates to world coordinates and vice-versa.
 * @Author Hugo Costelha
 */
namespace LocalFrameWorldFrameTransformations
{
  //
  // 2D+theta (Posture and VelocityPosture)
  //
  
  /**
   * Rotate a given frame. This rotation is perfomed centered on the
   * baseFrame origin, not the world frame
   * @Argument baseFrame - frame to be rotated.
   * @Argument rotationTheta - angle of rotation.
   */
  void rotateFrame (Posture* baseFrame, float rotationTheta);
  
  /**
   * Translate a given frame along the axis defined by the frame, i.e., we
   * will move X meters and Y meters along the frame defined by the posture.
   * @Argument frame - frame to be updated.
   * @Argument translationX - x translation value of the base frame.
   * @Argument translationY - y translation value of the base frame.
   */
  void translateFrame (Posture* frame,
                       float translationX,
                       float translationY);
                       
  /**
   * Translate a given frame along the axis defined by the frame (using polar
   * coordinates here)/
   * @Argument frame - frame to be updated.
   * @Argument translation - translation value of the base frame.
   */
  void translateFrame (Posture* frame, float translation);
  
  /**
   * Update a vector considering the given translation of its base frame ,
   * i.e., we will just decrease X meters and Y meters.
   * @Argument vectorInBaseFrame - vector to be updated.
   * @Argument translationX - x translation value of the local vector.
   * @Argument translationY - y translation value of the local vector.
   */
  void translateLocalVector (Posture* vectorInBaseFrame,
                             float translationX,
                             float translationY);
                             
  /**
   * Translate a given vector along the axis defined by its base frame, i.e.,
   * we will just increase X meters and Y meters.
   * @Argument vectorInBaseFrame - vector to be updated.
   * @Argument translation - translation value of the local vector.
   */
  void translateLocalVector (Posture* vectorInBaseFrame, float translation);
  
  /**
   * Rotate a given vector. This rotation is perfomed centered on the
   * baseFrame origin, not the world frame.
   * @Argument vectorInBaseFrame - vector to be rotated.
   * @Argument rotationTheta - angle of rotation of the vector.
   */
  void rotateLocalVector (Posture* vectorInBaseFrame, float rotationTheta);
  
  /**
   * This is similar to the previous one, but we also update a posture given
   * on the base frame.
   * @Argument baseFrame - base frame to be updated.
   * @Argument vectorInBaseFrame - vector, in the base frame, to update.
   * @Argument translationX - x translation value of the base frame.
   * @Argument translationY - y translation value of the base frame.
   */
  void translateBaseFrameAndVector (Posture* baseFrame,
                                    Posture* vectorInBaseFrame,
                                    float translationX,
                                    float translationY);
                                    
  /**
   * This is similar with the previous one, but we also update a given
   * posture, measured in the base frame.
   * @Argument localFrame - frame to be rotated.
   * @Argument vectorInBaseFrame - posture, measured in the local frame, to
   * be updated.
   * @Argument rotationTheta - angle of rotation.
   */
  void rotateBaseFrameAndVector (Posture* baseFrame,
                                 Posture* vectorInBaseFrame,
                                 float rotationTheta);
                                 
  /**
   * Given a vector, measured in its base frame, a new base frame, change
   * the vector to the new base frame.
   * @Argument newBaseFrame - new base frame of the vector.
   * @Argument vectorInBaseFrame - vector to be updated.
   */
  void changeToGivenBaseFrame (const Posture newBaseFrame,
                               Posture* vectorInBaseFrame);
                               
  /**
   * Given a vector, measured in its base frame, and an old base frame, change
   * the vector back to its old base frame. Running changeToGivenBaseFrame
   * followed by changeFromGivenBaseFrame would yeld no changes.
   * @Argument oldBaseFrame - old base frame of the vector to return to.
   * @Argument vectorInBaseFrame - vector to be updated.
   */
  void changeFromGivenBaseFrame (const Posture oldBaseFrame,
                                 Posture* vectorInBaseFrame);
                                 
  /**
   * Transform a given posture in local coordinates, referenced to
   * baseFrame into global coordinates.
   * @Argument baseFrame - the value of the base frame
   * @Argument postureInBaseFrame - the posture coordinates in the base
   * frame
   * @Argument postureInWorldFrame - value to be returned, containing the
   * given local posture in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Posture postureInBaseFrame,
                    Posture& postureInWorldFrame);
                    
  /**
   * Transform a given posture in world coordinates, into a posture based
   * on the given baseFrame.
   * @Argument baseFrame - the value of the base frame
   * @Argument postureInWorldFrame - the posture coordinates in the world
   * frame.
   * @Argument postureInBaseFrame - value to be returned, containing the
   * given world posture in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Posture postureInWorldFrame,
                    Posture& postureInBaseFrame);
                    
  /**
   * Transform a given posture velocity in local coordinates, referenced to
   * baseFrame into global coordinates.
   * @Argument baseFrame - the value of the base frame
   * @Argument velocityInBaseFrame - the velocity coordinates in the base
   * frame
   * @Argument velocityInWorldFrame - value to be returned, containing the
   * given local velocity in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const VelocityPosture velocityInBaseFrame,
                    VelocityPosture& velocityInWorldFrame);
                    
  /**
   * Transform a given posture velocity in world coordinates, into a posture
   * based
   * on the given baseFrame.
   * @Argument baseFrame - the value of the base frame
   * @Argument velocityInWorldFrame - the velocity coordinates in the world
   * frame.
   * @Argument velocityInBaseFrame - value to be returned, containing the
   * given world velocity in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const VelocityPosture velocityInWorldFrame,
                    VelocityPosture& velocityInBaseFrame);
                    
  /**
   * Transform a given control in local coordinates, referenced to
   * baseFrame into global coordinates.
   * @Argument baseFrame - the value of the base frame
   * @Argument controlInBaseFrame - the control coordinates in the base
   * frame
   * @Argument controlInWorldFrame - value to be returned, containing the
   * given local control in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const ControlInput controlInBaseFrame,
                    ControlInput& controlInWorldFrame);
                    
  /**
   * Transform a given control in world coordinates, into a control
   * based
   * on the given baseFrame.
   * @Argument baseFrame - the value of the base frame
   * @Argument controlInWorldFrame - the control coordinates in the world
   * frame.
   * @Argument controlInBaseFrame - value to be returned, containing the
   * given world control in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const ControlInput controlInWorldFrame,
                    ControlInput& controlInBaseFrame);
                    
  /**
   * Transform a given velocity in local coordinates, referenced to
   * baseFrame into global coordinates. In this case, we tak into account the
   * velocity of the observer, i.e., of the base frame.
   * @Argument baseFrame - the value the base frame
   * @Argument baseFrameVelocity - velocity of the base frame in local
   * coordinates.
   * @Argument velocityInBaseFrame - the velocity coordinates in the base
   * frame
   * @Argument velocityInWorldFrame - value to be returned, containing the
   * given local velocity in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Posture postureInBaseFrame,
                    const VelocityPosture baseFrameVelocity,
                    const VelocityPosture velocityInBaseFrame,
                    VelocityPosture& velocityInWorldFrame);
                    
  /**
   * Transform a given velocity in world coordinates, into a velocity based
   * on the given baseFrame.  In this case, we tak into account the
   * velocity of the observer, i.e., of the base frame.
   * @Argument baseFrame - the value of the base frame
   * @Argument baseFrameVelocity - velocity of the base frame in local
   * coordinates.
   * @Argument velocityInWorldFrame - the velocity coordinates in the world
   * frame.
   * @Argument velocityInBaseFrame - value to be returned, containing the
   * given world velocity in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Posture postureInBaseFrame,
                    const VelocityPosture baseFrameVelocity,
                    const VelocityPosture velocityInWorldFrame,
                    VelocityPosture& velocityInBaseFrame);
                    
  //
  // 2D
  //
  
  /**
   * Translate a given vector along the axis defined by its base frame, i.e.,
   * we will just increase X meters and Y meters.
   * @Argument vectorInBaseFrame - vector to be updated.
   * @Argument translationX - x translation value of the local vector.
   * @Argument translationY - y translation value of the local vector.
   */
  void translateLocalVector (Position2D* vectorInBaseFrame,
                             float translationX,
                             float translationY);
                             
  /**
   * Translate a given vector along the axis defined by its base frame (here
   * in polar coordinates).
   * @Argument vectorInBaseFrame - vector to be updated.
   * @Argument translation - translation value of the local vector.
   */
  void translateLocalVector (Position2D* vectorInBaseFrame,
                             float translation);
                             
  /**
   * Rotate a given vector. This rotation is perfomed centered on the
   * baseFrame origin, not the world frame.
   * @Argument vectorInBaseFrame - vector to be rotated.
   * @Argument rotationTheta - angle of rotation of the vector.
   */
  void rotateLocalVector (Position2D* vectorInBaseFrame,
                          float rotationTheta);
                          
  /**
   * This is similar to the previous one, but instead of updating a posture
   * given on the base frame, we update a vector.
   * @Argument frame - frame to be updated.
   * @Argument vectorInBaseFrame - vector, in the base frame, to update.
   * @Argument translationX - x translation value of the base frame.
   * @Argument translationY - y translation value of the base frame.
   */
  void translateBaseFrameAndVector (Posture* baseFrame,
                                    Position2D* vectorInBaseFrame,
                                    float translationX,
                                    float translationY);
                                    
  /**
   * This is similar with the previous one, but instead of updating a given
   * posture, measured in the base frame, we update a vector (also measured
   * in the base frame).
   * @Argument baseFrame - frame to be rotated.
   * @Argument vectorInBaseFrame - vector, measured in the base frame, to
   * be updated.
   * @Argument rotationTheta - angle of rotation.
   */
  void rotateBaseFrameAndVector (Posture* baseFrame,
                                 Position2D* vectorInBaseFrame,
                                 float rotationTheta);
                                 
  /**
   * Given a vector, measured in its base frame, a new base frame, change
   * the vector to the new base frame.
   * @Argument newBaseFrame - new base frame of the vector.
   * @Argument vectorInBaseFrame - vector to be updated.
   */
  void changeToGivenBaseFrame (const Posture newBaseFrame,
                               Position2D* vectorInBaseFrame);
                               
  /**
   * Given a vector, measured in its base frame, and an old base frame, change
   * the vector back to its old base frame. Running changeToGivenBaseFrame
   * followed by changeFromGivenBaseFrame would yeld no changes.
   * @Argument oldBaseFrame - old base frame of the vector to return to.
   * @Argument vectorInBaseFrame - vector to be updated.
   */
  void changeFromGivenBaseFrame (const Posture oldBaseFrame,
                                 Position2D* vectorInBaseFrame);
                                 
  /**
   * Transform a given vector in local coordinates, referenced to
   * localFrame into global coordinates.
   * @Argument baseFrame - the values of the base frame
   * @Argument vectorInLocalFrame - the vector coordinates in the basee
   * frame
   * @Argument vectorInWorldFrame - value to be returned, containing the
   * given local vector in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Position2D vectorInBaseFrame,
                    Position2D& vectorInWorldFrame);
                    
  /**
   * Transform a given vector in world coordinates, into a vector based
   * on the given baseFrame.
   * @Argument baseFrame - the value of the base frame
   * @Argument vectorInWorldFrame - the vector coordinates in the world
   * frame.
   * @Argument vectorInBaseFrame - value to be returned, containing the
   * given world vector in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Position2D vectorInWorldFrame,
                    Position2D& vectorInBaseFrame);
                    
  /**
   * Transform a given 2D velocity in local coordinates, referenced to
   * localFrame into global coordinates.
   * @Argument baseFrame - the values of the base frame
   * @Argument velocityInLocalFrame - the velocity coordinates in the basee
   * frame
   * @Argument velocityInWorldFrame - value to be returned, containing the
   * given local velocity in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Velocity2D velocityInBaseFrame,
                    Velocity2D& velocityInWorldFrame);
                    
  /**
   * Transform a given velocity in world coordinates, into a velocity based
   * on the given baseFrame.
   * @Argument baseFrame - the value of the base frame
   * @Argument velocityInWorldFrame - the velocity coordinates in the world
   * frame.
   * @Argument velocityInBaseFrame - value to be returned, containing the
   * given world velocity in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Velocity2D velocityInWorldFrame,
                    Velocity2D& velocityInBaseFrame);
                    
  /**
   * Transform a given velocity in local coordinates, referenced to
   * baseFrame into global coordinates. In this case, we tak into account the
   * velocity of the observer, i.e., of the base frame.
   * @Argument baseFrame - the value the base frame
   * @Argument baseFrameVelocity - velocity of the base frame in local
   * coordinates.
   * @Argument velocityInBaseFrame - the velocity coordinates in the base
   * frame
   * @Argument velocityInWorldFrame - value to be returned, containing the
   * given local velocity in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Position2D positionInBaseFrame,
                    const VelocityPosture baseFrameVelocity,
                    const Velocity2D velocityInBaseFrame,
                    Velocity2D& velocityInWorldFrame);
                    
  /**
   * Transform a given velocity in world coordinates, into a velocity based
   * on the given baseFrame.  In this case, we tak into account the
   * velocity of the observer, i.e., of the base frame.
   * @Argument baseFrame - the value of the base frame
   * @Argument baseFrameVelocity - velocity of the base frame in local
   * coordinates.
   * @Argument velocityInWorldFrame - the velocity coordinates in the world
   * frame.
   * @Argument velocityInBaseFrame - value to be returned, containing the
   * given world velocity in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Position2D positionInBaseFrame,
                    const VelocityPosture baseFrameVelocity,
                    const Velocity2D velocityInWorldFrame,
                    Velocity2D& velocityInBaseFrame);
  //
  // 3D
  //
  
  /**
   * Translate a given vector along the axis defined by its base frame, i.e.,
   * we will just increase X, Y and Z meters.
   * @Argument vectorInBaseFrame - vector to be updated.
   * @Argument translationX - x translation value of the local vector.
   * @Argument translationY - y translation value of the local vector.
   * @Argument translationz - z translation value of the local vector.
   */
  void translateLocalVector (Position3D* vectorInBaseFrame,
                             float translationX,
                             float translationY,
                             float translationZ);
                             
  /**
   * Translate a given vector along the axis defined by its base frame, i.e.,
   * we will just increase X, Y meters. (Z will not be changed)
   * @Argument vectorInBaseFrame - vector to be updated.
   * @Argument translationX - x translation value of the local vector.
   * @Argument translationY - y translation value of the local vector.
   */
  void translateLocalVector (Position3D* vectorInBaseFrame,
                             float translationX,
                             float translationY);
                             
  /**
   * Translate a given vector along the axis defined by its base frame (we
   * use spherical coordinates here).
   * @Argument vectorInBaseFrame - vector to be updated.
   * @Argument translation - translation value of the local vector.
   */
  void translateLocalVector (Position3D* vectorInBaseFrame,
                             float translation);
                             
  /**
   * Rotate a given vector in the XY plane. This rotation is perfomed centered
   * on the baseFrame origin, not the world frame.
   * @Argument vectorInBaseFrame - vector to be rotated.
   * @Argument rotationTheta - angle of rotation of the vector.
   */
  void rotateLocalVector (Position3D* vectorInBaseFrame,
                          float rotationTheta);
                          
  /**
   * This is similar to the previous one, but instead of updating a posture
   * given on the base frame, we update a vector.
   * @Argument frame - frame to be updated.
   * @Argument vectorInBaseFrame - vector, in the base frame, to update.
   * @Argument translationX - x translation value of the base frame.
   * @Argument translationY - y translation value of the base frame.
   */
  void translateBaseFrameAndVector (Posture* baseFrame,
                                    Position3D* vectorInBaseFrame,
                                    float translationX,
                                    float translationY);
                                    
  /**
   * This is similar with the previous one, but instead of updating a given
   * posture, measured in the base frame, we update a vector (also measured
   * in the base frame).
   * @Argument baseFrame - frame to be rotated.
   * @Argument vectorInBaseFrame - vector, measured in the base frame, to
   * be updated.
   * @Argument rotationTheta - angle of rotation.
   */
  void rotateBaseFrameAndVector (Posture* baseFrame,
                                 Position3D* vectorInBaseFrame,
                                 float rotationTheta);
                                 
  /**
   * Given a vector, measured in its base frame, a new base frame, change
   * the vector to the new base frame.
   * @Argument newBaseFrame - new base frame of the vector.
   * @Argument vectorInBaseFrame - vector to be updated.
   */
  void changeToGivenBaseFrame (const Posture newBaseFrame,
                               Position3D* vectorInBaseFrame);
                               
  /**
   * Given a vector, measured in its base frame, and an old base frame, change
   * the vector back to its old base frame. Running changeToGivenBaseFrame
   * followed by changeFromGivenBaseFrame would yeld no changes.
   * @Argument oldBaseFrame - old base frame of the vector to return to.
   * @Argument vectorInBaseFrame - vector to be updated.
   */
  void changeFromGivenBaseFrame (const Posture oldBaseFrame,
                                 Position3D* vectorInBaseFrame);
                                 
  /**
   * Transform a given vector in local coordinates, referenced to
   * baseFrame into global coordinates.
   * @Argument baseFrame - the value the base frame
   * @Argument vectorInBaseFrame - the vector coordinates in the base
   * frame
   * @Argument vectorInWorldFrame - value to be returned, containing the
   * given local vector in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Position3D vectorInBaseFrame,
                    Position3D& vectorInWorldFrame);
                    
  /**
   * Transform a given vector in world coordinates, into a vector based
   * on the given baseFrame.
   * @Argument baseFrame - the value of the base frame
   * @Argument vectorInWorldFrame - the vector coordinates in the world
   * frame.
   * @Argument vectorInBaseFrame - value to be returned, containing the
   * given world vector in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Position3D vectorInWorldFrame,
                    Position3D& vectorInBaseFrame);
                    
  /**
   * Transform a given velocity in local coordinates, referenced to
   * baseFrame into global coordinates.
   * @Argument baseFrame - the value the base frame
   * @Argument velocityInBaseFrame - the velocity coordinates in the base
   * frame
   * @Argument velocityInWorldFrame - value to be returned, containing the
   * given local velocity in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Velocity3D velocityInBaseFrame,
                    Velocity3D& velocityInWorldFrame);
                    
  /**
   * Transform a given velocity in world coordinates, into a velocity based
   * on the given baseFrame.
   * @Argument baseFrame - the value of the base frame
   * @Argument velocityInWorldFrame - the velocity coordinates in the world
   * frame.
   * @Argument velocityInBaseFrame - value to be returned, containing the
   * given world velocity in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Velocity3D vectorInWorldFrame,
                    Velocity3D& vectorInBaseFrame);
                    
  /**
   * Transform a given velocity in local coordinates, referenced to
   * baseFrame into global coordinates. In this case, we tak into account the
   * velocity of the observer, i.e., of the base frame.
   * @Argument baseFrame - the value the base frame
   * @Argument baseFrameVelocity - velocity of the base frame in local
   * coordinates.
   * @Argument velocityInBaseFrame - the velocity coordinates in the base
   * frame
   * @Argument velocityInWorldFrame - value to be returned, containing the
   * given local velocity in world coordinates
   */
  void local2World (const Posture baseFrame,
                    const Position3D positionInBaseFrame,
                    const VelocityPosture baseFrameVelocity,
                    const Velocity3D velocityInBaseFrame,
                    Velocity3D& velocityInWorldFrame);
                    
  /**
   * Transform a given velocity in world coordinates, into a velocity based
   * on the given baseFrame.  In this case, we tak into account the
   * velocity of the observer, i.e., of the base frame.
   * @Argument baseFrame - the value of the base frame
   * @Argument baseFrameVelocity - velocity of the base frame in local
   * coordinates.
   * @Argument velocityInWorldFrame - the velocity coordinates in the world
   * frame.
   * @Argument velocityInBaseFrame - value to be returned, containing the
   * given world velocity in local coordinates
   */
  void world2Local (const Posture baseFrame,
                    const Position3D positionInBaseFrame,
                    const VelocityPosture baseFrameVelocity,
                    const Velocity3D velocityInWorldFrame,
                    Velocity3D& velocityInBaseFrame);
} // class LocalFrameWorldFrameTransformations

#endif
