/**
 * @Filename MirrorImageFrameRobotFrameTransformations.hpp
 * @Description Transform image frame coordinates to robot frame coordinates
 * and vice-versa for the camera+mirror setup.
 * @Status Implementing
 * @Version $Id: MirrorImageFrameRobotFrameTransformations.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _MIRROR_IMAGE_FRAME_ROBOT_FRAME_TRANSFORMATIONS_HPP_
#define _MIRROR_IMAGE_FRAME_ROBOT_FRAME_TRANSFORMATIONS_HPP_

#include "ImageFrameRobotFrameTransformations.hpp"



/**
 * @Class MirrorImageFrameRobotFrameTransformations ImageFrameRobotFrameTransformations.hpp "ImageFrameRobotFrameTransformations.hpp"
 * Class to transform image frame coordinates to robot frame coordinates
 * and vice-versa for the mirror setup.
 * @Author Hugo Costelha
 */
class MirrorImageFrameRobotFrameTransformations :
  public ImageFrameRobotFrameTransformations
{
  public:
  
    /**
     * This method transforms pixel distance from the center of the image from
     * pixels to meters.
     * @Argument gain - the gain of the camera+mirror [m/pixels]
     * @Argument xCenter - X robot position in the image [pixels]
     * @Argument yCenter - Y robot position in the image [pixels]
     */
    MirrorImageFrameRobotFrameTransformations (float gain,
        int xRobotInImg,
        int yRobotInImg);
        
    /**
     * Destructor
     */
    ~MirrorImageFrameRobotFrameTransformations();
    
    /**
     * Transfom a point in x,y coordinates from the image frame to the robot
     * frame. The image points are measured from the center of the image
     * (positive X from left to right, positive Y from up to bottom).
     * @Argument imgX - x value in the image [pixels]
     * @Argument imgY - y value in the image [pixels]
     * @Argument robotX - will hold the X value in the robot frame [meters]
     * @Argument robotY - will hold the Y value in the robot frame [meters]
     */
    void fromImageFrame2RobotFrame (int imgX, int imgY,
                                    float* robotX, float* robotY);
                                    
    /**
     * Transfom a distance from the image frame to the robot frame. The image
     * distance is measured from the center of the image. In this case (Raw)
     * the camera offset is not taken into account, as we would need the angle
     * for that.
     * @Argument imgDistance - Distance in pixels measured in the image.
     * @Argument robotDistance - will hold the distance in meters.
     */
    void fromImageFrame2RobotFrameRaw (float imgDistance, float* robotDistance);
    
    /**
     * Transfom a point in x,y coordinates from the robot frame to the image
     * frame. The image points are measured from the center of the image
     * (positive X from left to right, positive Y from up to bottom).
     * @Argument robotX - X value in the robot frame [meters]
     * @Argument robotY - Y value in the robot frame [meters]
     * @Argument imgX - will hold the X value in the image [pixels]
     * @Argument imgY - will hold the Y value in the image [pixels]
     */
    void fromRobotFrame2ImageFrame (float robotX, float robotY,
                                    float* imgX, float* imgY);
                                    
    /**
     * Transfom a distance from the robot frame to the image frame. The image
     * distance is measured from the center of the image. In this case (Raw)
     * the camera offset is not taken into account.
     * @Argument robotDistance - Distance from the robot in meters.
     * @Argument imgDistance - Distance in pixels measured in the image.
     */
    void fromRobotFrame2ImageFrameRaw (float robotDistance, float* imgDistance);
    
  private:
    float gain_; /*! Gain in the image [m/pixels] */
}; // class MirrorImageFrameRobotFrameTransformations
#endif
