#ifndef _WORLD_DEFINITION_HPP_
#define _WORLD_DEFINITION_HPP_

#include <string>

#include <Posture.hpp>

#include <boost/lexical_cast.hpp>

#include <ros/param.h>



#define MULTIROBOT_MULTIPLEXER(NAME, T, V1, V2, V3, V4, V5) \
  inline T NAME(int id){ \
    switch(id){ \
      case 1: return V1; \
      case 2: return V2; \
      case 3: return V3; \
      case 4: return V4; \
      case 5: return V5; \
    } \
    abort(); \
    return T(); \
  } \
  inline T NAME() { \
    return NAME(WD::intId()); \
  }



namespace WD
{
  inline std::string platform()
  {
    std::string ret;
    if (! ros::param::getCached ("platform_hack", ret)) {
      ROS_ERROR ("Could not find parameter platform_hack");
      abort();
    }
    return ret;
  }
  
  inline std::string RobotNamePrefix()
  {
    return "OMNI";
  }
  
  inline int intId()
  {
    int ret;
    if (! ros::param::getCached ("robot_number", ret)) {
      ROS_ERROR ("Could not find parameter robot_number");
      abort();
    }
    return ret;
  }
  
  inline std::string stringId()
  {
    return RobotNamePrefix() + boost::lexical_cast<std::string> (intId());
  }
  
  inline std::string configPath()
  {
    std::string ret;
    if (! ros::param::get ("config_path", ret)) {
      ROS_ERROR ("Could not find parameter config_path");
      abort();
    }
    return ret;
  }
  
  inline std::string logPath()
  {
    return "/tmp";
  }
  
  inline int GKId()
  {
    return 1;
  }
  
  inline bool GoalIsNorth()
  {
    bool ret;
    if (! ros::param::getCached ("goal_is_north", ret)) {
      ROS_ERROR ("Could not find parameter goal_is_north");
      abort();
    }
    return ret;
  }
  
  inline bool PlayingMagenta()
  {
    bool ret;
    if (! ros::param::getCached ("playing_magenta", ret)) {
      ROS_ERROR ("Could not find parameter playing_magenta");
      abort();
    }
    return ret;
  }
  
  inline float BallRadius()
  {
    return 0.10;
  }
  inline float RobotRadius()
  {
    return 0.20;
  }
  inline float WheelRadius()
  {
    return 0.10;
  }
  
  inline float maxAccel()
  {
    return 2.7; // value verified by simulation - 3.841
  }
  
  inline float stoppingTimeCompensator()
  {
    return 0.8;
  }
  
  /// Maximum linear velocity of the robot (m/s) //NOTE: depends on the battery life!
  inline float MaximumSpeed()
  {
    return 2.0;
  }
  
  /// Maximum angular acceleration for each wheel (rad/s^2) //NOTE: up to 20 is safe.
  inline float MaximumAccelerationPerWheel()
  {
    return 20.0;
  }
  
  /// Time to take foul - units: ms
  inline int FoulTime()
  {
    return 10000;
  }
  /// Time to take foul almost over - units: ms
  inline int FoulTimeExpiring()
  {
    return 4000;
  }
  /// Time to wait requests - units: ms
  inline int WaitForRequestsTime()
  {
    return 7000;
  }
  /// Time to wait accepts - units: ms
  inline int WaitForAcceptsTime()
  {
    return 30000;
  }
  /// Time to wait confirm or reject - units: ms
  inline int WaitTimeBeforeQuit()
  {
    return 8000;
  }
  /// Time to wait "commited" - units: ms
  inline int WaitForCommitedTime()
  {
    return 5000;
  }
  
  /// Indicates if we want the robots to do a long pass during the game (for now it is best to be false so only short pass is done)
  inline bool WeWantLongPass()
  {
    return false;
  }
  
  /// This margin is used to divide the field in three parts, creating two areas near the goals and one area in the midfield
  inline float NearGoalMargin()
  {
    return 5;
  }
  inline float NearGoalMarginGK()
  {
    return 3;
  }
  
  inline float Distance2CornerX()
  {
    return 0.5;
  }
  inline float Distance2CornerY()
  {
    return 0.5;
  }
  
  /// The initial Postures
  MULTIROBOT_MULTIPLEXER (StartPosition, Posture, Posture ("(-6,0,0)"), Posture ("(-1.5,2.0,0)"), Posture ("(-1.5,-2.0,0)"), Posture ("(-3.0,1.0,0)"), Posture ("(-3.0,-1.0,0)"))
  /// The home position for the defender during the game
  inline Posture DefenderPosition()
  {
    return Posture ("(-4.0,0,0)");
  }
  
  // LRM Field
  // This should simply be cached getters to ROS parameters
  inline double FieldLength()
  {
    double ret;
    if (! ros::param::getCached ("field/field_length", ret)) {
      ROS_ERROR ("Could not find parameter field/field_length");
      abort();
    }
    return ret;
  }
  inline double FieldWidth()
  {
    double ret;
    if (! ros::param::getCached ("field/field_width", ret)) {
      ROS_ERROR ("Could not find parameter field/field_width");
      abort();
    }
    return ret;
  }
  inline double MarginWidth()
  {
    double ret;
    if (! ros::param::getCached ("field/margin_width", ret)) {
      ROS_ERROR ("Could not find parameter field/margin_width");
      abort();
    }
    return ret;
  }
  inline double CircleRadius()
  {
    double ret;
    if (! ros::param::getCached ("field/circle_radius", ret)) {
      ROS_ERROR ("Could not find parameter field/circle_radius");
      abort();
    }
    return ret;
  }
  inline double CornerRadius()
  {
    double ret;
    if (! ros::param::getCached ("field/corner_radius", ret)) {
      ROS_ERROR ("Could not find parameter field/corner_radius");
      abort();
    }
    return ret;
  }
  inline double GoalWidth()
  {
    double ret;
    if (! ros::param::getCached ("field/goal_width", ret)) {
      ROS_ERROR ("Could not find parameter field/goal_width");
      abort();
    }
    return ret;
  }
  inline double GoalDepth()
  {
    double ret;
    if (! ros::param::getCached ("field/goal_depth", ret)) {
      ROS_ERROR ("Could not find parameter field/goal_depth");
      abort();
    }
    return ret;
  }
  inline double GoalAreaWidth()
  {
    double ret;
    if (! ros::param::getCached ("field/goal_area_width", ret)) {
      ROS_ERROR ("Could not find parameter field/goal_area_width");
      abort();
    }
    return ret;
  }
  inline double GoalAreaDepth()
  {
    double ret;
    if (! ros::param::getCached ("field/goal_area_depth", ret)) {
      ROS_ERROR ("Could not find parameter field/goal_area_depth");
      abort();
    }
    return ret;
  }
  inline double PenaltyAreaWidth()
  {
    double ret;
    if (! ros::param::getCached ("field/penalty_area_width", ret)) {
      ROS_ERROR ("Could not find parameter field/penalty_area_width");
      abort();
    }
    return ret;
  }
  inline double PenaltyAreaDepth()
  {
    double ret;
    if (! ros::param::getCached ("field/penalty_area_depth", ret)) {
      ROS_ERROR ("Could not find parameter field/penalty_area_depth");
      abort();
    }
    return ret;
  }
  inline double PenaltyMarkDistance()
  {
    double ret;
    if (! ros::param::getCached ("field/penalty_mark_distance", ret)) {
      ROS_ERROR ("Could not find parameter field/penalty_mark_distance");
      abort();
    }
    return ret;
  }
  inline double FieldLineThickness()
  {
    double ret;
    if (! ros::param::getCached ("field/field_line_thickness", ret)) {
      ROS_ERROR ("Could not find parameter field/field_line_thickness");
      abort();
    }
    return ret;
  }
  inline float FieldHeading()
  {
    double ret;
    if (! ros::param::getCached ("field/field_heading", ret)) {
      ROS_ERROR ("Could not find parameter field/field_heading");
      abort();
    }
    return ret;
  }
  inline float FieldHeadingStdev()
  {
    double ret;
    if (! ros::param::getCached ("field/field_heading_stdev", ret)) {
      ROS_ERROR ("Could not find parameter field/field_heading_stdev");
      abort();
    }
    return ret;
  }
  
  /********** Calibration **********/
  
  // Be careful with the direction of the second axis.
  // While X (or column) is always from left to right in our cases,
  // the Y (or row) is different in several situations and confusing.
  // Images are stored in memory as tables, so the top row is 0 and grows
  // towards the bottom. However, in trignometric functions the Y axis
  // points up. To make tings more confusing, in our case, in some
  // places the Y points down.
  // Pixel values here are table rows and columns, directly from GIMP.
  
  /// Usable image circle radius in pixels, relative to camera center.
  MULTIROBOT_MULTIPLEXER (SearchableRadius, double, 520, 520, 520, 520, 520)
  
  /// Field Of View constant.
  /// Distance from center to end of image divided by 2, for our projection model.
  MULTIROBOT_MULTIPLEXER (FOVConstant, double, 256, 258, 258, 258, 258)
  
  /// Height of the camera in meters.
  /// Since the camera has 180 degrees FOV, this is the height of the last
  /// visible pixel in every direction.
  MULTIROBOT_MULTIPLEXER (CameraHeight, double, 0.657, 0.657, 0.657, 0.657, 0.653)
  
  /// Camera center of the image in pixels.
  /// Where the origin for the camera transformation resides.
  // TODO For now, center of the image will do. Must be carefully finetuned.
  // OCamCalib is useless since it uses a different projection model
  MULTIROBOT_MULTIPLEXER (CameraCenterImgCol, double, 542.0, 529.0, 532.0, 530.0, 530.0)
  MULTIROBOT_MULTIPLEXER (CameraCenterImgRow, double, 536.0, 522.0, 525.0, 526.0, 536.0)
  
  /// Center of the robot in the image in pixels.
  /// Where the origin of the robot frame resides.
  MULTIROBOT_MULTIPLEXER (RobotCenterImgCol, double, 556.5, 546.0, 552.5, 547.5, 547.0)
  MULTIROBOT_MULTIPLEXER (RobotCenterImgRow, double, 560.0, 550.0, 556.5, 556.5, 567.0)
  
  /// Angles of the camera supports in radians, relative to the camera center.
  MULTIROBOT_MULTIPLEXER (Support1Min, double, 1.12, 1.13, 1.11, 1.12, 1.11)
  MULTIROBOT_MULTIPLEXER (Support1Max, double, 1.28, 1.25, 1.25, 1.22, 1.28)
  MULTIROBOT_MULTIPLEXER (Support2Min, double, 2.90, 2.92, 2.95, 2.89, 2.90)
  MULTIROBOT_MULTIPLEXER (Support2Max, double, 3.05, 3.06, 3.13, 3.08, 3.08)
  MULTIROBOT_MULTIPLEXER (Support3Min, double, 5.15, 5.18, 5.17, 5.18, 5.16)
  MULTIROBOT_MULTIPLEXER (Support3Max, double, 5.35, 5.29, 5.40, 5.33, 5.37)
}

#endif
