#include <cstdio>
#include <string>
#include <sstream>

#include <unistd.h>
#include <termios.h>

#include <ros/ros.h>

#include <socrob/math.h>
using namespace socrob;

#include <socrob_ros/ControlInput.h>
#include <std_msgs/Float32.h>



using namespace std;



const double INCR_LINEAR = 0.125;
const double INCR_ANGULAR = 0.3125;



void noisyPrint (string s)
{
  // Probably because of the terminal changes, we cannot use cout here.
  printf ("\n\r"
          "\n\r"
          "-------------------------------------------------------\n\r"
          "%s\n\r"
          "-------------------------------------------------------\n\r"
          "\n\r", s.c_str());
  fflush (stdout);
}



void
omniManual()
{
  ros::NodeHandle nh;
  ros::Publisher control_pub = nh.advertise<socrob_ros::ControlInput> ("control_input_manual", 1, true);
  ros::Publisher kick_pub = nh.advertise<std_msgs::Float32> ("kick", 1, false);
  
  socrob_ros::ControlInput control;
  control.type = socrob_ros::ControlInput::VELOCITY;
  control.priority = socrob_ros::ControlInput::BALANCED;
  control.linear.vX = 0;
  control.linear.vY = 0;
  control.angular = 0;
  
  std_msgs::Float32 kick;
  kick.data = 0;
  
  ros::spinOnce();
  noisyPrint ("COMMAND: STOP!");
  control.linear.vX = 0;
  control.linear.vY = 0;
  control.angular = 0;
  control_pub.publish (control);
  
  while (ros::ok()) {
    ros::spinOnce();
    
    int command = tolower (getchar_unlocked());
    switch (command) {
      case 'i': {
        noisyPrint ("COMMAND: FRONT!");
        control.linear.vX += INCR_LINEAR;
        break;
      }
      case 'k': {
        noisyPrint ("COMMAND: BACK!");
        control.linear.vX -= INCR_LINEAR;
        break;
      }
      case 'j': {
        noisyPrint ("COMMAND: LEFT!");
        control.linear.vY += INCR_LINEAR;
        break;
      }
      case 'l': {
        noisyPrint ("COMMAND: RIGHT!");
        control.linear.vY -= INCR_LINEAR;
        break;
      }
      case 'q': {
        noisyPrint ("COMMAND: ROTATE LEFT!");
        control.angular += INCR_ANGULAR;
        break;
      }
      case 'e': {
        noisyPrint ("COMMAND: ROTATE RIGHT!");
        control.angular -= INCR_ANGULAR;
        break;
      }
      case 'w': {
        noisyPrint ("COMMAND: STOP ROTATE!");
        control.angular = 0;
        break;
      }
      case ' ': {
        noisyPrint ("COMMAND: STOP!");
        control.linear.vX = 0;
        control.linear.vY = 0;
        control.angular = 0;
        break;
      }
      case 'a': {
        noisyPrint ("COMMAND: KICK 25%!");
        kick.data = 25;
        kick_pub.publish (kick);
        break;
      }
      case 's': {
        noisyPrint ("COMMAND: KICK 50%!");
        kick.data = 50;
        kick_pub.publish (kick);
        break;
      }
      case 'd': {
        noisyPrint ("COMMAND: KICK 75%!");
        kick.data = 75;
        kick_pub.publish (kick);
        break;
      }
      case 'f': {
        noisyPrint ("COMMAND: KICK 100%!");
        kick.data = 100;
        kick_pub.publish (kick);
        break;
      }
      case 'x': {
        noisyPrint ("COMMAND: QUIT!");
        control.linear.vX = 0;
        control.linear.vY = 0;
        control.angular = 0;
        control_pub.publish (control);
        return;
      }
      case 'h': {
        noisyPrint ("HELP:\n\r"
                    " -> Move: Forward ......... Key: i\n\r"
                    " -> Move: Backward ........ Key: k\n\r"
                    " -> Move: Left ............ Key: j\n\r"
                    " -> Move: Right ........... Key: l\n\r"
                    " -> Rotate: Left .......... Key: q\n\r"
                    " -> Rotate: Right ......... Key: e\n\r"
                    " -> Rotate: Stop Rotation . Key: w\n\r"
                    " -> Stop all movement ..... Key: Space\n\r"
                    " -> Kicker: Kick 25% ...... Key: a\n\r"
                    " -> Kicker: Kick 50% ...... Key: s\n\r"
                    " -> Kicker: Kick 75% ...... Key: d\n\r"
                    " -> Kicker: Kick 100% ..... Key: f\n\r"
                    " -> Exit .................. Key: x\n\r"
                    " -> Help .................. Key: h");
                    
        break;
      }
      default: {
        noisyPrint ("UNKNOWN COMMAND");
        break;
      }
    }
    
    control_pub.publish (control);
    
    {
      stringstream temp;
      temp << "Velocity: (Vx,Vy,Th) = ( "
           << control.linear.vX << " , "
           << control.linear.vY << " , "
           << control.angular << " )";
      noisyPrint (temp.str());
    }
  }
}



int
main (int argc,
      char** argv)
{
  ros::init (argc, argv, "omni_manual");
  
  std::cout << "SocRob OmniManual" << endl;
  
  struct termios old_tios;
  tcgetattr (STDIN_FILENO, &old_tios);
  
  {
    struct termios new_tios;
    tcgetattr (STDIN_FILENO, &new_tios);
    cfmakeraw (&new_tios);
    tcsetattr (STDIN_FILENO, TCSAFLUSH, &new_tios);
  }
  
  omniManual();
  
  tcsetattr (STDIN_FILENO, TCSAFLUSH, &old_tios);
  
  return 0;
}
