/**
 * @Filename ControlInput.cpp
 * @Description  ControlInput class implementation
 * @Status Finished
 * @Version $Id: ControlInput.cpp 5864 2012-12-08 14:40:54Z jreis $
 * @Maintainer João Messias
 */

#include "ControlInput.hpp"
#include <math.h>
#include <socrob/math.h>

#include <sstream>

#include <LocalFrameWorldFrameTransformations.hpp>



ControlInput::ControlInput() : Data(),
  type_ (VELOCITY), priority_ (BALANCED), linear_(), angular_ (0.0), uncertainty_ (0.0), trustAngular_ (1.0)
{
}



ControlInput::ControlInput (string dataString) : Data(),
  type_ (VELOCITY), priority_ (BALANCED), linear_(), angular_ (0.0), uncertainty_ (0.0), trustAngular_ (1.0)
{
  const char* data;
  data = dataString.c_str();
  float inMagnitude = 0.0, inAzimuth = 0.0, inAngular = 0.0;
  
  sscanf (data, "(%f,%f,%f)", &inMagnitude, &inAzimuth, &inAngular);
  
  setPolar (inMagnitude, inAzimuth, inAngular);
}



ControlInput::ControlInput (float magnitude, float azimuth, float angular,
                            bool isRect, CONTROL_TYPE type, CONTROL_PRIORITY priority) : Data(),
  type_ (type), priority_ (priority), linear_ (magnitude, azimuth, isRect), angular_ (angular), uncertainty_ (0.0),
  trustAngular_ (1.0)
{
}



ControlInput::ControlInput (CONTROL_TYPE type) : Data(),
  type_ (type), priority_ (BALANCED), linear_ (0.0, 0.0), angular_ (0.0), uncertainty_ (0.0),
  trustAngular_ (1.0)
{
}



ControlInput::~ControlInput()
{
}



Speed ControlInput::getLinearX() const
{
  return linear_.getVx();
}



Speed ControlInput::getLinearY() const
{
  return linear_.getVy();
}



Speed ControlInput::getMagnitude() const
{
  return linear_.getSpeed();
}



Angle ControlInput::getAzimuth() const
{
  return linear_.getAzimuth();
}



Rotation_Speed ControlInput::getAngular() const
{
  return angular_;
}



void ControlInput::setPolar (float magnitude, Angle azimuth,
                             float angular)
{
  setPolar (magnitude, azimuth);
  setAngular (angular);
}



void ControlInput::setPolar (float magnitude, Angle azimuth)
{
  linear_.setPolar (magnitude, azimuth);
}



void ControlInput::setMagnitude (float magnitude)
{
  linear_.setSpeed (magnitude);
}



void ControlInput::setAzimuth (Angle azimuth)
{
  linear_.setAzimuth (azimuth);
}



void ControlInput::setLinearXLinearYAngular (float linearx, float lineary,
    float angular)
{
  setLinearXLinearY (linearx, lineary);
  setAngular (angular);
}



void ControlInput::setLinearXLinearY (float linearx, float lineary)
{
  linear_.setVxVy (linearx, lineary);
}



void ControlInput::setLinearX (float linearx)
{
  linear_.setVx (linearx);
}



void ControlInput::setLinearY (float lineary)
{
  linear_.setVy (lineary);
}



void ControlInput::setAngular (float angular)
{
  angular_ = angular;
}



void ControlInput::setUncertainty (float value)
{
  uncertainty_ = value;
}



void ControlInput::setTrust (float trustLinearX, float trustLinearY, float trustAngular)
{
  linear_.setTrust (trustLinearX, trustLinearY);
  trustAngular_ = trustAngular;
}



void ControlInput::setType (CONTROL_TYPE type)
{
  type_ = type;
}



void ControlInput::setPriority (CONTROL_PRIORITY priority)
{
  priority_ = priority;
}



float ControlInput::getTrustLinearX() const
{
  return linear_.getTrustVx();
}



float ControlInput::getTrustLinearY() const
{
  return linear_.getTrustVy();
}



float ControlInput::getTrustAngular() const
{
  return trustAngular_;
}



float ControlInput::getUncertainty() const
{
  return uncertainty_;
}



CONTROL_TYPE ControlInput::getType() const
{
  return type_;
}



CONTROL_PRIORITY ControlInput::getPriority() const
{
  return priority_;
}



shared_ptr<Parameter> ControlInput::clone() const
{
  shared_ptr<ControlInput> newControlInput (new ControlInput (*this));
  return newControlInput;
}



string ControlInput::toString() const
{
  stringstream ret;
  ret << "(" << getMagnitude() << ","
      << getAzimuth() << ","
      << getAngular() << ")";
      
  return ret.str();
}



//eof
