/**
 * @Filename ControlInput.hpp
 * @Description ControlInput class header.
 * @Status Finished
 * @Version $Id: ControlInput.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer João Messias
 */

#ifndef _CONTROL_INPUT_HPP_
#define _CONTROL_INPUT_HPP_

#include <Data.hpp>
#include <Parameter.hpp>

#include <Defines.hpp>
#include <Velocity2D.hpp>

#include <socrob_ros/ControlInput.h>



/**
 * @Class ControlInput ControlInput.hpp "ControlInput.hpp"
 * This class defines a control input that may be sent to the robot's actuators. It may represent either an acceleration
 * control or a velocity control. It always possesses a 2D linear component and an angular component.
 * @Author João Messias
 */
class ControlInput : public Data
{
  public:
    /**
     * Class base constructor (everything is set to 0)
     */
    ControlInput();
    
    /**
     * Create the class from a string.
     * @Argument dataString - String with data in the format "(%f, %f, %f)",
     * with the linear control's magnitude and azimuth (in polar coordinates), followed by the angular control.
     */
    ControlInput (string dataString);
    
    /**
     * Create a new ControlInput with the given values in SI coordinates.
     * @Argument magnitude - rho value (or x if rectangular coordinates)
     * @Argument azimuth - angle value (or y if rectangular coordinates)
     * @Argument angular - Angular control
     * @Argument isRect - (optional) if false (default), then the arguments
     * refer to polar coordinates, otherwise they refer to rectangular
     * coordinates.
     */
    ControlInput (float magnitude, float azimuth, float angular,
                  bool isRect = false, CONTROL_TYPE type = VELOCITY, CONTROL_PRIORITY = BALANCED);
                  
    /**
     * Creates a ControlInput with the specified type, and every other field is initialized to default values.
     */
    ControlInput (CONTROL_TYPE type);
    /**
     * Destructor
     */
    ~ControlInput();
    
    /**
     * Retrieve the control input X value (rectangular coordinates)
     */
    float getLinearX() const;
    
    /**
     * Retrieve the control input Y value (rectangular coordinates)
     */
    Speed getLinearY() const;
    
    /**
     * Retreive the control input rho value (polar coordinates)
     */
    Speed getMagnitude() const;
    
    /**
     * Retrieve the azimuth angle (polar coordinates)
     */
    Angle getAzimuth() const;
    
    /**
     * Retrieve the angular control input
     */
    float getAngular() const;
    
    /**
     * Change the X,Y, Theta control values based on the given polar
     * coordinates.
     * @Argument magnitude - the linear control rho value in polar coordinates.
     * @Argument azimuth - the linear control angle in polar coordinate.
     * @Argument angular - new angular control value.
     */
    void setPolar (float magnitude, Angle azimuth, float angular);
    
    /**
     * Change the X,Y linear control values based on the given polar coordinates.
     * @Argument magnitude - the linear control rho value in polar coordinates.
     * @Argument azimuth - the linear control angle in polar coordinate.
     */
    void setPolar (float magnitude, Angle azimuth);
    
    /**
     * Change the linear control's magnitude.
     * @Argument magnitude - the rho value in polar coordinates.
     */
    void setMagnitude (float magnitude);
    
    /**
     * Change the linear control's azimuth.
     * @Argument azimuth - the new angle polar coordinate.
     */
    void setAzimuth (Angle azimuth);
    
    /**
     * Change the X,Y and Theta linear control values.
     * @Argument linearx - The new X linear control value.
     * @Argument lineary - The new Y linear control value.
     * @Argument angular - The new angular control value.
     */
    void setLinearXLinearYAngular (float linearx, float lineary,
                                   float angular);
                                   
    /**
     * Change the X,Y linear control values (cartesian coordinates).
     * @Argument linearx - The new X linear control value.
     * @Argument lineary - The new Y linear control value.
     */
    void setLinearXLinearY (float linearx, float lineary);
    
    /**
     * Change the X linear control value (cartesian coordinates).
     * @Argument vx - The new X linear control value.
     */
    void setLinearX (float linearx);
    
    /**
     * Change the Y linear control value (rectangulaar coordinates).
     * @Argument vy - The new Y linear control value.
     */
    void setLinearY (float lineary);
    
    /**
     * Change the angular control value.
     * @Argument angular - The angular control value.
     */
    void setAngular (float angular);
    
    /**
     * Retrieve the uncertainty
     */
    float getUncertainty() const;
    
    /**
     * Retrieve the X value certainty.
     */
    float getTrustLinearX() const;
    
    /**
     * Retrieve the Y value certainty.
     */
    float getTrustLinearY() const;
    
    /**
     * Retrieve the Theta value certainty.
     */
    float getTrustAngular() const;
    
    /**
     * Change the uncertainty associated with this vector. 0 means no
     * uncertainty, while 1 means full uncertainty.
     * @Argument value - new uncertainty value.
     */
    void setUncertainty (float value);
    
    /**
     * Change the cercainty associated with each component. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument trustLinearX - certainty of the X component.
     * @Argument trustLinearY - certainty of the Y component.
     * @Argument trustAngular - certainty of the Theta component.
     */
    void setTrust (float trustLinearX, float trustLinearY,
                   float trustAngular);
                   
    /**
     * Retrieve the control type.
     */
    CONTROL_TYPE getType() const;
    
    /**
     * Retrieve the control priority
     */
    CONTROL_PRIORITY getPriority() const;
    
    /**
      * Change the control type.
      * @Argument type - the desired control type (VELOCITY or ACCELERATION).
      */
    void setType (CONTROL_TYPE type);
    
    /**
      * Change the control priority.
      * @Argument priority - the desired control priority (BALANCED, LINEAR or ANGULAR).
      */
    void setPriority (CONTROL_PRIORITY priority);
    
    shared_ptr<Parameter> clone() const;
    
    /**
     * Returns the control input in string form (magnitude, azimuth, angular control).
     * @Returns the string with the control input.
     */
    string toString() const;
    
    void rosMsg (socrob_ros::ControlInput& msg) const {
      switch (type_) {
        case VELOCITY:
          msg.type = socrob_ros::ControlInput::VELOCITY;
          break;
        case ACCELERATION:
          msg.type = socrob_ros::ControlInput::ACCELERATION;
          break;
      }
      switch (priority_) {
        case BALANCED:
          msg.priority = socrob_ros::ControlInput::BALANCED;
          break;
        case LINEAR:
          msg.priority = socrob_ros::ControlInput::LINEAR;
          break;
        case ANGULAR:
          msg.priority = socrob_ros::ControlInput::ANGULAR;
          break;
      }
      linear_.rosMsg (msg.linear);
      msg.angular = angular_;
      msg.uncertainty = uncertainty_;
      msg.trustAngular = trustAngular_;
    }
    
    void fromRosMsg (const socrob_ros::ControlInput& msg) {
      switch (msg.type) {
        case socrob_ros::ControlInput::VELOCITY:
          type_ = VELOCITY;
          break;
        case socrob_ros::ControlInput::ACCELERATION:
          type_ = ACCELERATION;
          break;
      }
      switch (msg.priority) {
        case socrob_ros::ControlInput::BALANCED:
          priority_ = BALANCED;
          break;
        case socrob_ros::ControlInput::LINEAR:
          priority_ = LINEAR;
          break;
        case socrob_ros::ControlInput::ANGULAR:
          priority_ = ANGULAR;
          break;
      }
      linear_.fromRosMsg (msg.linear);
      angular_ = msg.angular;
      uncertainty_ = msg.uncertainty;
      trustAngular_ = msg.trustAngular;
    }
    
    static shared_ptr<ControlInput> newFromRosMsg (const socrob_ros::ControlInput& msg) {
      shared_ptr<ControlInput> tmp = shared_ptr<ControlInput> (new ControlInput ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
    
  private:
    CONTROL_TYPE type_; ///< the type of control, may be either VELOCITY or ACCELERATION.
    CONTROL_PRIORITY priority_; ///< Priority to robot velocity component (BALANCED, LINEAR, ANGULAR)
    Velocity2D linear_; ///< Linear control. In practice it may represent either a velocity or an acceleration.
    float angular_; ///< Angular control
    float uncertainty_; ///< Uncertainty associated with this object
    float trustAngular_; ///< Certainty of the angular control
};

#endif
