/**
* @Filename Data.hpp
* Data class header.
* @Status Implementing
* @Version $Id: Data.hpp 5750 2012-07-05 01:50:42Z jreis $
* @Maintainer Hugo Veiga (hasv@mega.ist.utl.pt)
*/

#ifndef _DATA_HPP_
#define _DATA_HPP_

#include <string>

#include "Parameter.hpp"
#include "TimeStamp.hpp"
#include <SocRobXMLNode.hpp>



/**
* @Class Data Data.hpp "Data.hpp"
* Class Data which represents all data that will be storable.
* @Author Hugo Veiga
*/
class Data : public Parameter
{
  public:
    /**
     * Constructor.
     * This method creates a new, empty Data.
     * The Data's timestamp is set to the the time of its creation.
     */
    Data();
    
    /**
     * Constructor.
     * This method creates a new, empty Data.
     * The Data's timestamp is defined by the given timestamp.
     * @param t The timestamp to be given to the data.
     */
    Data (TimeStamp t);
    
    
    /**
     * Destructor.
     * This method makes deinitializations for objects of this class.
     */
    virtual ~Data();
    
    /**
     * This method creates a new copy of itself and returns that copy. Note that.
     * if you wish to copy all the fields, then you should use this method.
     * Creating a copy of a Data object with a method, other than the clone()
     * one, does not guarantee that all the fields are copied. This method
     * also copies the timeStamp, so that the cloned object, and the clone
     * object, are in fact clones.
     */
    virtual shared_ptr<Parameter> clone() const = 0;
    
    virtual string toString() const = 0;
    
    /**
     * Set if this data is differential or not. Differential means that the
     * values are realtive values, so they only make sense if fully processed,
     * i.e., do sample should be discarded.
     * @Argument value - the boolean value to set.
     */
    void setDifferentialDataType (bool value);
    
    bool isDifferentialDataType();
    
    
    //this method exists and it job is not done in a Data constructor because It is necessary to return an error when xml has errors
    static shared_ptr<Data> buildData (SocRobXMLNode* dataNode, string& name);
    
    //     static void diffTime(Data* data1, Data* data2, float result);
    
    //! @ TODO move this to protected data
    TimeStamp timeStamp; ///< Time stamp used to control the data times.
    
    inline TimeStamp getTimeStamp() const {
      return timeStamp;
    }
    inline void setTimeStamp (TimeStamp t) {
      timeStamp = t;
    }
    
  protected:
  
    /**
     * Overload the copy constructor, so that we control the copies are made
     */
    Data (const Data& oldData);
    
    /**
     * Overload the = operator, so that we control how copies are made
     */
    Data& operator= (const Data& oldData);
    
  private:
    /// If differentialDataType is true, then it means that every value we
    /// obtain should be processed, i.e., if we receive 20 samples over a given
    /// period of time, then we should process those 20 samples.
    /// Note that this value is allways false, unless otherwise specified.
    bool isDifferentialDataType_;
    string name_;
};

#endif
