/**
* @Filename DataArray.cpp
* DataArray class implementation.
* @Status Implementing
* @Version $Id: DataArray.cpp 5661 2012-05-11 13:43:55Z jreis $
* @Maintainer Hugo Veiga (hasv@mega.ist.utl.pt)
*/

#include "DataArray.hpp"

#include <String.hpp>



DataArray::DataArray() :
  dataArray_(), actualPosition_ (-1), numMaxElements_ (-1), size_ (0),
  dynamicArray_ (true)
{
}



DataArray::DataArray (int numMaxElements) :
  dataArray_ (std::vector<shared_ptr<Data> > (numMaxElements)), actualPosition_ (-1),
  numMaxElements_ (numMaxElements), size_ (0), dynamicArray_ (false)
{
}



DataArray::~DataArray()
{
  dataArray_.clear();
}



void DataArray::addElement (shared_ptr<Data> data)
{
  //dataArray_.push_back(timedData); //insert int into the buffer
  
  if (dynamicArray_) {
    dataArray_.push_back (data);
    size_++;
  }
  else {
    actualPosition_ = (actualPosition_ + 1) % numMaxElements_;
    dataArray_[actualPosition_] = data;
    size_++;
    if (size_ > numMaxElements_) {
      size_ = numMaxElements_;
    }
  }
}



shared_ptr<Data> DataArray::getElement (int index)
{
  shared_ptr<Data> result;
  
  if (dynamicArray_) {
    result = dataArray_[dataArray_.size() - index - 1];
  }
  else {
    if ( (index < 0) || (index >= numMaxElements_) || (actualPosition_ == -1)) {
      return shared_ptr<Data>();
    }
    int relativeIndex = actualPosition_ - index;
    if (relativeIndex < 0) {
      relativeIndex = relativeIndex + numMaxElements_;
    }
    result = dataArray_[relativeIndex];
  }
  return result;
}



shared_ptr<Data const> DataArray::getElement (int index) const
{
  shared_ptr<Data const> result;
  
  if (dynamicArray_) {
    result = dataArray_[dataArray_.size() - index - 1];
  }
  else {
    if ( (index < 0) || (index >= numMaxElements_) || (actualPosition_ == -1)) {
      return shared_ptr<Data>();
    }
    int relativeIndex = actualPosition_ - index;
    if (relativeIndex < 0) {
      relativeIndex = relativeIndex + numMaxElements_;
    }
    result = dataArray_[relativeIndex];
  }
  return result;
}



shared_ptr<Data> DataArray::getMostRecentElement()
{
  return getElement (0);
}



shared_ptr<Parameter> DataArray::clone() const
{
  shared_ptr<DataArray> ret;
  int i;
  
  if (dynamicArray_) {
    ret = shared_ptr<DataArray> (new DataArray());
    for (i = 0; i < (int) dataArray_.size(); i++) {
      shared_ptr<Data const> dataTmp = getElement (i);
      ret->addElement (dynamic_pointer_cast<Data> (dataTmp->clone()));
    }
  }
  else {
    ret = shared_ptr<DataArray> (new DataArray (numMaxElements_));
    
    int lastPosition = actualPosition_ - size_;
    if (lastPosition < 0) {
      lastPosition = lastPosition + numMaxElements_ + 1;
    }
    for (i = 0; i < size_; i++) {
      shared_ptr<Data const> dataTmp = getElement ( (lastPosition + i) % numMaxElements_);
      ret->addElement (dynamic_pointer_cast<Data> (dataTmp->clone()));
    }
  }
  
  return ret;
}



string DataArray::toString() const
{
  ostringstream s;
  s << "[";
  bool isFirst = true;
  for (int i = 0; i < size_; i++) {
    if (!isFirst) {
      s << ",";
    }
    shared_ptr<Data const> dataTmp = getElement (i);
    s << " " << dataTmp->toString();
    isFirst = false;
  }
  s << " ]";
  return s.str();
  
#if 0
  string ret = "";
  if (size_ > 0) {
    shared_ptr<Data const> dataTmp = getElement (0);
    ret += dataTmp->toString();
  }
  for (int i = 1; i < size_; i++) {
    shared_ptr<Data const> dataTmp = getElement (i);
    ret += ":" + dataTmp->toString();
  }
  return ret;
#endif
}



int DataArray::getNumMaxElements()
{
  return numMaxElements_;
}



void DataArray::incNumMaxElements (int numOfElements2Add)
{
  if (numMaxElements_ == size_) {
    actualPosition_ = size_ - 1; //this don't guarantee the temporal order, so be carefull when you use this function.
  }
  numMaxElements_ += numOfElements2Add;
}



int DataArray::getSize() const
{
  int sizeOfArray;
  
  if (dynamicArray_) {
    sizeOfArray = dataArray_.size();
  }
  else {
    sizeOfArray = size_;
  }
  
  return sizeOfArray;
}



shared_ptr<DataArray>
DataArray::
from (const set<string>& s)
{
  shared_ptr<DataArray> array (new DataArray ());
  
  foreach (string p, s) {
    array->addElement (shared_ptr<String> (new String (p)));
  }
  
  return array;
}
