/**
 * @Filename FullPosture.hpp
 * @Description Vector Posture class header.
 * @Status Finished
 * @Version $Id: FullPosture.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _FULL_POSTURE_
#define _FULL_POSTURE_

#include <Data.hpp>
#include <Posture.hpp>
#include <VelocityPosture.hpp>

#include <socrob_ros/FullPosture.h>



/**
 * @Class FullPosture FullPosture.hpp "FullPosture.hpp"
 * This class is used simply to aggregate all the information regarding the
 * (robot) posture. This could be more elaborate, but time is too short.
 * @Author Hugo Costelha
 */
class FullPosture : public Data
{
  public:
    /**
     * Constructor
     * Set everything to 0.
     */
    FullPosture();
    
    /**
     * Constructor
     * @Argument newPosture - posture in world coordinates
     * @Argument newVelocity - velocity in local or world coordinates;
     * @Argument newVelocityInLocalCoordinates - (optional, true by default) if
     * true, then the velocity is given in local coordinates, otherwise it is
     * given in world coordinates.
     * @Argument newPosture - (optional, 0 by default) posture in local or world
     */
    FullPosture (Posture newPosture,
                 VelocityPosture newVelocity,
                 bool newVelocityInLocalCoordinates = true);
                 
    FullPosture (string dataString);
    
    shared_ptr<Parameter> clone() const;
    string toString() const;
    
    Posture worldPosture; ///< Robot posture in world coordinates.
    VelocityPosture localVelocity; ///< Robot velocity in local coordinates.
    VelocityPosture worldVelocity; ///< Robot velocity in world coordinates.
    bool isLocalized; ///< True if the uncertainty associated with the base
    ///< posture is low enough, false otherwise.
    
    void rosMsg (socrob_ros::FullPosture& msg) const {
      worldPosture.rosMsg (msg.worldPosture);
      localVelocity.rosMsg (msg.localVelocity);
      worldVelocity.rosMsg (msg.worldVelocity);
      msg.isLocalized = isLocalized;
    }
    
    void fromRosMsg (const socrob_ros::FullPosture& msg) {
      worldPosture.fromRosMsg (msg.worldPosture);
      localVelocity.fromRosMsg (msg.localVelocity);
      worldVelocity.fromRosMsg (msg.worldVelocity);
      isLocalized = msg.isLocalized;
    }
    
    static shared_ptr<FullPosture> newFromRosMsg (const socrob_ros::FullPosture& msg) {
      shared_ptr<FullPosture> tmp = shared_ptr<FullPosture> (new FullPosture ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
};

#endif
