/**
* @Filename GaussianMixture.cpp
* GaussianMixture Data class implementation.
* @Status Implementing
* @Version $Id: GaussianMixture.cpp
* @Maintainer Joao Santos (jsantos@isr.ist.utl.pt)
*/

#include <GaussianMixture.hpp>

#include <sstream>



GaussianMixture::
GaussianMixture (void) :
  Data(),
  nModes_ (0),
  GmmMean_ (0),
  GmmWeight_ (0),
  GmmCov_ (0),
  likelihood_ (0)
{
}



GaussianMixture::
GaussianMixture (int nModes) :
  Data(),
  nModes_ (nModes),
  GmmMean_ ( (float*) calloc (nModes * 2 , sizeof (float))),
  GmmWeight_ ( (float*) calloc (nModes , sizeof (float))),
  GmmCov_ ( (float*) calloc (nModes * 2 * 2 , sizeof (float))),
  likelihood_ (0)
{
}



GaussianMixture::
GaussianMixture (string dataString) :
  Data(),
  nModes_(),
  GmmMean_(),
  GmmWeight_(),
  GmmCov_(),
  likelihood_ (0)
{
  stringstream data;
  data << dataString;
  
  data >> nModes_;
  
  GmmMean_ = (float*) calloc (nModes_ * 2 , sizeof (float));
  GmmWeight_ = (float*) calloc (nModes_ , sizeof (float));
  GmmCov_ = (float*) calloc (nModes_ * 2 * 2 , sizeof (float));
}



GaussianMixture::
~GaussianMixture (void)
{
  free (GmmMean_);
  free (GmmWeight_);
  free (GmmCov_);
}



void
GaussianMixture::
setMean (int mode,
         float mu_x,
         float mu_y)
{
  GmmMean_[mode * 2] = mu_x;
  GmmMean_[mode * 2 + 1] = mu_y;
}



void
GaussianMixture::
setWeight (int mode,
           float weight)
{
  GmmWeight_[mode] = weight;
}



void
GaussianMixture::
setCov (int mode,
        float xx,
        float xy,
        float yx,
        float yy)
{
  GmmCov_[mode * 2 * 2] = xx;
  GmmCov_[mode * 2 * 2 + 1] = xy;
  GmmCov_[mode * 2 * 2 + 2] = yx;
  GmmCov_[mode * 2 * 2 + 3] = yy;
}



void
GaussianMixture::
setLikelihood (float value)
{
  likelihood_ = value;
}



void
GaussianMixture::
getMean (int mode,
         float* mu_x,
         float* mu_y)
{
  *mu_x = GmmMean_[mode * 2];
  *mu_y = GmmMean_[mode * 2 + 1];
}



float
GaussianMixture::
getWeight (int mode)
{
  return GmmWeight_[mode];
}



void
GaussianMixture::
getCov (int mode,
        float* xx,
        float* xy,
        float* yx,
        float* yy)
{
  *xx = GmmCov_[mode * 2 * 2];
  *xy = GmmCov_[mode * 2 * 2 + 1];
  *yx = GmmCov_[mode * 2 * 2 + 2];
  *yy = GmmCov_[mode * 2 * 2 + 3];
}



float
GaussianMixture::
getLikelihood()
{
  return likelihood_;
}



int
GaussianMixture::
getNumberOfModes()
{
  return nModes_;
}



string
GaussianMixture::
toString()
{
  stringstream ret;
  
  int i;
  for (i = 0; i < nModes_; i++) {
    ret << i << "\n";
    ret << GmmWeight_[i] << "\n";
    ret << GmmMean_[i * 2] << " " << GmmMean_[i * 2 + 1] << "\n";
    ret << GmmCov_[i * 2 * 2] << " " << GmmCov_[i * 2 * 2 + 1] << " " << GmmCov_[i * 2 * 2 + 2] << " " << GmmCov_[i * 2 * 2 + 3] << "\n";
  }
  
  return ret.str();
}



shared_ptr<Parameter>
GaussianMixture::
clone() const
{
  // This is obsolete and not corretly implemeted. Do not use this method.
  abort();
  
  return shared_ptr<Parameter>();
}
