/**
* @Filename GaussianMixture.hpp
* GaussianMixture Data class header.
* @Status Implementing
* @Version $Id: GaussianMixture.hpp
* @Maintainer Joao Santos (jsantos@isr.ist.utl.pt)
*/

#ifndef _GAUSSIAN_MIXTURE_HPP_
#define _GAUSSIAN_MIXTURE_HPP_

#include "Data.hpp"



/**
* @Class GaussianMixture GaussianMixture.hpp "GaussianMixture.hpp"
* Class GaussianMixture Data implements a 2D Gaussian Mixture Model
* @Author Joao Santos
*/
class GaussianMixture : public Data
{
  public:
    /**
     * Constructor.
     * This method creates a new GaussianMixture.
     */
    GaussianMixture (void);
    
    /**
     * Constructor.
     * This method creates a new GaussianMixture with a specified number of components.
     */
    GaussianMixture (int nModes);
    
    /**
     * Constructor.
     * This method creates a new GaussianMixture with a specified number of components, given from a string value.
     */
    GaussianMixture (string dataString);
    
    /**
     * Destructor.
     * This method makes deinitializations for objects of this class.
     */
    ~GaussianMixture (void);
    
    /**
     * This method sets the mean values of the specified mode
     * @Argument mode - number of the component
     * @Argument mu_x - mean value x coordinate
     * @Argument mu_y - mean value y coordinate
     */
    void setMean (int mode, float mu_x, float mu_y);
    
    /**
     * This method sets the weight of the specified mode
     * @Argument mode - number of the component
     * @Argument weight - relative weight of the mode in the GMM
     */
    void setWeight (int mode, float weight);
    
    /**
     * This method sets the covariance matrix of the specified mode
     * @Argument mode - number of the component
     * @Argument xx - cov matrix elem xx
     * @Argument xy - cov matrix elem xy
     * @Argument xx - cov matrix elem yx
     * @Argument xy - cov matrix elem yy
     */
    void setCov (int mode, float xx, float xy, float yx, float yy);
    
    /**
     * This method sets the likelihood of the GMM
     * @Argument value - likelihood
     */
    void setLikelihood (float value);
    
    /**
     * This method gets the mean values of the specified mode
     * @Argument mode - number of the component
     * @Argument mu_x - returned mean value x coordinate
     * @Argument mu_y - returned mean value y coordinate
     */
    void getMean (int mode, float* mu_x, float* mu_y);
    
    /**
     * This method gets the weight of the specified mode
     * @Argument mode - number of the component
     * @Returns weight - relative weight of the mode in the GMM
     */
    float getWeight (int mode);
    
    /**
     * This method gets the covariance matrix of the specified mode
     * @Argument mode - number of the component
     * @Argument xx - returned cov matrix elem xx
     * @Argument xy - returned cov matrix elem xy
     * @Argument xx - returned cov matrix elem yx
     * @Argument xy - returned cov matrix elem yy
     */
    void getCov (int mode, float* xx, float* xy, float* yx, float* yy);
    
    /**
     * This method gets the likelihood of the GMM
     * @Returns likelihood of the GMM
     */
    float getLikelihood();
    
    /**
     * This method gets the number of components in the GMM
     * @Returns number of modes
     */
    int getNumberOfModes();
    
    
    /**
     * Converts the GMM info to a string
     * Should be used only for debug purposes!
     * @Returns all the info in one string
     */
    string toString();
    
    
    /**
     * This method create a new object that is a clone of this
     * @Returns clone Parameter
     */
    shared_ptr<Parameter> clone() const;
    
    
  private:
    int nModes_;  ///< Number of components
    float* GmmMean_; ///< vector array of the GMM means
    float* GmmWeight_; ///< vector array of the GMM weight
    float* GmmCov_; ///< vector array of the GMM covariance matrices
    float likelihood_;  ///< GMM likelihood
    
    // Since there are pointer members, copy must be forbidden.
    // If needed, create this two correctly according to the cookbook.
    GaussianMixture (const GaussianMixture&);
    GaussianMixture& operator= (const GaussianMixture&);
};
#endif



//eof
