/**
 * @Filename Position2D.cpp
 * Position2D Data class implementation.
 * @Status Implementing
 * @Version $Id: Position2D.cpp 5661 2012-05-11 13:43:55Z jreis $
 * @Maintainer Hugo Costelha
 */
#include <Position2D.hpp>
#include <FrameTransformations2D.hpp>

#include <sstream>



Position2D::Position2D (void) : Data(),
  x_ (0), y_ (0), rho_ (0), angle_ (0), trustX_ (1), trustY_ (1), uncertainty_ (0)
{
}



Position2D::Position2D (string dataString) : Data(),
  x_ (0), y_ (0), rho_ (0), angle_ (0), trustX_ (1), trustY_ (1), uncertainty_ (0)
{
  const char* data;
  data = dataString.c_str();
  float inX = 0, inY = 0;
  
  sscanf (data, "(%f, %f)", &inX, &inY);
  setXY (inX, inY);
}



Position2D::Position2D (float x, float y, bool isRect) : Data(),
  x_ (0), y_ (0), rho_ (0), angle_ (0), trustX_ (1), trustY_ (1), uncertainty_ (0)
{
  if (isRect) {
    setXY (x, y);
  }
  else {
    setPolar (x, y);
  }
}



Position2D::~Position2D (void)
{
}



Position2D& Position2D::operator= (const Position2D& pos)
{
  if (this == &pos) {
    return *this;
  }
  
  setXY (pos.getX(), pos.getY());
  setTrust (pos.getTrustX(), pos.getTrustY());
  setUncertainty (pos.getUncertainty());
  return *this;
}



Position2D& Position2D::operator+= (const Position2D& pos)
{
  setXY (getX() + pos.getX(), getY() + pos.getY());
  //The trust values have to be recalculated, cant be added.
  return *this;
}



Position2D& Position2D::operator-= (const Position2D& pos)
{
  setXY (getX() - pos.getX(), getY() - pos.getY());
  //The trust values have to be recalculated, cant be added.
  return *this;
}



const Position2D Position2D::operator+ (const Position2D& pos) const
{
  Position2D result;
  result.setXY (getX() + pos.getX(), getY() + pos.getY());
  //The trust values have to be recalculated, cant be added.
  return result;
}



const Position2D Position2D::operator- (const Position2D& pos) const
{
  Position2D result;
  result.setXY (getX() - pos.getX(), getY() - pos.getY());
  //The trust values have to be recalculated, cant be added.
  return result;
}



const Position2D Position2D::operator* (const float& value) const
{
  Position2D result;
  result.setXY (getX() *value, getY() *value);
  //The trust values have to be recalculated, cant be added.
  return result;
}



const Position2D Position2D::operator/ (const float& value) const
{
  Position2D result;
  result.setXY (getX() / value, getY() / value);
  //The trust values have to be recalculated, cant be added.
  return result;
}



Distance Position2D::getX() const
{
  return x_;
}



Distance Position2D::getY() const
{
  return y_;
}



Distance Position2D::getDistance() const
{
  return rho_;
}



Angle Position2D::getAzimuth() const
{
  return angle_;
}



void Position2D::setXY (Distance x, Distance y)
{
  x_ = x;
  y_ = y;
  updatePolar_();
}



void Position2D::setX (Distance x)
{
  setXY (x, y_);
}



void Position2D::setY (Distance y)
{
  setXY (x_, y);
}



void Position2D::setPolar (Distance rho, Angle angle)
{
  rho_ = rho;
  angle_ = angle;
  updateRect_();
}



void Position2D::setDistance (Distance rho)
{
  setPolar (rho, angle_);
}



void Position2D::setAzimuth (Angle angle)
{
  setPolar (rho_, angle);
}



void Position2D::updateRect_()
{
  FrameTransformations2D::polar2Rect (rho_, angle_, &x_, &y_);
}



void Position2D::updatePolar_()
{
  FrameTransformations2D::rect2Polar (x_, y_, &rho_, &angle_);
}



void Position2D::setTrust (float trustX, float trustY)
{
  trustX_ = trustX;
  trustY_ = trustY;
}



float Position2D::getTrustX() const
{
  return trustX_;
}



float Position2D::getTrustY() const
{
  return trustY_;
}



void Position2D::setUncertainty (float value)
{
  uncertainty_ = value;
}



float Position2D::getUncertainty() const
{
  return uncertainty_;
}



shared_ptr<Parameter> Position2D::clone() const
{
  shared_ptr<Position2D> tmp (new Position2D (*this));
  return tmp;
}



string Position2D::toString() const
{
  stringstream ret;
  ret << "(" << getX() << ", " << getY() << ")";
  
  return ret.str();
}
