/**
 * @Filename Position2D.hpp
 * Position2D Data class header.
 * @Status Implementing
 * @Version $Id: Position2D.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _POSITION_2D_HPP_
#define _POSITION_2D_HPP_

#include <Data.hpp>
#include <Defines.hpp>

#include <socrob_ros/Position2D.h>



/**
 * @Class Position2D Position2D.hpp "Position2D.hpp"
 * This class is used to store 2 dimentions floats, aimed mainly to store x,y
 * (or r,theta) coordinates in a local frame. It is basically the same as
 * Vector2D, but created to better distinguish from the Velocity2D.
 * @Author Hugo Costelha
 */
class Position2D : public Data
{
  public:
    /**
     * Constructor.
     * This method creates a new Position2D.
     */
    Position2D (void);
    
    /**
     * Constructor.
     * This method creates a new Position2D from a string. We assume rectangular
     * coordinates here.
     * @Argument stringData - string with X Y values
     */
    Position2D (string stringData);
    
    /**
     * Constructor.
     * This method creates a new Position2D using rectangular or polar
     * coordinates.
     * @Argument x - x value of the rectangular coordinates, or, if using polar
     * coordinates, the distance.
     * @Argument y - y value of the rectangular coordinates, or, if using polar
     * coordinates, the angle.
     * @Argument isRect - (optional argument) if true (default), the arguments
     * are used as rectangular coordinates, otherwise they are used as polar
     * coordinates.
     */
    Position2D (float x, float y, bool isRect = true);
    
    /**
     * Destructor.
     * This method makes deinitializations for objects of this class.
     */
    ~Position2D (void);
    
    Position2D& operator= (const Position2D& pos);
    Position2D& operator+= (const Position2D& pos);
    Position2D& operator-= (const Position2D& pos);
    const Position2D operator+ (const Position2D& original) const;
    const Position2D operator- (const Position2D& original) const;
    const Position2D operator* (const float& value) const;
    const Position2D operator/ (const float& value) const;
    
    /**
     * This method retrieves the x position value (rectangular coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getX() const;
    
    /**
     * This method retrieves the y position value (rectangular coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getY() const;
    
    /**
     * This method retrieves the rho position value (polar coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getDistance() const;
    
    /**
     * This method retrieves the angle position value (polar coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Angle getAzimuth() const;
    
    /**
     * This method sets the position values (rectangular coordinates).
     * @Argument x - the X value to store.
     * @Argument y - the Y value to store.
     */
    void setXY (Distance x, Distance y);
    
    /**
     * This method sets the X position value (rectangular coordinates).
     * @Argument x - the X value to store.
     */
    void setX (Distance x);
    
    /**
     * This method sets the Y position value (rectangular coordinates).
     * @Argument y - the Y value to store.
     */
    void setY (Distance y);
    
    /**
     * This method sets the position values (polar coordinates).
     * @Argument distance - the rho value to store.
     * @Argument azimuth - the angle value to store.
     */
    void setPolar (Distance distance, Angle azimuth);
    
    /**
     * This method sets the position distance value (polar coordinates).
     * @Argument distance - the rho value to store.
     */
    void setDistance (Distance distance);
    
    /**
     * This method sets the position angle value (polar coordinates).
     * @Argument azimuth - the angle value to store.
     */
    void setAzimuth (Angle azimuth);
    
    /**
     * Change the uncertainty. 0 means no uncertainty, while 1 means full
     * uncertainty.
     * @Argument value - new uncertainty value.
     */
    void setUncertainty (float value);
    
    /**
     * Retrieve the uncertainty
     */
    float getUncertainty() const;
    
    /**
     * Change the certainty associated with each component. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument trustX - certainty of the X component.
     * @Argument trustY - certainty of the Y component.
     */
    void setTrust (float trustX, float trustY);
    
    /**
     * Retrieve the X value certainty.
     */
    float getTrustX() const;
    
    /**
     * Retrieve the Y value certainty.
     */
    float getTrustY() const;
    
    shared_ptr<Parameter> clone() const;
    
    /**
     * Get the rectangular coordinated in string format.
     * @return string with the rectangular coodinates.
     */
    string toString() const;
    
    void rosMsg (socrob_ros::Position2D& msg) const {
      msg.x = x_;
      msg.y = y_;
      msg.rho = rho_;
      msg.angle = angle_;
      msg.trustX = trustX_;
      msg.trustY = trustY_;
      msg.uncertainty = uncertainty_;
    }
    
    void fromRosMsg (const socrob_ros::Position2D& msg) {
      x_ = msg.x;
      y_ = msg.y;
      rho_ = msg.rho;
      angle_ = msg.angle;
      trustX_ = msg.trustX;
      trustY_ = msg.trustY;
      uncertainty_ = msg.uncertainty;
    }
    
    static shared_ptr<Position2D> newFromRosMsg (const socrob_ros::Position2D& msg) {
      shared_ptr<Position2D> tmp = shared_ptr<Position2D> (new Position2D ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
  private:
    /**
     * Update the stored rectangular values using the polar ones.
     */
    void updateRect_();
    
    /**
     * Update the stored polar values using the rectangular ones.
     */
    void updatePolar_();
    
    float x_;  ///< X value (rectangular coordinates)
    float y_;  ///< Y value (rectangular coordinates)
    float rho_;  ///< Rho value (polar coordinates)
    float angle_;  ///< Angle value (polar coordinates)
    
    float trustX_; ///< Certainty of the X value
    float trustY_; ///< Certainty of the Y value
    
    float uncertainty_; ///< Uncertainty associated with this object
};
#endif
