/**
 * @Filename Position3D.cpp
 * Position3D Data class implementation.
 * @Status Finished
 * @Version $Id: Position3D.cpp 5353 2011-11-11 05:52:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#include <Position3D.hpp>
#include <FrameTransformations2D.hpp>
#include <FrameTransformations3D.hpp>

#include <sstream>



Position3D::
Position3D (void) :
  Data(),
  x_ (0),
  y_ (0),
  z_ (0),
  rho_ (0),
  azimuth_ (0),
  zenith_ (0),
  rhoInXYPlane_ (0),
  uncertainty_ (0),
  confidence_ (0),
  trustX_ (1),
  trustY_ (1),
  trustZ_ (1)
{
}



Position3D::
Position3D (string dataString) :
  Data(),
  x_ (0),
  y_ (0),
  z_ (0),
  rho_ (0),
  azimuth_ (0),
  zenith_ (0),
  rhoInXYPlane_ (0),
  uncertainty_ (0),
  confidence_ (0),
  trustX_ (1),
  trustY_ (1),
  trustZ_ (1)
{
  const char* data;
  data = dataString.c_str();
  float inX = 0, inY = 0, inZ = 0;
  
  sscanf (data, "(%f, %f, %f)", &inX, &inY, &inZ);
  setXYZ (inX, inY, inZ);
}



Position3D::
Position3D (Distance x,
            Distance y,
            Distance z,
            bool isCartesian) :
  Data(),
  x_ (0),
  y_ (0),
  z_ (0),
  rho_ (0),
  azimuth_ (0),
  zenith_ (0),
  rhoInXYPlane_ (0),
  uncertainty_ (0),
  confidence_ (0),
  trustX_ (1),
  trustY_ (1),
  trustZ_ (1)
{
  if (isCartesian) {
    setXYZ (x, y, z);
  }
  else {
    setSpherical (x, y, z);
  }
}



Position3D::
~Position3D (void)
{
}



shared_ptr<Parameter>
Position3D::
clone() const
{
  shared_ptr<Position3D> tmp (new Position3D (*this));
  return tmp;
}



Position3D&
Position3D::
operator= (const Position3D& pos)
{
  if (this == &pos) {
    return *this;
  }
  
  setXYZ (pos.getX(), pos.getY(), pos.getZ());
  setTrust (pos.getTrustX(), pos.getTrustY(), pos.getTrustZ());
  setUncertainty (pos.getUncertainty());
  return *this;
}



Position3D&
Position3D::
operator+= (const Position3D& pos)
{
  setXYZ (getX() + pos.getX(),
          getY() + pos.getY(),
          getZ() + pos.getZ());
  //The trust values have to be recalculated, cant be added.
  return *this;
}



Position3D&
Position3D::
operator-= (const Position3D& pos)
{
  setXYZ (getX() - pos.getX(),
          getY() - pos.getY(),
          getZ() - pos.getZ());
  //The trust values have to be recalculated, cant be added.
  return *this;
}



const Position3D
Position3D::
operator+ (const Position3D& pos)
const
{
  Position3D result;
  result.setXYZ (getX() + pos.getX(),
                 getY() + pos.getY(),
                 getZ() + pos.getZ());
  //The trust values have to be recalculated, cant be added.
  return result;
}



const Position3D
Position3D::
operator- (const Position3D& pos)
const
{
  Position3D result;
  result.setXYZ (getX() - pos.getX(),
                 getY() - pos.getY(),
                 getZ() - pos.getZ());
  //The trust values have to be recalculated, cant be added.
  return result;
}



const Position3D
Position3D::
operator* (const float& value)
const
{
  Position3D result;
  result.setXYZ (getX() *value, getY() *value, getZ() *value);
  //The trust values have to be recalculated, cant be added.
  return result;
}



const Position3D
Position3D::
operator/ (const float& value)
const
{
  Position3D result;
  result.setXYZ (getX() / value, getY() / value, getZ() / value);
  //The trust values have to be recalculated, cant be added.
  return result;
}



Distance
Position3D::
getX()
const
{
  return x_;
}



Distance
Position3D::
getY()
const
{
  return y_;
}



Distance
Position3D::
getZ()
const
{
  return z_;
}



Distance
Position3D::
getDistance()
const
{
  return rho_;
}



Distance
Position3D::
getDistanceInXYPlane()
const
{
  return rhoInXYPlane_;
}



Angle
Position3D::
getAzimuth()
const
{
  return azimuth_;
}



Angle
Position3D::
getZenith()
const
{
  return zenith_;
}



void
Position3D::
setXYZ (Distance x,
        Distance y,
        Distance z)
{
  x_ = x;
  y_ = y;
  z_ = z;
  updateSpherical_();
}



void
Position3D::
setX (Distance x)
{
  setXYZ (x, y_, z_);
}



void
Position3D::
setY (Distance y)
{
  setXYZ (x_, y, z_);
}



void
Position3D::
setZ (Distance z)
{
  setXYZ (x_, y_, z);
}



void
Position3D::
setSpherical (Distance rho,
              Angle azimuth,
              Angle zenith)
{
  rho_ = rho;
  azimuth_ = azimuth;
  zenith_ = zenith;
  
  // We need to update the rho in the XYPlane so, we update the
  // cartesian coordinates, and then compute that rho.
  updateCartesian_();
  Angle temp;
  FrameTransformations2D::rect2Polar (x_, y_,
                                      &rhoInXYPlane_, &temp);
}



void
Position3D::
setDistance (Distance rho)
{
  setSpherical (rho, azimuth_, zenith_);
}



void
Position3D::
setAzimuth (Angle azimuth)
{
  setSpherical (rho_, azimuth, zenith_);
}



void
Position3D::
setZenith (Angle zenith)
{
  setSpherical (rho_, azimuth_, zenith);
}



void
Position3D::
updateCartesian_()
{
  FrameTransformations3D::spherical2Cartesian (rho_, azimuth_, zenith_,
      &x_, &y_, &z_);
}



void
Position3D::
updateSpherical_()
{
  FrameTransformations3D::cartesian2Spherical (x_, y_, z_,
      &rho_, &azimuth_, &zenith_);
  // Update also the rho value in the XYPlane
  Angle temp;
  FrameTransformations2D::rect2Polar (x_, y_,
                                      &rhoInXYPlane_, &temp);
}



void
Position3D::
setUncertainty (float value)
{
  uncertainty_ = value;
}



float
Position3D::
getUncertainty()
const
{
  return uncertainty_;
}



void
Position3D::
setConfidence (float value)
{
  confidence_ = value;
}



float
Position3D::
getConfidence()
const
{
  return confidence_;
}



void
Position3D::
setTrust (float trustX,
          float trustY,
          float trustZ)
{
  trustX_ = trustX;
  trustY_ = trustY;
  trustZ_ = trustZ;
}



float
Position3D::
getTrustX()
const
{
  return trustX_;
}



float
Position3D::
getTrustY()
const
{
  return trustY_;
}



float
Position3D::
getTrustZ()
const
{
  return trustZ_;
}



string
Position3D::
toString() const
{
  stringstream ret;
  ret << "(" << getX() << ", " << getY() << ", " << getZ() << ")";
  
  return ret.str();
}
