/**
 * @Filename Position3D.hpp
 * Position3D Data class header.
 * @Status Finished
 * @Version $Id: Position3D.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _POSITION_3D_HPP_
#define _POSITION_3D_HPP_

#include <Data.hpp>
#include <Defines.hpp>

#include <socrob_ros/Position3D.h>



/**
 * @Class Position3D Position3D.hpp "Position3D.hpp"
 * This class is used to store 3 dimentions floats, aimed mainly to store x,y,z
 * (or rho,zenith,azimuth) coordinates in a local frame.
 * @Author Hugo Costelha
 */
class Position3D : public Data
{
  public:
    /**
     * Constructor.
     * This method creates a new Position3D.
     */
    Position3D (void);
    
    /**
     * Constructor.
     * This method creates a new Position3D from a string. We assume cartesian
     * coordinates here.
     * @Argument stringData - string with X, Y and Z values
     */
    Position3D (string stringData);
    
    /**
     * Constructor.
     * This method creates a new Position3D using cartesian or spherical
     * coordinates.
     * @Argument x - x value of the cartesian coordinates, or, if using
     * spherical coordinates, the rho value.
     * @Argument y - y value of the cartesian coordinates, or, if using
     * spherical coordinates, the azimuth angle (measured from the x axis).
     * @Argument z - z value of the cartesian coordinates, or, if using
     * spherical coordinates, the zenith (measured from the z axis).
     * @Argument isCartesian - (optional argument) if true (default), the
     * arguments are used as cartesian coordinates, otherwise they are used as
     * spherical coordinates.
     */
    Position3D (Distance x, Distance y, Distance z, bool isCartesian = true);
    
    /**
     * Destructor.
     * This method makes deinitializations for objects of this class.
     */
    ~Position3D (void);
    
    Position3D& operator= (const Position3D& pos);
    Position3D& operator+= (const Position3D& pos);
    Position3D& operator-= (const Position3D& pos);
    const Position3D operator+ (const Position3D& original) const;
    const Position3D operator- (const Position3D& original) const;
    const Position3D operator* (const float& value) const;
    const Position3D operator/ (const float& value) const;
    
    /**
     * This method retrieves the x value of the position in cartesian
     * coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getX() const;
    
    /**
     * This method retrieves the y position value in cartesian coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getY() const;
    
    /**
     * This method retrieves the z position value in cartesian coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getZ() const;
    
    /**
     * This method retrieves the position rho value in spherical coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getDistance() const;
    
    /**
     * This method retrieves the position rho value in spherical coordinates,
     * on the XY plane (ignoring the z value).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Distance getDistanceInXYPlane() const;
    
    /**
     * This method retrieves the azimuth, measured from the x-axis in the XY
     * plane (spherical coordinates). This angle is the same has in cartesian
     * coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Angle getAzimuth() const;
    
    /**
     * This method retrieves the angle zenith, measured from the z-axis
     * (spherical coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Angle getZenith() const;
    
    /**
     * This method sets the position values in cartesian coordinates.
     * @Argument x - the position X value to store.
     * @Argument y - the position Y value to store.
     * @Argument z - the position Z value to store.
     */
    void setXYZ (Distance x, Distance y, Distance z);
    
    /**
     * This method sets the x position value (cartesian coordinates).
     * @Argument x - the position X value to store.
     */
    void setX (Distance x);
    
    /**
     * This method sets the y position value (cartesian coordinates).
     * @Argument y - the position Y value to store.
     */
    void setY (Distance y);
    
    /**
     * This method sets the z position value (cartesian coordinates).
     * @Argument vz - the position Z value to store.
     */
    void setZ (Distance z);
    
    /**
     * This method sets the value in spherical coordinates.
     * @Argument rho - the rho value to store.
     * @Argument azimuth - the angle from the X axis (in XY plane)
     * @Argument zentih - the angle from the z axis.
     */
    void setSpherical (Distance distance, Angle azimuth, Angle zenith);
    
    /**
     * This method sets the position distance value (spherical coordinates).
     * @Argument distance - the position distance value to store.
     */
    void setDistance (Distance distance);
    
    /**
     * This method sets the azimuth value (spherical coordinates).
     * @Argument azimuth - the angle from the X axis (in XY plane)
     */
    void setAzimuth (Angle azimuth);
    
    /**
     * This method sets the zenith value (spherical coordinates).
     * @Argument zentih - the angle from the z axis.
     */
    void setZenith (Angle zenith);
    
    /**
     * Change the uncertainty associated with this object. 0 means no
     * uncertainty, while 1 means full uncertainty.
     * @Argument value - new uncertainty value.
     */
    void setUncertainty (float value);
    
    /**
     * Retrieve the uncertainty
     */
    float getUncertainty() const;
    /**
     * Change the Confidence associated with this object. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument value - new uncertainty value.
     */
    void setConfidence (float value);
    
    /**
     * Retrieve the uncertainty
     */
    float getConfidence() const;
    
    
    /**
     * Change the cercainty associated with each component. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument trustX - certainty of the X component.
     * @Argument trustY - certainty of the Y component.
     * @Argument trustTheta - certainty of the Theta component.
     */
    void setTrust (float trustX, float trustY, float trustTheta);
    
    /**
     * Retrieve the X value certainty.
     */
    float getTrustX() const;
    
    /**
     * Retrieve the Y value certainty.
     */
    float getTrustY() const;
    
    /**
     * Retrieve the Theta value certainty.
     */
    float getTrustZ() const;
    
    shared_ptr<Parameter> clone() const;
    
    /**
     * Get the cartesian coordinated in string format.
     * @Returns string with the cartesian coodinates.
     */
    string toString() const;
    
    void rosMsg (socrob_ros::Position3D& msg) const {
      msg.x = x_;
      msg.y = y_;
      msg.z = z_;
      msg.rho = rho_;
      msg.azimuth = azimuth_;
      msg.zenith = zenith_;
      msg.rhoInXYPlane = rhoInXYPlane_;
      
      msg.uncertainty = uncertainty_;
      msg.confidence = confidence_;
      msg.trustX = trustX_;
      msg.trustY = trustY_;
      msg.trustZ = trustZ_;
    }
    
    void fromRosMsg (const socrob_ros::Position3D& msg) {
      x_ = msg.x;
      y_ = msg.y;
      z_ = msg.z;
      rho_ = msg.rho;
      azimuth_ = msg.azimuth;
      zenith_ = msg.zenith;
      rhoInXYPlane_ = msg.rhoInXYPlane;
      
      uncertainty_ = msg.uncertainty;
      confidence_ = msg.confidence;
      trustX_ = msg.trustX;
      trustY_ = msg.trustY;
      trustZ_ = msg.trustZ;
    }
    
    static shared_ptr<Position3D> newFromRosMsg (const socrob_ros::Position3D& msg) {
      shared_ptr<Position3D> tmp = shared_ptr<Position3D> (new Position3D ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
  private:
    /**
     * Update the stored cartesian values using the spherical ones.
     */
    void updateCartesian_();
    
    /**
     * Update the stored spherical values using the cartesian ones.
     */
    void updateSpherical_();
    
    float x_;  ///< X value (cartesian coordinates)
    float y_;  ///< Y value (cartesian coordinates)
    float z_;  ///< Y value (cartesian coordinates)
    float rho_;  ///<Rho value (shperical coordinates)
    float azimuth_;  ///< Angle value in XY plane (spherical coordinates)
    float zenith_;  ///< Angle value (shperical coordinates)
    float rhoInXYPlane_;  ///< Rho value (polar coordinates in XY plane)
    
    float uncertainty_; ///< Uncertainty associated with this object
    float confidence_;  ///< Confidence associated with this object
    float trustX_; ///< Certainty of the X value
    float trustY_; ///< Certainty of the Y value
    float trustZ_; ///< Certainty of the Z value
};
#endif
