/**
 * @Filename Posture.hpp
 * @Description Posture class header.
 * @Status Finished
 * @Version $Id: Posture.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _POSTURE_HPP_
#define _POSTURE_HPP_

#include <Data.hpp>
#include <Parameter.hpp>

#include <Defines.hpp>

#include <Position2D.hpp>

#include <iostream>
#include <string>

#include <socrob_ros/Posture.h>



using namespace std;



/**
 * @Class Posture Posture.hpp "Posture.hpp"
 * This class is used to store postures (x,y,theta) values. Typically, the
 * stored posture is in local coorditates. This class also provides methods to
 * return the posture value in world/local coordinates.
 * @Author Hugo Costelha
 */
class Posture : public Data
{
  public:
    /**
     * Class base constructor (everything is set to 0)
     */
    Posture();
    
    /**
     * Create the class from a string.
     * @Argument dataString - String with data in the format "(%f, %f, %f)".
     */
    Posture (string dataString);
    
    /**
     * Create a new posture with the given values in SI coordinates.
     * @Argument x - X value (or rho if polar coordinates)
     * @Argument y - Y value (or angle if polar coordinates)
     * @Argument theta - Theta value
     * @Argument isRect - (optional) if true (default), then the arguments
     * refer to rectangular coordinates, otherwise they refer to polar
     * coordinates.
     */
    Posture (float x, float y, float theta, bool isRect = true);
    
    /**
     * Destructor
     */
    ~Posture();
    
    /**
     * Retrieve the posture X value (rectangular coordinates)
     */
    Distance getX() const;
    
    /**
     * Retrieve the posture Y value (rectangular coordinates)
     */
    Distance getY() const;
    
    /**
     * Retreive the posture rho value (polar coordinates)
     */
    Distance getDistance() const;
    
    /**
     * Retrieve the azimuth posture value (polar coordinates)
     */
    Angle getAzimuth() const;
    
    /**
     * Retrieve the posture orienation value
     */
    Angle getTheta() const;
    
    /**
     * Change the X,Y, Theta posture values based on the given polar
     * coordinates.
     * @Argument distance - the theta rho value in polar coordinates.
     * @Argument azimuth - the posture angle in polar coordinate.
     * @Argument theta - new posture orientation value.
     */
    void setPolar (Distance distance, Angle azimuth, Angle theta);
    
    /**
     * Change the X,Y posture values based on the given polar coordinates.
     * @Argument distance - the posture rho value in polar coordinates.
     * @Argument azimuth - the posture angle in polar coordinate.
     */
    void setPolar (Distance distance, Angle azimuth);
    
    /**
     * Change the posture value (polar coordinates).
     * @Argument speed - the rho value in polar coordinates.
     */
    void setDistance (Distance distance);
    
    /**
     * Change the posture angle value (polar coordinates).
     * @Argument azimuth - the new angle polar coordinate.
     */
    void setAzimuth (Angle azimuth);
    
    /**
     * Change the X,Y and Theta posture values.
     * @Argument x - The new X posture value.
     * @Argument y - The new Y posture value.
     * @Argument theta - The new posture orientation value.
     */
    void setXYTheta (Distance x, Distance y, Angle theta);
    
    /**
     * Change the X,Y posture values (cartesian coordinates).
     * @Argument x - The new X posture value.
     * @Argument y - The new Y posture value.
     */
    void setXY (Distance x, Distance y);
    
    /**
     * Change the X posture value (cartesian coordinates).
     * @Argument x - The new X posture value.
     */
    void setX (Distance x);
    
    /**
     * Change the Y posture value (rectangulaar coordinates).
     * @Argument y - The new Y posture value.
     */
    void setY (Distance y);
    
    /**
     * Change the orientation value.
     * @Argument theta - The new posture orientation value.
     */
    void setTheta (Angle theta);
    
    /**
     * Retrieve the uncertainty
     */
    float getUncertainty() const;
    
    /**
     * Retrieve the X value certainty.
     */
    float getTrustX() const;
    
    /**
     * Retrieve the Y value certainty.
     */
    float getTrustY() const;
    
    /**
     * Retrieve the Theta value certainty.
     */
    float getTrustTheta() const;
    
    /**
     * Change the uncertainty associated with this vector. 0 means no
     * uncertainty, while 1 means full uncertainty.
     * @Argument value - new uncertainty value.
     */
    void setUncertainty (float value);
    
    /**
     * Change the cercainty associated with each component. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument trustX - certainty of the X component.
     * @Argument trustY - certainty of the Y component.
     * @Argument trustTheta - certainty of the Theta component.
     */
    void setTrust (float trustX, float trustY, float trustTheta);
    
    shared_ptr<Parameter> clone() const;
    
    /**
     * Return the posture value in X,Y,Theta form.
     * @Returns the string with the posture valu in the X,Y,Theta form.
     */
    string toString() const;
    
    void rosMsg (socrob_ros::Posture& msg) const {
      pos2D_.rosMsg (msg.pos2D);
      msg.theta = theta_;
      msg.uncertainty = uncertainty_;
      msg.trustTheta = trustTheta_;
    }
    
    void fromRosMsg (const socrob_ros::Posture& msg) {
      pos2D_.fromRosMsg (msg.pos2D);
      theta_ = msg.theta;
      uncertainty_ = msg.uncertainty;
      trustTheta_ = msg.trustTheta;
    }
    
    static shared_ptr<Posture> newFromRosMsg (const socrob_ros::Posture& msg) {
      shared_ptr<Posture> tmp = shared_ptr<Posture> (new Posture ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
  private:
    Position2D pos2D_; ///< 2D value of the vector
    float theta_; ///< Angle vector value
    float uncertainty_; ///< Uncertainty associated with this object
    float trustTheta_; ///< Certainty of the Theta value
};

#endif
