/**
 * @Filename TimeStamp.cpp
 * TimeStamp class implementation.
 * @Status Finished
 * @Version $Id: TimeStamp.cpp 5661 2012-05-11 13:43:55Z jreis $
 * @Maintainer Hugo Veiga (hasv@mega.ist.utl.pt)
 */

#include "TimeStamp.hpp"
#include <sys/time.h>
#include <math.h>
#include <stdlib.h>

#include <sstream>



using namespace std;



TimeStamp::TimeStamp() :
  timeStamp_()
{
  setTimeToNow();
}



TimeStamp::TimeStamp (double value) :
  timeStamp_()
{
  setTime (value);
}



TimeStamp::~TimeStamp()
{
}



double TimeStamp::getTime() const
{
  return (double) (timeStamp_.tv_sec * 1e6 + timeStamp_.tv_usec);
}



double TimeStamp::getTimeInMs() const
{
  return (double) (timeStamp_.tv_sec * 1e3 + timeStamp_.tv_usec / 1000.0);
}



bool TimeStamp::olderThan (TimeStamp val) const
{
  if (timeStamp_.tv_sec < val.timeStamp_.tv_sec) {
    return true;
  }
  else if (timeStamp_.tv_sec == val.timeStamp_.tv_sec) {
    if (timeStamp_.tv_usec < val.timeStamp_.tv_usec) {
      return true;
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
}



bool TimeStamp::newerThan (TimeStamp val) const
{
  return ! (olderThan (val)); // maybe not 100% correct....
}



static inline bool time_gt_ (const struct timeval& a, const struct timeval& b)
{
  // The semantic of this function is exactly a > b
  if (a.tv_sec > b.tv_sec) {
    return true;
  }
  if (a.tv_sec < b.tv_sec) {
    return false;
  }
  if (a.tv_usec > b.tv_usec) {
    return true;
  }
  return false;
}
static inline void time_diff_ (struct timeval& res, const struct timeval& a, const struct timeval& b)
{
  // BE CAREFUL! a really must be greater than b, or else 0 is returned
  // This is because the semantic is exactly res = a - b
  if (time_gt_ (b, a)) {
    res.tv_sec = 0;
    res.tv_usec = 0;
    return;
  }
  if (a.tv_usec < b.tv_usec) {
    res.tv_usec = 1000000 + a.tv_usec - b.tv_usec;
    res.tv_sec = a.tv_sec - b.tv_sec - 1;
  }
  else {
    res.tv_usec = a.tv_usec - b.tv_usec;
    res.tv_sec = a.tv_sec - b.tv_sec;
  }
}
bool TimeStamp::deltaSmallerThan (const TimeStamp& val, int msec) const
{
  struct timeval valueTV;
  valueTV.tv_sec = msec / 1000;
  valueTV.tv_usec = (msec * 1000) - (valueTV.tv_sec * 1000000);
  
  struct timeval deltaTV;
  if (time_gt_ (timeStamp_, val.timeStamp_)) {
    time_diff_ (deltaTV, timeStamp_, val.timeStamp_);
  }
  else {
    time_diff_ (deltaTV, val.timeStamp_, timeStamp_);
  }
  
  if (time_gt_ (valueTV, deltaTV)) {
    return true;
  }
  return false;
}



void TimeStamp::setTimeToNow()
{
  gettimeofday (&timeStamp_, 0);
}



void TimeStamp::setTimeStamp (TimeStamp timeStamp)
{
  timeStamp_ = timeStamp.timeStamp_ ;
}



void TimeStamp::setTime (double value)
{
  timeStamp_.tv_sec = (time_t) (value / 1e6);
  timeStamp_.tv_usec = (suseconds_t) (value - timeStamp_.tv_sec * 1e6);
}



void TimeStamp::timeDiff (const TimeStamp& timeStamp2, TimeStamp timeStamp1, struct timeval* result)
{
  /* Perform the carry for the later subtraction by updating y. */
  if (timeStamp2.timeStamp_.tv_usec < timeStamp1.timeStamp_.tv_usec) {
    int nsec = (timeStamp1.timeStamp_.tv_usec - timeStamp2.timeStamp_.tv_usec) / 1000000 + 1;
    timeStamp1.timeStamp_.tv_usec -= 1000000 * nsec;
    timeStamp1.timeStamp_.tv_sec += nsec;
  }
  if (timeStamp2.timeStamp_.tv_usec - timeStamp1.timeStamp_.tv_usec > 1000000) {
    int nsec = (timeStamp2.timeStamp_.tv_usec - timeStamp1.timeStamp_.tv_usec) / 1000000;
    timeStamp1.timeStamp_.tv_usec += 1000000 * nsec;
    timeStamp1.timeStamp_.tv_sec -= nsec;
  }
  
  // Compute the time remaining to wait.
  // tv_usec is certainly positive.
  result->tv_sec = timeStamp2.timeStamp_.tv_sec - timeStamp1.timeStamp_.tv_sec;
  result->tv_usec = timeStamp2.timeStamp_.tv_usec - timeStamp1.timeStamp_.tv_usec;
}



void TimeStamp::timeDiff (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, long* result)
{
  struct timeval deltaTime;
  timeDiff (timeStamp2, timeStamp1, &deltaTime);
  *result = static_cast<long> (deltaTime.tv_sec) * 1000000 + deltaTime.tv_usec;
}



void TimeStamp::timeDiffms (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, long int* result)
{
  struct timeval deltaTime;
  timeDiff (timeStamp2, timeStamp1, &deltaTime);
  *result = deltaTime.tv_sec * 1000 + deltaTime.tv_usec / 1000;
}



void TimeStamp::timeDiff (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, double* result)
{
  struct timeval deltaTime;
  timeDiff (timeStamp2, timeStamp1, &deltaTime);
  *result = deltaTime.tv_sec + deltaTime.tv_usec / 1000000.0;
}



void TimeStamp::timeDiff (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, float* result)
{
  struct timeval deltaTime;
  timeDiff (timeStamp2, timeStamp1, &deltaTime);
  *result = deltaTime.tv_sec + deltaTime.tv_usec / 1000000.0;
}



string TimeStamp::toString() const
{
  stringstream ret;
  ret << "[";
  ret << timeStamp_.tv_sec;
  ret << ",";
  ret << timeStamp_.tv_usec;
  ret << "]";
  return ret.str();
}



//eof
