/**
* @Filename TimeStamp.hpp
* TimeStamp class header.
* @Status Finished
* @Version $Id: TimeStamp.hpp
* @Maintainer Hugo Costelha
*/

#ifndef _TIMESTAMP_HPP_
#define _TIMESTAMP_HPP_

#include <sys/time.h>

#include <string>



using namespace std;



/**
 * @Class TimeStamp TimeStamp.hpp "TimeStamp.hpp"
 * Class TimeStamp stores a timeStamp, which is used to hold the time value
 * associated with a given object, differences between timeStamps, elapsed
 * times, etc..
 * @Author Hugo Veiga
 */
class TimeStamp
{
  public:
    /**
     * Constructor.
     * This method creates a new TimeStamp.
     */
    TimeStamp();
    
    /**
     * Constructor.
     * This method creates a new TimeStamp.
     * Receives as a parameter a double with a TimeStamp in usec
     * @Argument timestamp - ?
     */
    TimeStamp (double timestamp);
    
    /**
     * Destructor.
     * This method makes deinitializations for objects of this class.
     */
    ~TimeStamp();
    
    /**
     * @Returns value of the time stamp in usec.
     */
    double getTime() const;
    
    /**
     * @Returns value of the time stamp in ms.
     */
    double getTimeInMs() const;
    
    /**
     * Change the time stamp value to the current time.
     */
    void setTimeToNow();
    
    /**
     * Change the time stamp value to the given time.
     * @Argument timeStamp - new time stamp value.
     */
    void setTimeStamp (TimeStamp timeStamp);
    
    /**
     * Change the time stamp value to the given time (in usec).
     * @Argument timeInMs - new time stamp value (in usec).
     */
    void setTime (double newTime);
    
    /**
     * Check if this object's time stamp is older than the given time stamp.
     * @Argument timestamp - The time stamp to compare with.
     * @Returns Boolean saying if this timestamp is older than the given one
     */
    bool olderThan (TimeStamp timestamp) const;
    
    /**
     * Check if this object's time stamp is newer than the given time stamp.
     * @Argument timestamp - The time stamp to compare with.
     * @Returns Boolean saying if this timestamp is newer than the given one
     */
    bool newerThan (TimeStamp timestamp) const;
    
    /**
     * Check if the this object's time stamp minus the given time stamp is less
     * than a given delta (in msec).
     * @Argument timestamp - the time stamp to compare with.
     * @Argument msec - the time delta value to be tested.
     * @returns true if this time stamp minus the given time stamp is less than
     * the given delta.
     */
    bool deltaSmallerThan (const TimeStamp& timestamp, int msec) const;
    
    /**
     * Convert the object's time stamp to a string.
     * @returns return string representation of TimeStamp
     */
    std::string toString() const;
    
    /**
     * Compute the timeStamp2 (first argument) minus timeStamp1 (second
     * argument), storing the result as struct timeval (third argument)
     * @Argument timeStamp2 - the time stamp to subtract from.
     * @Argument timeStamp1 - the time stamp to subtract.
     * @Argument result - the resulting difference of timeStamp2-timeStamp1.
     */
    static void timeDiff (const TimeStamp& timeStamp2, TimeStamp timeStamp1, struct timeval* result);
    static void timeDiff (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, long* result);
    long timeDiff (const TimeStamp& timeStamp1) {
      long result;
      TimeStamp::timeDiff (*this, timeStamp1, &result);
      return result;
    }
    
    /**
     * Compute the timeStamp2 (first argument) minus timeStamp1 (second
     * argument), storing the result in msec (third argument)
     * @Argument timeStamp2 - the time stamp to subtract from.
     * @Argument timeStamp1 - the time stamp to subtract.
     * @Argument result - the resulting difference of timeStamp2-timeStamp1.
     */
    static void timeDiffms (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, long int* result);
    
    /**
     * Compute the timeStamp2 (first argument) minus timeStamp1 (second
     * argument), storing the result in sec (third argument, float precision)
     * @Argument timeStamp2 - the time stamp to subtract from.
     * @Argument timeStamp1 - the time stamp to subtract.
     * @Argument result - the resulting difference of timeStamp2-timeStamp1.
     */
    static void timeDiff (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, float* result);
    
    /**
     * Compute the timeStamp2 (first argument) minus timeStamp1 (second
     * argument), storing the result in sec (third argument, double precision)
     * @Argument timeStamp2 - the time stamp to subtract from.
     * @Argument timeStamp1 - the time stamp to subtract.
     * @Argument result - the resulting difference of timeStamp2-timeStamp1.
     */
    static void timeDiff (const TimeStamp& timeStamp2, const TimeStamp& timeStamp1, double* result);
    
  private:
    struct timeval timeStamp_; ///< The actual time value of the time stamp
};

#endif
