/**
 * @Filename TrackedPosition3D.hpp
 * TrackedPosition3D Data class header.
 * @Status Implementing
 * @Version $Id: TrackedPosition3D.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _TRACKED_POSITION_3D_HPP_
#define _TRACKED_POSITION_3D_HPP_

#include <Data.hpp>
#include <Position3D.hpp>
#include <Velocity3D.hpp>
#include <FullPosture.hpp>

#include <socrob_ros/TrackedPosition3D.h>



/**
 * @Class TrackedPosition3D TrackedPosition3D.hpp "TrackedPosition3D.hpp"
 * This class is used to aggregate all the information regarding a tracked
 * object in a 3D space. The observer is typically a robot, having a full
 * posture reference frame.
 * @Author Hugo Costelha
 */
class TrackedPosition3D : public Data
{
  public:
    /**
     * Constructor.
     * Everything set to 0/false.
     */
    TrackedPosition3D();
    
    /**
     * Constructor.
     * @Argument newBaseFrame - base frame info;
     * @Argument newPosition - ball position in local or world coordinates;
     * @Argument newVelocity - ball velocity in local or world coordinates;
     * @Argument newSeeTrackedObject - if true, then the tracked object was
     * observed, otherwise it was etimated indirectly from other sources.
     * @Argument inLocalCoordinates - (optional, true by default) if true, then
     * the given position and velocity are in local coordinates, otherwise they
     * are in world coordinates.
     */
    TrackedPosition3D (FullPosture newBaseFrame, Position3D newPosition,
                       Velocity3D newVelocity, bool newSeeTrackedObject,
                       bool inLocalCoordinates = true);
                       
    shared_ptr<Parameter> clone() const;
    string toString() const;
    
    void rosMsg (socrob_ros::TrackedPosition3D& msg) const {
      localPosition.rosMsg (msg.localPosition);
      localVelocity.rosMsg (msg.localVelocity);
      baseFrame.rosMsg (msg.baseFrame);
      worldPosition.rosMsg (msg.worldPosition);
      worldVelocity.rosMsg (msg.worldVelocity);
      msg.seeTrackedObject = seeTrackedObject;
      msg.seeTrackedObjectLocally = seeTrackedObjectLocally;
      msg.seeTrackedObjectBySharing = seeTrackedObjectBySharing;
    }
    
    void fromRosMsg (const socrob_ros::TrackedPosition3D& msg) {
      localPosition.fromRosMsg (msg.localPosition);
      localVelocity.fromRosMsg (msg.localVelocity);
      baseFrame.fromRosMsg (msg.baseFrame);
      worldPosition.fromRosMsg (msg.worldPosition);
      worldVelocity.fromRosMsg (msg.worldVelocity);
      seeTrackedObject = msg.seeTrackedObject;
      seeTrackedObjectLocally = msg.seeTrackedObjectLocally;
      seeTrackedObjectBySharing = msg.seeTrackedObjectBySharing;
    }
    
    static shared_ptr<TrackedPosition3D> newFromRosMsg (const socrob_ros::TrackedPosition3D& msg) {
      shared_ptr<TrackedPosition3D> tmp = shared_ptr<TrackedPosition3D> (new TrackedPosition3D ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
    Position3D localPosition; ///< Position of the ball in local coordinates.
    Velocity3D localVelocity; ///< Velocity of the ball in local coordinates
    FullPosture baseFrame; ///< Base frame info
    Position3D worldPosition; ///< Position of the ball in world coordinates.
    Velocity3D worldVelocity; ///< Velocity of the ball in world coordinates.
    bool seeTrackedObject; ///< If true, the tracked objects was observed by shared vision if some sharing/fusuion is on,
    ///< otherwise it was estimated from other sources.
    bool seeTrackedObjectLocally; ///< If true, the tracked objects was observed by Local Camera's vision,
    ///< .
    bool seeTrackedObjectBySharing; ///< If true, the tracked objects was observed sharing among robots,
    ///< .
};

#endif
