/**
 * @Filename Velocity2D.hpp
 * Velocity2D Data class header.
 * @Status Implementing
 * @Version $Id: Velocity2D.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _VELOCITY_2D_HPP_
#define _VELOCITY_2D_HPP_

#include <Data.hpp>
#include <Defines.hpp>

#include <socrob_ros/Velocity2D.h>



/**
 * @Class Velocity2D Velocity2D.hpp "Velocity2D.hpp"
 * This class is used to store 2 dimentions floats, aimed mainly to store x,y
 * (or r,theta) coordinates in a local frame. It is basically the same as
 * Vector2D, but created to better distinguish from the Velocity2D.
 * @Author Hugo Costelha
 */
class Velocity2D : public Data
{
  public:
    /**
     * Constructor.
     * This method creates a new Velocity2D.
     */
    Velocity2D (void);
    
    /**
     * Constructor.
     * This method creates a new Velocity2D from a string. We assume polar
     * coordinates here.
     * @Argument stringData - string with X Y values
     */
    Velocity2D (string stringData);
    
    /**
     * Constructor.
     * This method creates a new Velocity2D using rectangular or polar
     * coordinates.
     * @Argument x - rho (distance) value of the polar coordinates, or, if using
     * rectangular coordinates, the x value.
     * @Argument y - angle value of the polar coordinates, or, if using
     * rectangular coordinates, the y value.
     * @Argument isRect - (optional argument) if false (default), the arguments
     * are used as polar coordinates, otherwise they are used as rectangular
     * coordinates.
     */
    Velocity2D (float rho, float angle, bool isRect = false);
    
    /**
     * Destructor.
     * This method makes deinitializations for objects of this class.
     */
    ~Velocity2D (void);
    
    /**
     * This method retrieves the x velocity value (rectangular coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getVx() const;
    
    /**
     * This method retrieves the y velocity value (rectangular coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getVy() const;
    
    /**
     * This method retrieves the rho velocity value (polar coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getSpeed() const;
    
    /**
     * This method retrieves the angle velocity value (polar coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Angle getAzimuth() const;
    
    /**
     * This method sets the velocity values (rectangular coordinates).
     * @Argument vx - the X value to store.
     * @Argument vy - the Y value to store.
     */
    void setVxVy (Speed vx, Speed vy);
    
    /**
     * This method sets the X velocity value (rectangular coordinates).
     * @Argument x - the X value to store.
     */
    void setVx (Speed vx);
    
    /**
     * This method sets the Y velocity value (rectangular coordinates).
     * @Argument y - the Y value to store.
     */
    void setVy (Speed vy);
    
    /**
     * This method sets the velocity values (polar coordinates).
     * @Argument distance - the rho value to store.
     * @Argument azimuth - the angle value to store.
     */
    void setPolar (Speed speed, Angle azimuth);
    
    /**
     * This method sets the velocity value (polar coordinates).
     * @Argument speed - the rho value to store.
     */
    void setSpeed (Speed speed);
    
    /**
     * This method sets the velocity angle value (polar coordinates).
     * @Argument azimuth - the angle value to store.
     */
    void setAzimuth (Angle azimuth);
    
    /**
     * Change the uncertainty. 0 means no uncertainty, while 1 means full
     * uncertainty.
     * @Argument value - new uncertainty value.
     */
    void setUncertainty (float value);
    
    /**
     * Retrieve the uncertainty
     */
    float getUncertainty() const;
    
    /**
     * Change the certainty associated with each component. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument trustX - certainty of the X component.
     * @Argument trustY - certainty of the Y component.
     */
    void setTrust (float trustVx, float trustVy);
    
    /**
     * Retrieve the X value certainty.
     */
    float getTrustVx() const;
    
    /**
     * Retrieve the Y value certainty.
     */
    float getTrustVy() const;
    
    shared_ptr<Parameter> clone() const;
    
    /**
     * Get the polar coordinated in string format.
     * @return string with the polar coodinates.
     */
    string toString() const;
    
    void rosMsg (socrob_ros::Velocity2D& msg) const {
      msg.vX = vX_;
      msg.vY = vY_;
      msg.speed = speed_;
      msg.azimuth = azimuth_;
      msg.trustVx = trustVx_;
      msg.trustVy = trustVy_;
      msg.uncertainty = uncertainty_;
    }
    
    void fromRosMsg (const socrob_ros::Velocity2D& msg) {
      vX_ = msg.vX;
      vY_ = msg.vY;
      speed_ = msg.speed;
      azimuth_ = msg.azimuth;
      trustVx_ = msg.trustVx;
      trustVy_ = msg.trustVy;
      uncertainty_ = msg.uncertainty;
    }
    
    static shared_ptr<Velocity2D> newFromRosMsg (const socrob_ros::Velocity2D& msg) {
      shared_ptr<Velocity2D> tmp = shared_ptr<Velocity2D> (new Velocity2D ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
  private:
    /**
     * Update the stored rectangular values using the polar ones.
     */
    void updateRect_();
    
    /**
     * Update the stored polar values using the rectangular ones.
     */
    void updatePolar_();
    
    float vX_;  ///< X velocity value (rectangular coordinates)
    float vY_;  ///< Y velocity value (rectangular coordinates)
    float speed_;  ///< Rho velocity value (polar coordinates)
    float azimuth_;  ///< Angle velocity value (polar coordinates)
    
    float trustVx_; ///< Certainty of the X value
    float trustVy_; ///< Certainty of the Y value
    float uncertainty_; ///< Uncertainty associated with this object
};
#endif
