/**
 * @Filename Velocity3D.cpp
 * Velocity3D Data class implementation.
 * @Status Implementing
 * @Version $Id: Velocity3D.cpp 5661 2012-05-11 13:43:55Z jreis $
 * @Maintainer Hugo Costelha
 */
#include "Velocity3D.hpp"

#include <FrameTransformations2D.hpp>
#include <FrameTransformations3D.hpp>
#include <sstream>



Velocity3D::Velocity3D (void) : Data(),
  vX_ (0), vY_ (0), vZ_ (0), speed_ (0), azimuth_ (0), zenith_ (0),
  speedInXYPlane_ (0), uncertainty_ (0), trustVx_ (1), trustVy_ (1), trustVz_ (1)



{
}



Velocity3D::Velocity3D (string dataString) : Data(),
  vX_ (0), vY_ (0), vZ_ (0), speed_ (0), azimuth_ (0), zenith_ (0),
  speedInXYPlane_ (0), uncertainty_ (0), trustVx_ (1), trustVy_ (1), trustVz_ (1)
{
  const char* data;
  data = dataString.c_str();
  float inSpeed = 0, inAngleAzimuth = 0, inAngleZenith = 0;
  
  sscanf (data, "(%f, %f, %f)", &inSpeed, &inAngleAzimuth, &inAngleZenith);
  setSpherical (inSpeed, inAngleAzimuth, inAngleZenith);
}



Velocity3D::Velocity3D (Speed speed, Angle azimuth, Angle zenith,
                        bool isCartesian) : Data(),
  vX_(), vY_(), vZ_(), speed_(), azimuth_(), zenith_(), speedInXYPlane_(),
  uncertainty_ (0.0), trustVx_ (1.0), trustVy_ (1.0), trustVz_ (1.0)
{
  if (isCartesian) {
    setVxVyVz (speed, azimuth, zenith);
  }
  else {
    setSpherical (speed, azimuth, zenith);
  }
}



Velocity3D::~Velocity3D (void)
{
}



Speed Velocity3D::getVx() const
{
  return vX_;
}



Speed Velocity3D::getVy() const
{
  return vY_;
}



Speed Velocity3D::getVz() const
{
  return vZ_;
}



Speed Velocity3D::getSpeed() const
{
  return speed_;
}



Speed Velocity3D::getSpeedInXYPlane() const
{
  return speedInXYPlane_;
}



Angle Velocity3D::getAzimuth() const
{
  return azimuth_;
}



Angle Velocity3D::getZenith() const
{
  return zenith_;
}



void Velocity3D::setVxVyVz (Speed vX, Speed vY, Speed vZ)
{
  vX_ = vX;
  vY_ = vY;
  vZ_ = vZ;
  updateSpherical_();
}



void Velocity3D::setVx (Speed vX)
{
  setVxVyVz (vX, vY_, vZ_);
}



void Velocity3D::setVy (Speed vY)
{
  setVxVyVz (vX_, vY, vZ_);
}



void Velocity3D::setVz (Speed vZ)
{
  setVxVyVz (vX_, vY_, vZ);
}



void Velocity3D::setSpherical (Speed speed, Angle azimuth, Angle zenith)
{
  speed_ = speed;
  azimuth_ = azimuth;
  zenith_ = zenith;
  
  // We need to update the rho in the XYPlane so, we update the
  // cartesian coordinates, and then compute that rho.
  updateCartesian_();
  Angle temp;
  FrameTransformations2D::rect2Polar (vX_, vY_,
                                      &speedInXYPlane_, &temp);
}



void Velocity3D::setSpeed (Speed speed)
{
  setSpherical (speed, azimuth_, zenith_);
}



void Velocity3D::setAzimuth (Angle azimuth)
{
  setSpherical (speed_, azimuth, zenith_);
}



void Velocity3D::setZenith (Angle zenith)
{
  setSpherical (speed_, azimuth_, zenith);
}



void Velocity3D::updateCartesian_()
{
  FrameTransformations3D::spherical2Cartesian (speed_, azimuth_, zenith_,
      &vX_, &vY_, &vZ_);
}



void Velocity3D::updateSpherical_()
{
  FrameTransformations3D::cartesian2Spherical (vX_, vY_, vZ_,
      &speed_, &azimuth_, &zenith_);
  // Update also the rho value in the XYPlane
  Angle temp;
  FrameTransformations2D::rect2Polar (vX_, vY_,
                                      &speedInXYPlane_, &temp);
}



void Velocity3D::setUncertainty (float value)
{
  uncertainty_ = value;
}



float Velocity3D::getUncertainty() const
{
  return uncertainty_;
}



void Velocity3D::setTrust (float trustVx, float trustVy, float trustVz)
{
  trustVx_ = trustVx;
  trustVy_ = trustVy;
  trustVz_ = trustVz;
}



float Velocity3D::getTrustVx() const
{
  return trustVx_;
}



float Velocity3D::getTrustVy() const
{
  return trustVy_;
}



float Velocity3D::getTrustVz() const
{
  return trustVz_;
}



shared_ptr<Parameter> Velocity3D::clone() const
{
  shared_ptr<Velocity3D> tmp (new Velocity3D (*this));
  return tmp;
}



string Velocity3D::toString() const
{
  stringstream ret;
  ret << "(" << getSpeed() << ", "
      << getAzimuth() << ", "
      << getZenith() << ")";
      
  return ret.str();
}
