/**
 * @Filename Velocity3D.hpp
 * Velocity3D Data class header.
 * @Status Implementing
 * @Version $Id: Velocity3D.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _VELOCITY_3D_HPP_
#define _VELOCITY_3D_HPP_

#include <Data.hpp>
#include <Defines.hpp>

#include <socrob_ros/Velocity3D.h>



/**
 * @Class Velocity3D Velocity3D.hpp "Velocity3D.hpp"
 * This class is used to store 3 dimentions floats, aimed mainly to store x,y,z
 * (or rho,zenith,azimuth) coordinates in a local frame.
 * @Author Hugo Costelha
 */
class Velocity3D : public Data
{
  public:
    /**
     * Constructor.
     * This method creates a new Velocity3D.
     */
    Velocity3D (void);
    
    /**
     * Constructor.
     * This method creates a new Velocity3D from a string. We assume spherical
     * coordinates here.
     * @Argument stringData - string with rho, azimuth and zenith values
     */
    Velocity3D (string stringData);
    
    /**
     * Constructor.
     * This method creates a new Velocity3D using cartesian or spherical
     * coordinates.
     * @Argument rho - rho value of the spherical coordinates, or, if using
     * cartesian coordinates, the x value.
     * @Argument azimuth - azimuth value of the spherical coordinates (measured
     * from the x axis), or, if using cartesian coordinates, the y value .
     * @Argument zenith - zenith value of the spherical coordinates (measured
     * from the z axis), or, if using cartesian coordinates, the z value.
     * @Argument isCartesian - (optional argument) if false (default), the
     * arguments are used as spherical coordinates, otherwise they are used as
     * cartesian coordinates.
     */
    Velocity3D (Speed speed, Angle azimuth, Angle zenith,
                bool isCartesian = false);
                
    /**
     * Destructor.
     * This method makes deinitializations for objects of this class.
     */
    ~Velocity3D (void);
    
    /**
     * This method retrieves the x value of the velocity in cartesian
     * coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getVx() const;
    
    /**
     * This method retrieves the y velocity value in cartesian coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getVy() const;
    
    /**
     * This method retrieves the z velocity value in cartesian coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getVz() const;
    
    /**
     * This method retrieves the velocity rho value in spherical coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getSpeed() const;
    
    /**
     * This method retrieves the velocity rho value in spherical coordinates,
     * on the XY plane (ignoring the z value).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Speed getSpeedInXYPlane() const;
    
    /**
     * This method retrieves the azimuth, measured from the x-axis in the XY
     * plane (spherical coordinates). This angle is the same has in cartesian
     * coordinates.
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Angle getAzimuth() const;
    
    /**
     * This method retrieves the angle zenith, measured from the z-axis
     * (spherical coordinates).
     * @Returns value held by this Data, 0.0 if uninitialized.
     */
    Angle getZenith() const;
    
    /**
     * This method sets the velocity values in cartesian coordinates.
     * @Argument vx - the velocity X value to store.
     * @Argument vy - the velocity Y value to store.
     * @Argument vz - the velocity Z value to store.
     */
    void setVxVyVz (Speed vx, Speed vy, Speed vz);
    
    /**
     * This method sets the x velocity value (cartesian coordinates).
     * @Argument vx - the velocity X value to store.
     */
    void setVx (Speed vx);
    
    /**
     * This method sets the y velocity value (cartesian coordinates).
     * @Argument vy - the velocity Y value to store.
     */
    void setVy (Speed vy);
    
    /**
     * This method sets the z velocity value (cartesian coordinates).
     * @Argument vz - the velocity Z value to store.
     */
    void setVz (Speed vz);
    
    /**
     * This method sets the value in spherical coordinates.
     * @Argument rho - the rho value to store.
     * @Argument azimuth - the angle from the X axis (in XY plane)
     * @Argument zentih - the angle from the z axis.
     */
    void setSpherical (Speed speed, Angle azimuth, Angle zenith);
    
    /**
     * This method sets the speed velocity value (spherical coordinates).
     * @Argument speed - the velocity speed value to store.
     */
    void setSpeed (Speed speed);
    
    /**
     * This method sets the azimuth value (spherical coordinates).
     * @Argument azimuth - the angle from the X axis (in XY plane)
     */
    void setAzimuth (Angle azimuth);
    
    /**
     * This method sets the zenith value (spherical coordinates).
     * @Argument zentih - the angle from the z axis.
     */
    void setZenith (Angle zenith);
    
    /**
     * Change the uncertainty associated with this object. 0 means no
     * uncertainty, while 1 means full uncertainty.
     * @Argument value - new uncertainty value.
     */
    void setUncertainty (float value);
    
    /**
     * Retrieve the uncertainty
     */
    float getUncertainty() const;
    
    /**
     * Change the cercainty associated with each component. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument trustX - certainty of the X component.
     * @Argument trustY - certainty of the Y component.
     * @Argument trustTheta - certainty of the Theta component.
     */
    void setTrust (float trustVx, float trustVy, float trustVtheta);
    
    /**
     * Retrieve the X value certainty.
     */
    float getTrustVx() const;
    
    /**
     * Retrieve the Y value certainty.
     */
    float getTrustVy() const;
    
    /**
     * Retrieve the Theta value certainty.
     */
    float getTrustVz() const;
    
    shared_ptr<Parameter> clone() const;
    
    /**
     * Get the cartesian coordinated in string format.
     * @return string with the cartesian coodinates.
     */
    string toString() const;
    
    void rosMsg (socrob_ros::Velocity3D& msg) const {
      msg.vX = vX_;
      msg.vY = vY_;
      msg.vZ = vZ_;
      msg.speed = speed_;
      msg.azimuth = azimuth_;
      msg.zenith = zenith_;
      msg.speedInXYPlane = speedInXYPlane_;
      
      msg.uncertainty = uncertainty_;
      msg.trustVx = trustVx_;
      msg.trustVy = trustVy_;
      msg.trustVz = trustVz_;
    }
    
    void fromRosMsg (const socrob_ros::Velocity3D& msg) {
      vX_ = msg.vX;
      vY_ = msg.vY;
      vZ_ = msg.vZ;
      speed_ = msg.speed;
      azimuth_ = msg.azimuth;
      zenith_ = msg.zenith;
      speedInXYPlane_ = msg.speedInXYPlane;
      
      uncertainty_ = msg.uncertainty;
      trustVx_ = msg.trustVx;
      trustVy_ = msg.trustVy;
      trustVz_ = msg.trustVz;
    }
    
    static shared_ptr<Velocity3D> newFromRosMsg (const socrob_ros::Velocity3D& msg) {
      shared_ptr<Velocity3D> tmp = shared_ptr<Velocity3D> (new Velocity3D ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
  private:
    /**
     * Update the stored cartesian values using the spherical ones.
     */
    void updateCartesian_();
    
    /**
     * Update the stored spherical values using the cartesian ones.
     */
    void updateSpherical_();
    
    float vX_;  ///< velocity X value (cartesian coordinates)
    float vY_;  ///< velocity Y value (cartesian coordinates)
    float vZ_;  ///< velocity Z value (cartesian coordinates)
    float speed_;  ///< Rho value (shperical coordinates)
    float azimuth_;  ///< Angle value in XY plane (spherical coordinates)
    float zenith_;  ///< Angle value (shperical coordinates)
    float speedInXYPlane_;  ///< Rho value (polar coordinates in XY plane)
    
    float uncertainty_; ///< Uncertainty associated with this object
    float trustVx_; ///< Certainty of the X value
    float trustVy_; ///< Certainty of the Y value
    float trustVz_; ///< Certainty of the Z value
};
#endif
