/**
 * @Filename VelocityPosture.cpp
 * @Description  VelocityPosture class implementation
 * @Status Finished
 * @Version $Id: VelocityPosture.cpp 5864 2012-12-08 14:40:54Z jreis $
 * @Maintainer Hugo Costelha
 */

#include "VelocityPosture.hpp"
#include <math.h>
#include <socrob/math.h>

#include <sstream>

#include <LocalFrameWorldFrameTransformations.hpp>



VelocityPosture::VelocityPosture() : Data(),
  vel2D_(), vtheta_ (0.0), uncertainty_ (0.0), trustVTheta_ (1.0)
{
}



VelocityPosture::VelocityPosture (string dataString) : Data(),
  vel2D_(), vtheta_ (0.0), uncertainty_ (0.0), trustVTheta_ (1.0)
{
  const char* data;
  data = dataString.c_str();
  float inSpeed = 0.0, inAzimuth = 0.0, inVTheta = 0.0;
  
  sscanf (data, "(%f,%f,%f)", &inSpeed, &inAzimuth, &inVTheta);
  
  setPolar (inSpeed, inAzimuth, inVTheta);
}



VelocityPosture::VelocityPosture (float speed, float azimuth, float vangular,
                                  bool isRect) : Data(),
  vel2D_ (speed, azimuth, isRect), vtheta_ (vangular), uncertainty_ (0.0),
  trustVTheta_ (1.0)
{
}



VelocityPosture::~VelocityPosture()
{
}



Speed VelocityPosture::getVx() const
{
  return vel2D_.getVx();
}



Speed VelocityPosture::getVy() const
{
  return vel2D_.getVy();
}



Speed VelocityPosture::getSpeed() const
{
  return vel2D_.getSpeed();
}



Angle VelocityPosture::getAzimuth() const
{
  return vel2D_.getAzimuth();
}



Rotation_Speed VelocityPosture::getAngularSpeed() const
{
  return vtheta_;
}



void VelocityPosture::setPolar (Speed speed, Angle azimuth,
                                Rotation_Speed angularSpeed)
{
  setPolar (speed, azimuth);
  setAngularSpeed (angularSpeed);
}



void VelocityPosture::setPolar (Speed speed, Angle azimuth)
{
  vel2D_.setPolar (speed, azimuth);
}



void VelocityPosture::setSpeed (Speed speed)
{
  vel2D_.setSpeed (speed);
}



void VelocityPosture::setAzimuth (Angle azimuth)
{
  vel2D_.setAzimuth (azimuth);
}



void VelocityPosture::setVxVyVang (Speed vx, Speed vy,
                                   Rotation_Speed vtheta)
{
  setVxVy (vx, vy);
  setAngularSpeed (vtheta);
}



void VelocityPosture::setVxVy (Speed vx, Speed vy)
{
  vel2D_.setVxVy (vx, vy);
}



void VelocityPosture::setVx (Speed vx)
{
  vel2D_.setVx (vx);
}



void VelocityPosture::setVy (Speed vy)
{
  vel2D_.setVy (vy);
}



void VelocityPosture::setAngularSpeed (Rotation_Speed vtheta)
{
  vtheta_ = vtheta;
}



void VelocityPosture::setUncertainty (float value)
{
  uncertainty_ = value;
}



void VelocityPosture::setTrust (float trustVx, float trustVy, float trustVtheta)
{
  vel2D_.setTrust (trustVx, trustVy);
  trustVTheta_ = trustVtheta;
}



float VelocityPosture::getTrustVx() const
{
  return vel2D_.getTrustVx();
}



float VelocityPosture::getTrustVy() const
{
  return vel2D_.getTrustVy();
}



float VelocityPosture::getTrustVtheta() const
{
  return trustVTheta_;
}



float VelocityPosture::getUncertainty() const
{
  return uncertainty_;
}



shared_ptr<Parameter> VelocityPosture::clone() const
{
  shared_ptr<VelocityPosture> newVelocityPosture (new VelocityPosture (*this));
  return newVelocityPosture;
}



string VelocityPosture::toString() const
{
  stringstream ret;
  ret << "(" << getSpeed() << ","
      << getAzimuth() << ","
      << getAngularSpeed() << ")";
      
  return ret.str();
}
