/**
 * @Filename VelocityPosture.hpp
 * @Description VelocityPosture class header.
 * @Status Finished
 * @Version $Id: VelocityPosture.hpp 5750 2012-07-05 01:50:42Z jreis $
 * @Maintainer Hugo Costelha
 */

#ifndef _VELOCITY_POSTURE_HPP_
#define _VELOCITY_POSTURE_HPP_

#include <Data.hpp>
#include <Parameter.hpp>

#include <Defines.hpp>
#include <Velocity2D.hpp>

#include <socrob_ros/VelocityPosture.h>



/**
 * @Class VelocityPosture VelocityPosture.hpp "VelocityPosture.hpp"
 * This class is used to store postures (x,y,theta) values. Typically, the
 * stored posture is in local coorditates. This class also provides methods to
 * return the posture value in world/local coordinates.
 * @Author Hugo Costelha
 */
class VelocityPosture : public Data
{
  public:
    /**
     * Class base constructor (everything is set to 0)
     */
    VelocityPosture();
    
    /**
     * Create the class from a string.
     * @Argument dataString - String with data in the format "(%f, %f, %f)",
     * with Velocity and Angle (in polar coordinates), followed by theta.
     */
    VelocityPosture (string dataString);
    
    /**
     * Create a new posture with the given values in SI coordinates.
     * @Argument velocity - rho value (or x if rectangular coordinates)
     * @Argument angle - angle value (or y if rectangular coordinates)
     * @Argument vangular - Angular velocity
     * @Argument isRect - (optional) if false (default), then the arguments
     * refer to polar coordinates, otherwise they refer to rectangular
     * coordinates.
     */
    VelocityPosture (float speed, float azimuth, float vangular,
                     bool isRect = false);
                     
    /**
     * Destructor
     */
    ~VelocityPosture();
    
    /**
     * Retrieve the velocity X value (rectangular coordinates)
     */
    Speed getVx() const;
    
    /**
     * Retrieve the velocity Y value (rectangular coordinates)
     */
    Speed getVy() const;
    
    /**
     * Retreive the velocity rho value (polar coordinates)
     */
    Speed getSpeed() const;
    
    /**
     * Retrieve the azimuth angle (polar coordinates)
     */
    Angle getAzimuth() const;
    
    /**
     * Retrieve the angular velocity
     */
    Rotation_Speed getAngularSpeed() const;
    
    /**
     * Change the X,Y, Theta velocity values based on the given polar
     * coordinates.
     * @Argument speed - the velocity rho value in polar coordinates.
     * @Argument azimuth - the velocity angle in polar coordinate.
     * @Argument vangular - new angular velocity value.
     */
    void setPolar (Speed speed, Angle azimuth, Rotation_Speed vangular);
    
    /**
     * Change the X,Y velocity values based on the given polar coordinates.
     * @Argument speed - the velocity rho value in polar coordinates.
     * @Argument azimuth - the velocity angle in polar coordinate.
     */
    void setPolar (Speed speed, Angle azimuth);
    
    /**
     * Change the velocity value (polar coordinates).
     * @Argument speed - the rho value in polar coordinates.
     */
    void setSpeed (Speed speed);
    
    /**
     * Change the velocity angle value (polar coordinates).
     * @Argument azimuth - the new angle polar coordinate.
     */
    void setAzimuth (Angle azimuth);
    
    /**
     * Change the X,Y and Theta velocity  values.
     * @Argument vx - The new X velocity value.
     * @Argument vy - The new Y velocity value.
     * @Argument vangular - The new angular velocity value.
     */
    void setVxVyVang (Speed vx, Speed vy, Rotation_Speed vangular);
    
    /**
     * Change the X,Y velocity values (cartesian coordinates).
     * @Argument vx - The new X velocity value.
     * @Argument vy - The new Y velocity value.
     */
    void setVxVy (Speed vx, Speed vy);
    
    /**
     * Change the X velocity value (cartesian coordinates).
     * @Argument vx - The new X velocity value.
     */
    void setVx (Speed vx);
    
    /**
     * Change the Y velocity value (rectangulaar coordinates).
     * @Argument vy - The new Y velocity value.
     */
    void setVy (Speed vy);
    
    /**
     * Change the angular velocity value.
     * @Argument vangular - The angular velocity value.
     */
    void setAngularSpeed (Rotation_Speed vangular);
    
    /**
     * Retrieve the uncertainty
     */
    float getUncertainty() const;
    
    /**
     * Retrieve the X value certainty.
     */
    float getTrustVx() const;
    
    /**
     * Retrieve the Y value certainty.
     */
    float getTrustVy() const;
    
    /**
     * Retrieve the Theta value certainty.
     */
    float getTrustVtheta() const;
    
    /**
     * Change the uncertainty associated with this vector. 0 means no
     * uncertainty, while 1 means full uncertainty.
     * @Argument value - new uncertainty value.
     */
    void setUncertainty (float value);
    
    /**
     * Change the cercainty associated with each component. 0 means no
     * confidence, while 1 means full confidence.
     * @Argument trustVx - certainty of the X component.
     * @Argument trustVy - certainty of the Y component.
     * @Argument trustVTheta - certainty of the Theta component.
     */
    void setTrust (float trustVx, float trustVy, float trustVtheta);
    
    shared_ptr<Parameter> clone() const;
    
    /**
     * Return the velocity value in Velocity,Angle,Theta form, where velocity
     * and angle are in polar coordinates.
     * @Returns the string with the posture value in the Velocity, Angle,
     * Theta form.
     */
    string toString() const;
    
    void rosMsg (socrob_ros::VelocityPosture& msg) const {
      vel2D_.rosMsg (msg.vel2D);
      msg.vtheta = vtheta_;
      msg.uncertainty = uncertainty_;
      msg.trustVTheta = trustVTheta_;
    }
    
    void fromRosMsg (const socrob_ros::VelocityPosture& msg) {
      vel2D_.fromRosMsg (msg.vel2D);
      vtheta_ = msg.vtheta;
      uncertainty_ = msg.uncertainty;
      trustVTheta_ = msg.trustVTheta;
    }
    
    static shared_ptr<VelocityPosture> newFromRosMsg (const socrob_ros::VelocityPosture& msg) {
      shared_ptr<VelocityPosture> tmp = shared_ptr<VelocityPosture> (new VelocityPosture ());
      tmp->fromRosMsg (msg);
      return tmp;
    }
    
  private:
    Velocity2D vel2D_; ///< 2D velocity
    float vtheta_; ///< Angular velocity
    float uncertainty_; ///< Uncertainty associated with this object
    float trustVTheta_; ///< Certainty of the angular velocity
};

#endif
