/**
 * @Filename WorldInfo.cpp
 * WorldInfo class header.
 * @Status Implementing
 * @Version $Id: WorldInfo.cpp 6062 2013-05-23 09:14:21Z jreis $
 * @Maintainer Joao Reis <joaocgreis@ist.utl.pt>
 */

#include "WorldInfo.hpp"

#include <string>
#include <sstream>
#include <list>
#include <map>
#include <algorithm>

#include <boost/thread/mutex.hpp>

#include <Boolean.hpp>
#include <DataArray.hpp>
#include <DataMap.hpp>
#include <Float.hpp>
#include <FullPosture.hpp>
#include <Integer.hpp>
#include <Obstacle.hpp>
#include <Position2D.hpp>
#include <Position3D.hpp>
#include <Posture.hpp>
#include <String.hpp>
#include <TrackedPosition2D.hpp>
#include <TrackedPosition3D.hpp>
#include <TrackedPosture.hpp>
#include <Velocity2D.hpp>
#include <Velocity3D.hpp>
#include <VelocityPosture.hpp>

#include <SocRobService.hpp>

#include <ros/ros.h>
#include <std_msgs/Bool.h>
#include <std_msgs/String.h>
#include <socrob_ros/FullPostureStamped.h>
#include <socrob_ros/OtherRobotsInfoStamped.h>
#include <socrob_ros/TrackedPosition3DStamped.h>
#include <socrob_ros/Position3DStamped.h>
#include <socrob_ros/VelocityPostureStamped.h>
#include <socrob_ros/PostureStamped.h>
#include <socrob_ros/ObstaclesStamped.h>



namespace
{
  boost::mutex worldinfo_mutex;
  
  void registerCallbacks();
  
  typedef map<string, shared_ptr<Data const> > cache_t;
  typedef pair<string, shared_ptr<Data const> > cache_pair_t;
  cache_t cache;
  
  map<string, ros::Publisher> publishers;
  
  ros::Publisher&
  getPublisher (string const& dataName)
  {
    map<string, ros::Publisher>::iterator pub = publishers.find (dataName);
    
    if (pub == publishers.end()) {
      cache_t::iterator data_pos = cache.find (dataName);
      if (data_pos == cache.end()) {
        ROS_ERROR_STREAM (ros::this_node::getName()
                          << ": Cannot create publisher for "
                          << dataName
                          << ". It does not exist.");
        return publishers[dataName];
      }
      shared_ptr<Data const> data = data_pos->second;
      
      if (typeid (*data) == typeid (String)) {
        publishers[dataName] = SocRobService::nh().advertise<std_msgs::String> ("world_info/" + dataName, 1, true);
      }
      else if (typeid (*data) == typeid (Boolean)) {
        publishers[dataName] = SocRobService::nh().advertise<std_msgs::Bool> ("world_info/" + dataName, 1, true);
      }
      else if (typeid (*data) == typeid (Position3D)) {
        publishers[dataName] = SocRobService::nh().advertise<socrob_ros::Position3DStamped> ("world_info/" + dataName, 1, true);
      }
      else if (typeid (*data) == typeid (Posture)) {
        publishers[dataName] = SocRobService::nh().advertise<socrob_ros::PostureStamped> ("world_info/" + dataName, 1, true);
      }
      else if (typeid (*data) == typeid (TrackedPosition3D)) {
        publishers[dataName] = SocRobService::nh().advertise<socrob_ros::TrackedPosition3DStamped> ("world_info/" + dataName, 1, true);
      }
      else if (typeid (*data) == typeid (VelocityPosture)) {
        publishers[dataName] = SocRobService::nh().advertise<socrob_ros::VelocityPostureStamped> ("world_info/" + dataName, 1, true);
      }
      else if (typeid (*data) == typeid (Obstacles)) {
        publishers[dataName] = SocRobService::nh().advertise<socrob_ros::ObstaclesStamped> ("world_info/" + dataName, 1, true);
      }
      else if ("OtherRobotsInfoMap" == dataName) {
        publishers[dataName] = SocRobService::nh().advertise<socrob_ros::OtherRobotsInfoStamped> ("world_info/" + dataName, 1, true);
      }
      else {
        ROS_ERROR_STREAM ("Found unimplemented type "
                          << typeid (*data).name()
                          << " while initializing publishers in World Info.");
      }
    }
    
    return publishers[dataName];
  }
  
  map<string, set<Receiver<MessageWorldInfoUpdateData>*> > registeredServices;
}



void
WorldInfo::
initialize()
{
  boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
  
  /*
  awk 'BEGIN {FS="[<>]"} /Name/ {print $5}' WorldInfoInit.xml > VARNAMES
  awk 'BEGIN {FS="[<>]"} /Type/ {print $9}' WorldInfoInit.xml > VARTYPES
  awk 'BEGIN {FS="[<>]"} /Value/ {print $13}' WorldInfoInit.xml > VARINIT
  sed -i 's/\./_/g' VARNAMES
  paste VARTYPES VARNAMES VARINIT | sort | awk '{print "cache[\""$2"\"] = shared_ptr<Data const> (new "$1" (\""$3"\"));"}'
   */
  cache["Accept"] = shared_ptr<Data const> (new Boolean (false));
  cache["BallOurs"] = shared_ptr<Data const> (new Boolean (false));
  cache["CollisionDetected"] = shared_ptr<Data const> (new Boolean (false));
  cache["Commited"] = shared_ptr<Data const> (new Boolean (false));
  cache["Confirm"] = shared_ptr<Data const> (new Boolean (false));
  cache["DynamicPassEnded"] = shared_ptr<Data const> (new Boolean (false));
  cache["FoulTaken"] = shared_ptr<Data const> (new Boolean (false));
  cache["isRunning"] = shared_ptr<Data const> (new Boolean (false));
  cache["KickerOn"] = shared_ptr<Data const> (new Boolean (false));
  cache["LongPassAccept"] = shared_ptr<Data const> (new Boolean (false));
  cache["LongPassAvailable"] = shared_ptr<Data const> (new Boolean (false));
  cache["LongPass"] = shared_ptr<Data const> (new Boolean (false));
  cache["OMNI1_seeGaussBall"] = shared_ptr<Data const> (new Boolean (false));
  cache["OMNI2_seeGaussBall"] = shared_ptr<Data const> (new Boolean (false));
  cache["OMNI3_seeGaussBall"] = shared_ptr<Data const> (new Boolean (false));
  cache["OMNI4_seeGaussBall"] = shared_ptr<Data const> (new Boolean (false));
  cache["OMNI5_seeGaussBall"] = shared_ptr<Data const> (new Boolean (false));
  cache["PassDone"] = shared_ptr<Data const> (new Boolean (false));
  cache["Ready2Receive"] = shared_ptr<Data const> (new Boolean (false));
  cache["Reject"] = shared_ptr<Data const> (new Boolean (false));
  cache["Request"] = shared_ptr<Data const> (new Boolean (false));
  cache["ShortPassAccept"] = shared_ptr<Data const> (new Boolean (false));
  cache["ShortPass"] = shared_ptr<Data const> (new Boolean (false));
  cache["ShortPassRequest"] = shared_ptr<Data const> (new Boolean (false));
  cache["OtherRobotsInfoMap"] = shared_ptr<Data const> (new DataMap ());
  cache["MyFullPosture"] = shared_ptr<Data const> (new FullPosture ());
  cache["Obstacles"] = shared_ptr<Data const> (new Obstacles ());
  cache["OMNI1_ballGaussParams"] = shared_ptr<Data const> (new Position3D ("(0,0,0)"));
  cache["OMNI2_ballGaussParams"] = shared_ptr<Data const> (new Position3D ("(0,0,0)"));
  cache["OMNI3_ballGaussParams"] = shared_ptr<Data const> (new Position3D ("(0,0,0)"));
  cache["OMNI4_ballGaussParams"] = shared_ptr<Data const> (new Position3D ("(0,0,0)"));
  cache["OMNI5_ballGaussParams"] = shared_ptr<Data const> (new Position3D ("(0,0,0)"));
  cache["opponentGoalPosture"] = shared_ptr<Data const> (new Posture ("(9.07,0,0)"));
  cache["ownGoalPosture"] = shared_ptr<Data const> (new Posture ("(-9.07,0,0)"));
  cache["ShootingSpot"] = shared_ptr<Data const> (new Posture ("(1,0,0)"));
  cache["CurrentRole"] = shared_ptr<Data const> (new String ("PlayingDead"));
  cache["GameState"] = shared_ptr<Data const> (new String ("STOPPED"));
  cache["InternalEvent"] = shared_ptr<Data const> (new String ("NONE"));
  cache["PartnerRobot"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameClosestToBall"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameFourthClosestToBall"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameSecondClosestToBall"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameSecondClosestToMiddleField"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameSecondClosestToOppGoal"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameSecondClosestToOwnGoal"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameShouldIGoToOppLeftCorner"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameShouldIGoToOppRightCorner"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameShouldIGoToOwnLeftCorner"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameShouldIGoToOwnRightCorner"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotNameThirdClosestToBall"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RobotThatSentOlderRequest"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RunningBehavior"] = shared_ptr<Data const> (new String ("Unknown"));
  cache["RunningPrimitiveActions"] = shared_ptr<Data const> (new String ("_this_string_is_here_because_of_a_bug_"));
  cache["BallFullPosition"] = shared_ptr<Data const> (new TrackedPosition3D ());
  cache["sharedBallFullPosition"] = shared_ptr<Data const> (new TrackedPosition3D ());
  cache["MyVelocity"] = shared_ptr<Data const> (new VelocityPosture ("(0,0,0)"));
  
  registerCallbacks();
}



void
WorldInfo::
registerService (string dataName,
                 Receiver<MessageWorldInfoUpdateData>* service)
{
  boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
  
  std::replace (dataName.begin(), dataName.end(), '.', '_');
  
  cache_t::iterator data_pos = cache.find (dataName);
  if (data_pos == cache.end()) {
    ROS_ERROR_STREAM (ros::this_node::getName()
                      << ": World Info: registerService: "
                      << dataName
                      << " does not exist.");
    return;
  }
  
  registeredServices[dataName].insert (service);
}



void
WorldInfo::
putData (string dataName,
         shared_ptr<Data const> const& data)
{
  boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
  
  std::replace (dataName.begin(), dataName.end(), '.', '_');
  
  cache_t::iterator data_pos = cache.find (dataName);
  if (data_pos == cache.end()) {
    ROS_ERROR_STREAM (ros::this_node::getName()
                      << ": Cannot insert "
                      << dataName
                      << " in World Info. If you are creating something new, use a ROS Topic directly.");
    return;
  }
  
  if (typeid (*data) != typeid (* (data_pos->second))) {
    ROS_ERROR_STREAM (ros::this_node::getName()
                      << ": WorldInfo: Data "
                      << dataName
                      << " is of type "
                      << typeid (* (data_pos->second)).name()
                      << ". Cannot insert a "
                      << typeid (*data).name()
                      << " in its place.");
    return;
  }
  
  if (typeid (*data) == typeid (String)) {
    std_msgs::String msg;
    msg.data = dynamic_pointer_cast<String const> (data)->stringValue();
    getPublisher (dataName).publish (msg);
  }
  else if (typeid (*data) == typeid (Boolean)) {
    std_msgs::Bool msg;
    msg.data = dynamic_pointer_cast<Boolean const> (data)->boolValue();
    getPublisher (dataName).publish (msg);
  }
  else if (typeid (*data) == typeid (Position3D)) {
    socrob_ros::Position3DStamped msg;
    msg.header.stamp = ros::Time::now();
    dynamic_pointer_cast<Position3D const> (data)->rosMsg (msg.position3D);
    getPublisher (dataName).publish (msg);
  }
  else if (typeid (*data) == typeid (Posture)) {
    socrob_ros::PostureStamped msg;
    msg.header.stamp = ros::Time::now();
    dynamic_pointer_cast<Posture const> (data)->rosMsg (msg.posture);
    getPublisher (dataName).publish (msg);
  }
  else if (typeid (*data) == typeid (TrackedPosition3D)) {
    socrob_ros::TrackedPosition3DStamped msg;
    msg.header.stamp = ros::Time::now();
    dynamic_pointer_cast<TrackedPosition3D const> (data)->rosMsg (msg.trackedPosition3D);
    getPublisher (dataName).publish (msg);
  }
  else if (typeid (*data) == typeid (VelocityPosture)) {
    socrob_ros::VelocityPostureStamped msg;
    msg.header.stamp = ros::Time::now();
    dynamic_pointer_cast<VelocityPosture const> (data)->rosMsg (msg.velocityPosture);
    getPublisher (dataName).publish (msg);
  }
  else if (typeid (*data) == typeid (Obstacles)) {
    socrob_ros::ObstaclesStamped msg;
    msg.header.stamp = ros::Time::now();
    dynamic_pointer_cast<Obstacles const> (data)->rosMsg (msg.obstacles);
    getPublisher (dataName).publish (msg);
  }
  else if ("OtherRobotsInfoMap" == dataName) {
    socrob_ros::OtherRobotsInfoStamped msg;
    msg.header.stamp = ros::Time::now();
    pair<string, shared_ptr<Data> > p;
    foreach (p, dynamic_pointer_cast<DataMap> (data->clone())->mapValue()) {
      socrob_ros::RobotInfo robotInfo;
      shared_ptr<DataMap> ri = dynamic_pointer_cast<DataMap> (p.second);
      robotInfo.name = p.first;
      robotInfo.id = dynamic_pointer_cast<Integer> (ri->getElement ("id"))->intValue();
      robotInfo.seeBall = dynamic_pointer_cast<Boolean> (ri->getElement ("seeBall"))->boolValue();
      dynamic_pointer_cast<Position3D> (ri->getElement ("BallLocalPosition"))->rosMsg (robotInfo.BallLocalPosition);
      dynamic_pointer_cast<Posture> (ri->getElement ("ownGoalPosture"))->rosMsg (robotInfo.ownGoalPosture);
      dynamic_pointer_cast<Posture> (ri->getElement ("opponentGoalPosture"))->rosMsg (robotInfo.opponentGoalPosture);
      dynamic_pointer_cast<Posture> (ri->getElement ("RobotPosture"))->rosMsg (robotInfo.RobotPosture);
      robotInfo.CurrentRole = dynamic_pointer_cast<String> (ri->getElement ("CurrentRole"))->stringValue();
      robotInfo.RunningBehavior = dynamic_pointer_cast<String> (ri->getElement ("RunningBehavior"))->stringValue();
      robotInfo.isRunning = dynamic_pointer_cast<Boolean> (ri->getElement ("isRunning"))->boolValue();
      robotInfo.BallOurs = dynamic_pointer_cast<Boolean> (ri->getElement ("BallOurs"))->boolValue();
      robotInfo.FoulTaken = dynamic_pointer_cast<Boolean> (ri->getElement ("FoulTaken"))->boolValue();
      robotInfo.PassDone = dynamic_pointer_cast<Boolean> (ri->getElement ("PassDone"))->boolValue();
      robotInfo.DynamicPassEnded = dynamic_pointer_cast<Boolean> (ri->getElement ("DynamicPassEnded"))->boolValue();
      robotInfo.Ready2Receive = dynamic_pointer_cast<Boolean> (ri->getElement ("Ready2Receive"))->boolValue();
      robotInfo.LongPassAvailable = dynamic_pointer_cast<Boolean> (ri->getElement ("LongPassAvailable"))->boolValue();
      robotInfo.LongPassAccept = dynamic_pointer_cast<Boolean> (ri->getElement ("LongPassAccept"))->boolValue();
      robotInfo.ShortPassRequest = dynamic_pointer_cast<Boolean> (ri->getElement ("ShortPassRequest"))->boolValue();
      robotInfo.ShortPassAccept = dynamic_pointer_cast<Boolean> (ri->getElement ("ShortPassAccept"))->boolValue();
      robotInfo.Request = dynamic_pointer_cast<Boolean> (ri->getElement ("Request"))->boolValue();
      robotInfo.Accept = dynamic_pointer_cast<Boolean> (ri->getElement ("Accept"))->boolValue();
      robotInfo.Reject = dynamic_pointer_cast<Boolean> (ri->getElement ("Reject"))->boolValue();
      robotInfo.Confirm = dynamic_pointer_cast<Boolean> (ri->getElement ("Confirm"))->boolValue();
      robotInfo.Commited = dynamic_pointer_cast<Boolean> (ri->getElement ("Commited"))->boolValue();
      msg.robotsInfo.push_back (robotInfo);
    }
    getPublisher (dataName).publish (msg);
  }
  else {
    ROS_ERROR_STREAM (ros::this_node::getName() << ": WorldInfo: Unimplemented putData for " << dataName);
    return;
  }
  
  // Published without problems, but it may still be needed in the caller before ROS has opportunity to spin.
  cache[dataName] = data;
}



shared_ptr<Data const>
WorldInfo::
getData (string dataName)
{
  boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
  
  std::replace (dataName.begin(), dataName.end(), '.', '_');
  
  cache_t::iterator data_pos = cache.find (dataName);
  if (data_pos == cache.end()) {
    ROS_ERROR_STREAM (ros::this_node::getName()
                      << ": World Info: getData:  "
                      << dataName
                      << " not present in World Info.");
    return shared_ptr<Data const>();
  }
  
  return data_pos->second;
}



namespace
{
  void
  notifyUpdateData (string const& name, shared_ptr<Data const> data)
  {
    foreach (Receiver<MessageWorldInfoUpdateData>* s, registeredServices[name]) {
      s->updateData (name, data);
    }
  }
  
  void
  StringCallback (string const& name, const std_msgs::String::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = shared_ptr<String> (new String (msg->data));
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  BoolCallback (string const& name, const std_msgs::Bool::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = shared_ptr<Boolean> (new Boolean (msg->data));
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  Position3DCallback (string const& name, const socrob_ros::Position3DStamped::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = Position3D::newFromRosMsg (msg->position3D);
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  PostureCallback (string const& name, const socrob_ros::PostureStamped::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = Posture::newFromRosMsg (msg->posture);
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  TrackedPosition3DCallback (string const& name, const socrob_ros::TrackedPosition3DStamped::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = TrackedPosition3D::newFromRosMsg (msg->trackedPosition3D);
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  VelocityPostureCallback (string const& name, const socrob_ros::VelocityPostureStamped::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = VelocityPosture::newFromRosMsg (msg->velocityPosture);
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  FullPostureCallback (string const& name, const socrob_ros::FullPostureStamped::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = FullPosture::newFromRosMsg (msg->fullPosture);
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  ObstaclesCallback (string const& name, const socrob_ros::ObstaclesStamped::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      cache[name] = Obstacles::newFromRosMsg (msg->obstacles);
    }
    notifyUpdateData (name, cache[name]);
  }
  void
  OtherRobotsInfoCallback (string const& name, const socrob_ros::OtherRobotsInfoStamped::ConstPtr& msg)
  {{
      boost::lock_guard<boost::mutex> _ (worldinfo_mutex);
      shared_ptr<DataMap> tmp = shared_ptr<DataMap> (new DataMap ());
      foreach (socrob_ros::RobotInfo robotInfo, msg->robotsInfo) {
        shared_ptr<DataMap> ri = shared_ptr<DataMap> (new DataMap ());
        ri->addElement ("id", shared_ptr<Integer> (new Integer (robotInfo.id)));
        ri->addElement ("seeBall", shared_ptr<Boolean> (new Boolean (robotInfo.seeBall)));
        ri->addElement ("BallLocalPosition", Position3D::newFromRosMsg (robotInfo.BallLocalPosition));
        ri->addElement ("ownGoalPosture", Posture::newFromRosMsg (robotInfo.ownGoalPosture));
        ri->addElement ("opponentGoalPosture", Posture::newFromRosMsg (robotInfo.opponentGoalPosture));
        ri->addElement ("RobotPosture", Posture::newFromRosMsg (robotInfo.RobotPosture));
        ri->addElement ("CurrentRole", shared_ptr<String> (new String (robotInfo.CurrentRole)));
        ri->addElement ("RunningBehavior", shared_ptr<String> (new String (robotInfo.RunningBehavior)));
        ri->addElement ("isRunning", shared_ptr<Boolean> (new Boolean (robotInfo.isRunning)));
        ri->addElement ("BallOurs", shared_ptr<Boolean> (new Boolean (robotInfo.BallOurs)));
        ri->addElement ("FoulTaken", shared_ptr<Boolean> (new Boolean (robotInfo.FoulTaken)));
        ri->addElement ("PassDone", shared_ptr<Boolean> (new Boolean (robotInfo.PassDone)));
        ri->addElement ("DynamicPassEnded", shared_ptr<Boolean> (new Boolean (robotInfo.DynamicPassEnded)));
        ri->addElement ("Ready2Receive", shared_ptr<Boolean> (new Boolean (robotInfo.Ready2Receive)));
        ri->addElement ("LongPassAvailable", shared_ptr<Boolean> (new Boolean (robotInfo.LongPassAvailable)));
        ri->addElement ("LongPassAccept", shared_ptr<Boolean> (new Boolean (robotInfo.LongPassAccept)));
        ri->addElement ("ShortPassRequest", shared_ptr<Boolean> (new Boolean (robotInfo.ShortPassRequest)));
        ri->addElement ("ShortPassAccept", shared_ptr<Boolean> (new Boolean (robotInfo.ShortPassAccept)));
        ri->addElement ("Request", shared_ptr<Boolean> (new Boolean (robotInfo.Request)));
        ri->addElement ("Accept", shared_ptr<Boolean> (new Boolean (robotInfo.Accept)));
        ri->addElement ("Reject", shared_ptr<Boolean> (new Boolean (robotInfo.Reject)));
        ri->addElement ("Confirm", shared_ptr<Boolean> (new Boolean (robotInfo.Confirm)));
        ri->addElement ("Commited", shared_ptr<Boolean> (new Boolean (robotInfo.Commited)));
        tmp->addElement (robotInfo.name, ri);
      }
      cache[name] = tmp;
    }
    notifyUpdateData (name, cache[name]);
  }
  
  
  
  void
  registerCallbacks()
  {
    foreach (cache_pair_t const & i, cache) {
      if (typeid (* (i.second)) == typeid (String)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const std_msgs::String::ConstPtr&) > (boost::bind (StringCallback, i.first, _1))));
      }
      else if (typeid (* (i.second)) == typeid (Boolean)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const std_msgs::Bool::ConstPtr&) > (boost::bind (BoolCallback, i.first, _1))));
      }
      else if (typeid (* (i.second)) == typeid (Position3D)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const socrob_ros::Position3DStamped::ConstPtr&) > (boost::bind (Position3DCallback, i.first, _1))));
      }
      else if (typeid (* (i.second)) == typeid (Posture)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const socrob_ros::PostureStamped::ConstPtr&) > (boost::bind (PostureCallback, i.first, _1))));
      }
      else if (typeid (* (i.second)) == typeid (TrackedPosition3D)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const socrob_ros::TrackedPosition3DStamped::ConstPtr&) > (boost::bind (TrackedPosition3DCallback, i.first, _1))));
      }
      else if (typeid (* (i.second)) == typeid (VelocityPosture)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const socrob_ros::VelocityPostureStamped::ConstPtr&) > (boost::bind (VelocityPostureCallback, i.first, _1))));
      }
      else if (typeid (* (i.second)) == typeid (FullPosture)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const socrob_ros::FullPostureStamped::ConstPtr&) > (boost::bind (FullPostureCallback, i.first, _1))));
      }
      else if (typeid (* (i.second)) == typeid (Obstacles)) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const socrob_ros::ObstaclesStamped::ConstPtr&) > (boost::bind (ObstaclesCallback, i.first, _1))));
      }
      else if ("OtherRobotsInfoMap" == i.first) {
        new ros::Subscriber (SocRobService::nh().subscribe ("world_info/" + i.first, 1, *new boost::function<void (const socrob_ros::OtherRobotsInfoStamped::ConstPtr&) > (boost::bind (OtherRobotsInfoCallback, i.first, _1))));
      }
      else {
        ROS_ERROR_STREAM ("Found unimplemented type "
                          << typeid (* (i.second)).name()
                          << " while initializing callbacks in World Info.");
      }
    }
  }
}
